//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/*
Apple Macintosh Developer Technical Support

MultiFinder-Aware SoundApp Application

SoundAppSnds.r - MPW Rez Source

Jim Reekes - Macintosh Developer Technical Support
Copyright  1989-1994 Apple Computer, Inc.
All rights reserved.

Versions:
		1.03				January, 1990
		1.04				Sept, 1990
		1.2					August, 1994		translated to C

Components:
		SoundApp.c			January, 1990		MPW C source code
		SoundUnit.c			January, 1990		MPW C source code
		SoundUnit.h			January, 1990		MPW C source code
		SoundApp.r			January, 1990		MPW Rez source code
		SoundAppSnds.r		January, 1990		MPW Rez source code
		SoundApp.make		January, 1990		MPW build script

Formatting was done with FONT = Courier or Monaco, SIZE = 10, TABS = 4

SoundApp.p is a sample application source file for demonstrating the
Sound Manager.  It requires the use of the SoundUnit to handle all of
the sound routines.  This portion of the source code is the sources to
the 'snd ' resources for Rez.  Its not what I would want to use
continually as a snd compiler to say the least.  What I needed was an
easy way to build snd resources that simple contained a sequence of
sound commands.  I wanted to transcribed some given music, and didnt
want to enter all the data using hex. I did build a ResEdit template,
but even that was more gross that the method below.  I would really
like to see a MIDI file to snd file exchange utility.  So heres what
I came up with.

I first had to define all the possible frequencies that could be used in a
freqDurationCmd.  These are the 127 MIDI values.  There are 12 keys that
cover 11 octaves.  By using the system below, I can define the key I
want (such as D sharp) and then specify the octave (such as the octave
that has D sharp next to middle C, or the 6th octave). This helped me
to enter the frequencies by keeping in the proper key signature and range.
I wouldnt use this system to enter any music beyond the 19th century.

The next problem was to come up with a scheme for describing the
durations.  The Sound Manager is based on 2000 beats
per second.  Standard music is based on a number of beats per minute.
Use the number of Sound Manager beats in a minute and divide by the
number of beats you want per minute.  90 on the metronome is a typical
tempo.  To get this tempo, use (kSMBeatsPerMinute / 90).  Disco is
(kSMBeatsPerMinute / 120).

The result of this calculation will give you the duration of a
quarter note in 4/4 time.  If you wanted to play an eighth note, take
the songs beat and divide it by 2.  To get a whole note, multiply by
4.  If youre still with me or interested in this process, youre a sick
person.  I also found that putting the measure numbers in to the
sequence helped to keep things straight while looking at the music
manuscript.  If you want to change the tempo of one of the songs
below, just change the #define that calculates the beat duration.

Once I entered all the frequencies and got through a session with MPWs
Rez, I had to compare my results with the original music.  I had good
dictation drills while attending music courses at the university, so I
was prepared for this next part.  I had to listen to the output of the
Sound Manager and circle the incorrect note on the manuscript.  Go
back to MPW and correct the Rez source and start over.  Ive found
that restCmds dont cause the channel to be quiet unless preceded by a
quietCmd.  A quietCmd has no duration. At least tempos can easily be
adjusted.  If you want to transpose the key signature, you could write
a sound channel modifier when the Sound Manager supports them.

Jim Reekes E.O., Macintosh Developer Technical Support
Sunday, August 7, 1994 7:06:41 PM
*/
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// INCLUDES
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include "Types.r"
#include "SysTypes.r"


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// CONSTANTS
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define kOctave1		0					// octaves of MIDI values
#define kOctave2		12
#define kOctave3		24
#define kOctave4		36
#define kOctave5		48
#define kOctave6		60
#define kOctave7		72
#define kOctave8		84
#define kOctave9		96
#define kOctave10		108
#define kOctave11		120

#define Akey			-3					// the A key
#define Bbkey			-2					// the B flat key
#define Bkey			-1					// the B key
#define Ckey			0					// the C key
#define Dbkey			1					// the D flat key
#define Dkey			2					// the D key
#define Ebkey			3					// the D flat key
#define Ekey			4					// the E key
#define Fkey			5					// the F key
#define Gbkey			6					// the G flat key
#define Gkey			7					// the G key
#define Abkey			8					// the A flat key


// Using the constant below, I can define a tempo of 60 beats per minutes
// by using the following expressions: (kSMBeatsPerMinute / 60).

#define kSMBeatsPerMinute (2000	* 60)	// Sound Mgr beats per minute

// This will define the tempo for the rScaleSnd.
#define kScaleBeat	(kSMBeatsPerMinute / 240)

// This will define the tempo for the rMelodyPart1 - 4.
#define kMelodyBeat	(kSMBeatsPerMinute / 90)

// This will define the tempo for the rCounterPt1 - 4.
#define kBachBeat		(kSMBeatsPerMinute / 60)

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// RESOURCE ID NUMBERS
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// snd resource must be in the range 8192 or greater
#define rMoofSound		9000			// This is the about sound
#define rScaleSnd		9001			// snd containing a scale
#define rMelodyPart1	9002			// snd containing a melody
#define rMelodyPart2	9003			// snd containing the harmony
#define rMelodyPart3	9004			// snd containing the harmony
#define rMelodyPart4	9005			// snd containing the harmony
#define rWaveHarmony	9006			// snd containing waveTable for harmony
#define rWaveMelody		9007			// snd containing waveTable for melody
#define rCounterPt1		9008			// snd containing soprano part of counter point
#define rCounterPt2		9009			// snd containing alto part of counter point
#define rCounterPt3		9010			// snd containing tenor part of counter point
#define rCounterPt4		9011			// snd containing bass part of counter point
#define rSopranoVox		9012			// snd containing waveTable of soprano voice
#define rAltoVox		9013			// snd containing waveTable of alto voice
#define rTenorVox		9014			// snd containing waveTable of tenor voice
#define rBassVox		9015			// snd containing waveTable of bass voice

#define rSampleHarmony	9016			// snd containing the sampled sound harmony
#define rSampleMelody	9017			// snd containing the sampled sound melody


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// RESOURCE DEFINITIONS
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// This snd is used for holding the waveTable used in harmony sounds.
// Unfortunately, MPWs rIncludes doesnt properly define the complete snd
// resource format (who can blame them) and therefore defining a waveTable
// has to be done in the general data format.  The Sound Manager will re-sample
// a wave table to 512 bytes.  Realizing this, a wave table sounds best at
// 512 or even 256 bytes.

data 'snd ' (rWaveHarmony, "wave harmony", purgeable) {
	$"0001"					// format 1 snd
	$"0001"					// number of synths to be installed
	$"0003"					// snth to be used
	$"00000000"				// init option for synth
	$"0001"					// number of sound cmds
	$"803C 0100 00000014"	// waveTableCmd with dataPtr bit, 256 bytes, offset 20 bytes
							// the wave table data
	$"807F7C7A 78757371 706E6D6C 6B6A6A6B 6B6D6E71 73777A7E 82868B90 959A9FA4"
	$"A9AEB2B7 BBBFC1C5 C8CBCED1 D3D6D8DA DCDEE1E3 E5E7EAEC EEF0F3F5 F7F8FAFB"
	$"FDFEFEFE FFFEFEFD FCFAF8F6 F4F1EEEB E8E4E1DD D9D5D1CD C8C4BFBB B6B1ACA7"
	$"A29D9893 8E8A8682 7F7C7976 7472706F 6E6D6D6D 6D6E6F70 71737576 787A7D7F"
	$"80818386 888A8B8D 8F909192 93939393 9291908E 8C8A8784 817E7A76 726D6863"
	$"5E59544F 4A45413C 38332F2B 27231F1C 1815120F 0C0A0806 04030202 01020202"
	$"03050608 090B0D10 12141619 1B1D1F22 2426282A 2D2F3235 383B3F41 45494E52"
	$"575C6166 6B70757A 7E828689 8D8F9293 95959696 95949392 908F8D8B 88868481"
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// This snd is used for holding the waveTable used in melody sounds.

data 'snd ' (rWaveMelody, "wave melody", purgeable) {
	$"0001"					// format 1 snd
	$"0001"					// number of synths to be installed
	$"0003"					// snth to be used
	$"00000000"				// init option for synth
	$"0001"					// number of sound cmds
	$"803C 0100 00000014"	// waveTableCmd with dataPtr bit, 256 bytes, offset 20 bytes
							// the wave table data
	$"808A9399 9D9E9A93 897C6C5B 49392B20 19161920 2D3E526A 829BB3CA DDECF8FE"
	$"FFFBF3E7 D8C6B39F 8B796757 4A3F3630 2B282728 292B2E32 363C4248 5059626D"
	$"7884909C A8B4BFC9 D2D9DFE3 E6E8E8E6 E4E1DEDA D6D1CDC8 C4BFBAB5 B0AAA59E"
	$"98918A82 7B736B63 5B544C45 3E37312C 27221F1C 1B1B1D20 252C343E 49566372"
	$"808E9DAA B7C2CCD4 DBE0E3E5 E5E4E1DE D9D4CFC9 C2BBB4AC A59D958D 857E766F"
	$"68625B56 504B4641 3C38332F 2A26221F 1C1A1818 1A1D2127 2E37414C 5864707C"
	$"88939EA7 B0B8BEC4 CACED2D5 D7D8D9D8 D5D0CAC1 B6A99987 75614D3A 28190D05"
	$"01020814 23364D65 7E96AEC2 D3E0E7EA E7E0D5C7 B7A59484 776D6662 63676D76"
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// This snd is used for holding the waveTable used as the soprano voice.

data 'snd ' (rSopranoVox, "Soprano", purgeable) {
	$"0001"					// format 1 snd
	$"0001"					// number of synths to be installed
	$"0003"					// snth to be used
	$"00000000"				// init option for synth
	$"0001"					// number of sound cmds
	$"803C 0100 00000014"	// waveTableCmd with dataPtr bit, 256 bytes, offset 20 bytes
							// the wave table data
	$"80899198 9FA3A4A4 A39F9890 887E746B 635E5A57 575B606A 7481909F ADBCCBD7"
	$"E0E8EEEF EEE8E0D7 CABCAC9D 8D7E7266 5E575454 585F6773 808F9FAD BDCBD8E4"
	$"EEF6FBFE FFFEFBF7 F2EBE4DF D8D3CFCC CACACBCC D0D4DADF E6EBF0F6 FAFCFEFE"
	$"FCF8F3EC E4DACFC2 B5A7998B 7C6F6256 4B433B36 322F2F30 343A4048 525C6773"
	$"808D99A4 AEB8C0C6 CCD0D1D1 CECAC5BD B5AA9E91 84756759 4B3E3126 1C140D08"
	$"04020204 060A1015 1A21262C 30343536 3634312D 28211C15 0E090502 0102050A"
	$"121C2835 43536171 808D99A1 A8ACACA9 A29A8E82 73635444 36292018 12111218"
	$"20293544 5361707F 8C96A0A5 A9A9A6A2 9D958C82 78706861 5D5C5C5D 61686F77"
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// This snd is used for holding the waveTable used as the alto voice.

data 'snd ' (rAltoVox, "Alto", purgeable) {
	$"0001"					// format 1 snd
	$"0001"					// number of synths to be installed
	$"0003"					// snth to be used
	$"00000000"				// init option for synth
	$"0001"					// number of sound cmds
	$"803C 0100 00000014"	// waveTableCmd with dataPtr bit, 256 bytes, offset 20 bytes
							// the wave table data
	$"807A746E 69635D57 524C4742 3E3B3835 34333435 373A3E43 494E545B 61676C71"
	$"75787A7C 7C7B7976 736F6B66 625E5B58 56565759 5C61676E 77808A95 A0ACB7C2"
	$"CCD6DFE7 EEF4F8FB FEFFFFFE FDFBF9F6 F3F0ECE9 E6E3E0DD DAD7D4D1 CECAC6C2"
	$"BDB8B3AD A7A19A93 8D867F79 736D6863 5F5C5A58 5757585A 5C5F6367 6B70757B"
	$"80858B90 95999DA1 A4A6A8A9 A9A8A6A4 A19D9893 8D87817A 736D665F 59534D48"
	$"433E3A36 322F2C29 2623201D 1A171410 0D0A0705 03020101 0205080C 1219212A"
	$"343E4954 606B7680 8992999F A4A7A9AA AAA8A5A2 9E9A9591 8D8A8785 84848688"
	$"8B8F9499 9FA5ACB2 B7BDC2C6 C9CBCCCD CCCBC8C5 C2BEB9B4 AEA9A39D 97928C86"
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// This snd is used for holding the waveTable used as the tenor voice.

data 'snd ' (rTenorVox, "Tenor", purgeable) {
	$"0001"					// format 1 snd
	$"0001"					// number of synths to be installed
	$"0003"					// snth to be used
	$"00000000"				// init option for synth
	$"0001"					// number of sound cmds
	$"803C 0100 00000014"	// waveTableCmd with dataPtr bit, 256 bytes, offset 20 bytes
							// the wave table data
	$"8081807F 7B756D63 57493B2D 20150C05 02010206 0B111820 28313A44 4D57616A"
	$"737A8084 86878684 817D7873 6E69635E 58534D49 45434242 44474C52 59616A74"
	$"7F8B97A4 B1BDC8D1 D8DDE0E0 DFDDDAD8 D6D5D6D9 DCE1E6EB F0F3F6F7 F7F5F1EC"
	$"E6DFD7CF C6BFB7B1 ABA59F99 91897F73 66594B3F 352D2928 2B313943 4F5B6774"
	$"808C99A5 B1BDC7CF D5D8D7D3 CBC1B5A7 9A8D8177 6F67615B 554F4941 3A312921"
	$"1A140F0B 09090A0D 10151A1F 24272A2B 2A282623 21202023 282F3843 4F5C6975"
	$"818C969F A7AEB4B9 BCBEBEBD BBB7B3AD A8A29D97 928D8883 7F7C7A79 7A7C8086"
	$"8D969FA9 B3BCC6CF D8E0E8EF F5FAFEFF FEFBF4EB E0D3C5B7 A99D938B 8581807F"
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// This snd is used for holding the waveTable used as the bass voice.

data 'snd ' (rBassVox, "Bass", purgeable) {
	$"0001"					// format 1 snd
	$"0001"					// number of synths to be installed
	$"0003"					// snth to be used
	$"00000000"				// init option for synth
	$"0001"					// number of sound cmds
	$"803C 0100 00000014"	// waveTableCmd with dataPtr bit, 256 bytes, offset 20 bytes
							// the wave table data
	$"807B7671 6B655F58 50484038 3028211A 15110E0C 0D0E1216 1C232B33 3C454E56"
	$"5D646A6F 73767879 7A7B7B7C 7C7D7F82 85898E94 9BA3ABB3 BCC4CDD5 DDE4EBF0"
	$"F5F9FCFE FFFFFFFE FCFAF7F4 F0ECE8E3 DED9D3CE C8C2BCB6 B0A9A49E 98938E89"
	$"85817D79 76726F6C 6965625F 5B585451 4E4C4A48 4848484A 4D51555B 61687078"
	$"80889098 9FA5ABAF B3B6B8B8 B8B8B6B4 B2AFACA8 A5A19E9B 9794918E 8A87837F"
	$"7B77726D 68625C57 504A443E 38322D27 221D1814 100C0906 04020101 01020407"
	$"0B10151C 232B333C 444D555D 656C7277 7B7E8183 84848585 8687888A 8D91969C"
	$"A3AAB2BB C4CDD5DD E4EAEEF2 F3F4F2EF EBE6DFD8 D0C8C0B8 B0A8A19B 958F8A85"
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// This snd is used for holding the sample sound used as the harmony voice.

resource 'snd ' (rSampleHarmony, "sample harmony", purgeable) {
	FormatOne {
		{sampledSynth, 0}
	},
	{hasData,bufferCmd {20}		// buffer data offset is 20 bytes
	},
	{11553,						// number of samples
	 0x2B7745D1,				// sample rate
	 11551,						// loop starting point
	 11552,						// loop ending point
	 0,							// standard sound header
	 60,						// base frequency
		$"727474727372716F6D6C6866696A71818E9CB0B9BEC3BFB7AFA3978C827A736F"
		$"6D6A6A6D6B6D7270727673737673737372716F6D6B67676A6D798895A8B7BDC3"
		$"C3BCB4AA9D92877E77706F6C696C6C6B7071707474727574727372716F6D6C68"
		$"66696B73828F9FB1BBC0C4C0B8B0A4988D837B74706F6A6A6D6B6D716F717472"
		$"7274727272706F6D6C696567696D7A8996A9B7BEC3C3BCB4AA9D92877F78726F"
		$"6C696C6B6A6F6F6E72727072727070706F6D6B6A6664676A7282909FB1BBC1C4"
		$"C0B8AFA3978D837B74706E6A6A6C6A6C6F6E70727070727070706F6D6B6A6764"
		$"66696D7A8997A9B8BEC4C3BCB4A99D91877F78726F6C696B6A696D6D6D70706F"
		$"706F6E6F6E6D6B696764636669728290A0B2BBC2C5C0B8AFA3978D837C75716F"
		$"6A6A6C696B6D6C6E706F6F706E6E6E6D6C6968666364676D7B8A98ABB8C0C4C3"
		$"BCB4A99D92887F7873706D6A6C6A696C6C6C6F6E6D6F6E6D6D6C6C6967666362"
		$"6669738290A2B3BCC3C5C0B8AFA3978D847C76726F6B6B6B696B6D6C6D6F6D6E"
		$"6E6D6C6C6B6A6766646164676D7B8A99ACB9C0C5C3BCB4A99D9288817974726D"
		$"6B6C6A6A6C6C6C6E6D6D6D6D6C6C6B6A68666461616569738291A2B4BDC3C5C0"
		$"B8AEA3978D857E7773706C6C6C6A6B6C6B6D6E6C6D6D6C6C6B6A696665636063"
		$"666D7C8A9AACBAC1C6C3BCB4A99D938A827B76726F6C6D6A6A6C6C6C6D6C6C6C"
		$"6C6B6B6A6966656360616469748392A3B4BEC4C5C0B8AEA3988E857E7874706D"
		$"6D6C6A6C6C6B6C6D6C6C6C6B6A6A69676564616063666E7C8B9BADBAC1C6C3BC"
		$"B3A99D938A837C77736F6D6D6B6B6D6C6C6D6C6C6C6C6A6A6968666462606064"
		$"69758493A5B4BEC4C5C0B8AEA3988E877F7975726F6D6C6B6C6C6B6D6D6C6C6C"
		$"6A6A6968666463605F62666E7C8B9CAEBAC2C6C3BBB3A89D938B847D7875706F"
		$"6F6C6C6D6C6C6D6C6C6C6B6A6A69676563615F606369758494A6B5BEC4C4BFB7"
		$"AEA3998F88817B77736F6F6D6C6D6D6C6D6D6C6C6B6A6A69676563615F5E6166"
		$"6F7D8C9DAFBBC2C5C2BBB2A89E948B857E7976726F6F6D6D6D6C6D6D6C6C6C6B"
		$"6A6968666462605E5F6369758594A6B5BFC4C4BEB7ADA3999088827C78747070"
		$"6E6C6D6D6C6D6D6C6C6B6A6968676462605E5D60656F7D8D9EAFBBC2C5C1BBB2"
		$"A89E948C857F7B7673706F6D6D6D6D6D6D6C6C6C6A6A69676563615F5D5E6269"
		$"768595A7B5BFC4C4BEB6ADA399908A827C797572706F6D6D6D6C6D6D6C6C6A6A"
		$"69676663615F5D5D60646F7D8D9EAFBBC1C4C1BAB1A89E958D87807C78747270"
		$"6E6D6E6D6D6D6C6C6B6A6967666461605D5B5E6169768596A8B6BEC3C3BDB5AC"
		$"A39A918A847E7B7673726F6E6E6D6D6D6D6C6B6A6968666562605E5B5B5E646F"
		$"7D8D9FAFBAC1C3C0B9B1A79E968E87817C797573716F6F6F6D6D6D6C6C6B6A69"
		$"676663615E5C5B5D6169768596A8B5BEC3C2BDB5ACA39A928B857F7C78747370"
		$"6F6F6F6D6D6D6C6C6A69676664615F5D5B5B5E646F7E8D9FAFBAC0C2BEB8B0A7"
		$"9F968E88827E7A76747270706F6E6E6D6D6C6B6A69676563605D5B5A5B606976"
		$"8596A8B5BEC2C1BBB4ABA39A938C86817D7976747270706F6F6F6D6C6C6A6967"
		$"6664615E5B5A5A5D636F7E8D9FAFBAC0C1BDB7AFA69E968F89837F7B77757372"
		$"70706F6F6E6D6C6C6A686664615E5C5A585B6068768596A8B5BDC1BFBAB3ABA2"
		$"9A938D87827E7A7775737272706F6F6E6D6C6B69676563605D5B58595C636F7E"
		$"8E9FAEB9BFC0BCB6AFA69E96908A84807C787674727271706F6F6D6C6C6A6866"
		$"64615E5B58585A5E68768596A7B4BCC0BEB9B2AAA29A938D87827F7B78767372"
		$"727170706F6D6C6B696765625F5C5A57585B636F7E8E9FAEB8BEBEBBB5AEA69E"
		$"96908A85817D79787573737271706F6E6D6B6A686663605D5A5757595E677585"
		$"96A7B3BBBEBDB8B1A9A29A948E88847F7C79777573737271706F6D6C6A696764"
		$"615E5B5857575B626F7E8E9FAEB7BDBDBAB4ADA59E97918B86827E7B78767574"
		$"737271706F6D6C6A686563605C595655585E67758596A6B2BABDBBB7B0A9A29A"
		$"948E8985817D7A78767574737271706F6D6B696764615D5A5755565A626F7E8E"
		$"9FADB7BBBCB9B4ACA59E97928C87847F7C7A787675747372706F6E6C6A686562"
		$"5E5B585555585D67768596A6B1B8BBBAB5AFA8A19A948F8A85827E7B79777675"
		$"737372706F6D6B696664605D595654555A626E7E8E9FACB5BABAB7B2ACA59E98"
		$"928D8884817D7B79787675747372706F6C6A6865625E5A575454575D67768596"
		$"A5B1B7BAB8B4AEA8A19A94908A86827F7C7B787776757373716F6D6C69666360"
		$"5C585554555A626E7E8E9EABB4B8B9B6B1ABA49D97928D8885827E7C7A787876"
		$"757472716F6D6A6865615E5A565454575D67768596A5AFB5B8B7B2ADA7A09A94"
		$"908A8784807E7B79787776757372706E6C6966635F5B5754525459626F7E8D9D"
		$"AAB2B7B7B4AFAAA39D97928D8985827F7D7B79787776757372706D6B6865615D"
		$"59555353565D68768596A4AEB5B7B5B1ACA6A09A95908B8784817E7C7B797876"
		$"767472716F6C6A66635F5A5754525459626F7E8D9DA9B1B5B6B3AEA9A39D9793"
		$"8E8A8683807E7C7B797876767472706E6C6965615D58555252565D69768696A3"
		$"ADB4B5B4B1ABA6A09A95908C8885817F7D7B7A7978767573726F6D6A67635E5A"
		$"5653515458626F7E8D9CA8B0B4B4B2ADA8A29C97938E8A8784817F7C7B7B7978"
		$"767573716F6C6965615C58545252565D69768695A3ACB2B4B3AFABA59F9A9590"
		$"8D898582817E7C7B7A7978767572706D6A67635E5A5652515459626F7E8D9CA7"
		$"AFB2B3B1ACA7A29C97938E8B8784827F7E7C7B7A79787674726F6C6965615C57"
		$"545152565D69768594A2ABB1B2B1AEA9A49F9995918D8A8683817F7E7C7B7A79"
		$"777573716E6B67635E595552515459636F7E8D9BA6ADB1B1AFABA6A09B97938E"
		$"8B888482817E7D7C7B7978767572706D6965605B57545152565D69768594A0A9"
		$"AFB1AFACA8A39E9994908D8A868482807E7D7C7B79787674726F6C67635E5955"
		$"5151545963707E8C9AA5ACAFAFADA9A5A09A96928E8B888582817F7E7D7C7B79"
		$"787673706D6A65605B57535152565E697685939FA8AEAFAEABA6A29D9894908C"
		$"8A878482817F7E7D7C7B79787573706C68635E59555251545A63707E8C99A3AB"
		$"AEAEACA8A39F9A96918E8B88858381807F7E7C7B7A797675726E6A66615B5753"
		$"5152575E6A7785939FA7ACADACA9A5A09C9794908C8A87848281807F7E7C7C7B"
		$"797674716D69645E59555252545A64717E8C99A3A9ACACABA6A29D9995918E8B"
		$"88858482817F7F7E7C7B7A7875736F6B66615B57535152575F6A7885939DA5AB"
		$"ACABA8A49F9B97938F8C8987848382807F7F7E7C7B797775726D69645F5A5552"
		$"52555B65717E8C99A2A8ABABA9A5A19D9894918D8A8885848281807F7E7E7C7B"
		$"797673706C67615C5753515358606A7885919DA4A9AAA9A6A29E9A96928E8B89"
		$"8685848281807F7E7E7C7B7876726E69645F5A555252555C66727E8C97A0A6A9"
		$"A9A7A4A09C9794908D8A878584838281807F7E7E7C7A7875706C67615C575452"
		$"5458606B7885919CA3A7A8A7A5A19D9994918E8B88868584828181807F7F7D7C"
		$"7977736F6A655F5A565353565C66717E8B969FA5A8A8A5A29F9A9693908C8A87"
		$"85848382828181807F7E7B7976726D68635D5854525459616B7784909AA1A5A6"
		$"A6A39F9C9794918D8A888685848282818181807F7D7B7875706B66605B575454"
		$"575D66717D8A949DA3A5A6A3A09D9995928E8B8A878584848282828181807F7D"
		$"7B78736F6A645E595554555A616C76838F999FA3A5A4A29E9A9693908D8A8886"
		$"8584838282828181807E7C7976726C67615B575555575D66717D88939CA1A4A4"
		$"A29F9C9894918E8A8887858484838282828282817E7C7975706B655F5A575557"
		$"5B626C77828E979EA2A3A2A09D9995918E8B898785848483828282828281807E"
		$"7B78736E69625D585555585E66717C88929A9FA2A2A09D9A9693908C8A888685"
		$"8483838383828282817F7E7B76726C66605B5756575B636C76828D969DA0A2A1"
		$"9F9B9794908D8A888785848483838383838382817F7C79746F69635E5A575759"
		$"5E67717C8791999EA0A09F9C9995918E8B88878584848383838384848483817F"
		$"7C78736D67615C5857585C636C76818C949B9FA09F9D9996938F8C8987868484"
		$"838283848484848483817E7B76716B645F5B58575A5F67707B8690979C9F9F9D"
		$"9A9793908C8A8786848383828384848485858483817E79756F69635E5A58595D"
		$"636C76818B939A9D9E9D9B9894918D8A88868584838383848485858585858280"
		$"7C78726C66615C59595B6067707B858F969B9D9D9B9895918E8B888685848282"
		$"828384858586868685827F7B76706A645F5B5A5A5D646C76808A92989C9C9C99"
		$"96938F8B89878584828282838485858787878785827E79756E68635E5B5A5C61"
		$"68707B858E94999C9B999693908C898785848282828283848586878787868481"
		$"7D78726C66615D5B5B5F646D767F8991969A9A9A9794908D8A87858482828282"
		$"83848586888888888784807B76706A64605C5C5D6269717A848D939799999794"
		$"918E8A8785838281818182838485878888888886837F79736D67625E5C5D6065"
		$"6D757F88909598999895928E8B88858382818181818284858788898A8A888582"
		$"7D78726B66615E5D5E6369727A848B9296989896938F8C888583828180808181"
		$"82848587898A8A898785807B756F6964605E5E61666D757F878E949697969390"
		$"8D8986848281807F808182838587888A8B8B8A87847F79736D6763605E60646A"
		$"727A838B9195969694918E8A878482817F7F7F8081828486888A8B8B8B898682"
		$"7D77706A6661606062676D767E878E93959694918E8B878582817F7E7E7F8081"
		$"838587898B8C8C8B8985807B756F6964616061646A727B838A9094959493908C"
		$"888582817F7E7E7E7F81828486888A8C8D8C8B88847E79726C6763616163686E"
		$"767E868D91949493908D898683817F7E7D7D7E7F818385888A8C8D8D8D8A8782"
		$"7C76706A66636163666B727A828A8F939493918E8A8784817F7E7D7C7D7E7F82"
		$"8487898B8D8E8E8C8A85807A746E6965636364696F767E858C909393918E8B88"
		$"84817F7D7C7C7C7D7E808285888A8D8E8E8E8B88847E78726C68646364676C72"
		$"7A82898E9192918F8C8885827F7D7C7B7B7C7E7F8184878A8C8E8F8F8E8B8782"
		$"7C76706B67646465696F767E858B8F91918F8D8985827F7D7B7B7A7B7C7D7F82"
		$"85888B8D8F908F8D8A857F79736E69666465676D737A81888D9091908D8A8783"
		$"807D7B7A79797B7C7E8184878A8D8F90908F8D88837E78726D696666676A7076"
		$"7E858A8E90908E8B8884817E7B7A7979797B7C7F8285888B8E9090908E8B8681"
		$"7B75706C686667696D737A81888D8F908E8C8985827E7C7A797878797B7D8084"
		$"878A8D909191908E8A857F79736F6B6867686C71777E858A8E8F8F8D8A87827F"
		$"7C7A78787878797B7E8185888B8E9191918F8C87827C77726D6968686A6E747A"
		$"81878C8E8E8D8B8784807D7A78777777787A7C7F83878A8D909292918F8B8681"
		$"7B75706C6A696A6D72787E858A8D8E8E8C8985817E7B7877767677787B7D8185"
		$"888C8F919392908D89847E78736F6C6A6A6B6F757B81878B8E8E8C8A86827F7B"
		$"797776757677797B7F82878A8E91939392908C87827C76726E6B6A6B6E72787E"
		$"858A8D8D8D8B8884807C79777675757677797D8184888C90929393918E8A857F"
		$"7A75706D6B6B6C70757B81878B8D8D8B8885817D7A777574747475787B7E8286"
		$"8A8E91939393908D88837E78736F6D6C6C6F73787F84898B8D8C8A86827F7B78"
		$"767473737476787C7F84888C90939494928F8B87817B76726F6D6C6D71767C82"
		$"878A8C8C8A87847F7C7876737272737476797D81858A8E91939494918E8A847F"
		$"7975716F6D6D7074797F84898B8C8B8885817D7976747372727375787B7F8488"
		$"8D9093949493908D88827D7874706F6E6F72767C82878A8B8B8A87827E7B7774"
		$"727171727375787C81858A8E91949494928E8A85817B7773706F6F71757A7F84"
		$"898B8C8A8884807C7875737170707273767A7E82878C9093959594918D89847F"
		$"7A767270707173787D82878A8B8B8986827E797673717070707274787B80858A"
		$"8E9294959593908B87827D787572707072767B8085898B8B8A87847F7B777471"
		$"706F6F707275797D82878B9093959594928E8A85807C787472717275797E8388"
		$"8A8B8B8985817D797572706F6E6F7073767A7F84898E9194959593908D88837E"
		$"7A7673727273777B8185898B8B8A87837E7A7673706F6D6D6F7173787C81868B"
		$"9093959594928F8B86817D7975737273767A7F84888A8B8B8885807C7874706F"
		$"6D6D6D6F7275797E84888D9194959593918D8984807B7875737375787C818689"
		$"8B8B8987827E7975726F6D6C6C6D6F72767B80868B8F9395959493908C87837E"
		$"7A77757375777B7F84888B8B8B8884807B76736F6D6C6C6C6D7074787E83888D"
		$"9194959694918E8A85817D7976757576797E82878A8B8B8986817D7874706D6C"
		$"6A6A6C6E72757A7F858A8F9395969593908D8884807C78767576787C8085888B"
		$"8B8A88847F7A76726F6C6B6A6B6C6F73777C82888D9194969694928F8B87837F"
		$"7B787676787B7E83878A8B8B8986817C78736F6D6A69696A6D7074797F858A8E"
		$"9294969593918E8A85817E7A7877787A7E8286898B8C8A88837E7975706D6A69"
		$"69696B6E72767C82878C919495969493908C8884807C797878797C8084888B8C"
		$"8B8985817B76726E6B696868696C6F73787E848A8E9294959593918E8A87827F"
		$"7B7978797B7F83878A8C8C8A87827E7973706C69686768696D70757B81878B90"
		$"9395969493908D8985817E7B79797B7E8285898B8D8B8985807B76716D6A6867"
		$"66686A6D72787E84898E9194959593918E8B8784807C7B797B7D8084888B8D8D"
		$"8A87827D78736F6B68666666686C6F757A80868B909394959493908D8A86827F"
		$"7C7B7B7C7F83878A8D8D8C8985807B75706C6967666566696D71767C83888D91"
		$"93949493918F8C8885817E7C7B7C7E8285898C8D8D8A87827C77726D6A676564"
		$"65676A6F73797F858B8F9294949493908E8A8784807E7C7C7E8185888B8E8E8C"
		$"89857F7A756F6C6866646466686C70767C82888D919394949492908D8985827F"
		$"7E7D7E8083878A8D8E8D8B87827C76716D696664636466696D73797F858A8F91"
		$"94949493918E8B8885817F7E7E7F82868A8C8E8E8D89847F79746F6A67646363"
		$"64676B70767C82878D909394949492908D8A8783817F7E7F8185888B8E8E8E8B"
		$"87827C76706C686463626364686D72787E848A8E9193949493918F8C89858281"
		$"7F7F8184878B8D8F8E8D89847F79736D696663626263666A6F757B81878B9092"
		$"94949493908E8B88848281808183868A8D8E8F8E8B87817C75706B6763616161"
		$"63676C71777E84888D9093949493918F8C8A86848281818285888B8E908F8D89"
		$"847E78726D69656261606164696E747A81868B8F9193949493918E8B88858382"
		$"818284888B8E90908E8B87817B756F6A666361606063666A70767C82888D9092"
		$"939393918F8D8A878583828284878A8D9091908D89847E78726C676461605F61"
		$"64676D73797F858A8E9193939393918E8C8A878483838486898C8F91918F8C87"
		$"817B756F696561605E6061656A6F767C82878C8F9193939392908D8B88858483"
		$"8485888B8E9091918E8A847E78726C6763605E5E6063676C72797F848A8D9092"
		$"939393918F8D8A8785848485888B8E909292908C87817B756F6964615F5E5F61"
		$"64696F757B81878B8E9193939392908E8B8987858485878A8D909293918E8A84"
		$"7E78716B66625F5E5D5F62666C72787E84888D9091939393918F8D8A88878585"
		$"87888C8F919393918D88827B746E6964605E5D5E6064696E757B81868A8E9092"
		$"939392908E8C8A88868587888B8E919393928F8A847E77706A65615E5D5D5E61"
		$"666A71777D82888C8F9192939291908D8B89878686888A8D90939393918D8781"
		$"7B746D6863605D5C5D5F63686D747A80858A8D9092939392908E8C8A88878787"
		$"898C8F91949493908A847E77706A64615D5C5C5D60646A70767C82878B8E9092"
		$"929291908E8B8A888787898B8E91939494918D88817A736D67635E5D5B5C5E62"
		$"676D73797F84898D8F91929292918F8D8B898888888A8D9093949493908B857E"
		$"76706A64605D5B5B5D60646A70767C81868A8D9091929291908E8C8A8988888A"
		$"8C9092949594928E88817A736D67625E5B5B5B5E61676C72797E84888C8E9192"
		$"9292918F8D8B8A89898A8B8F9294969694918B857E76706A64605C5B5B5C6064"
		$"696F757B81858A8D8F91919191908E8C8B8A898A8B8D9193969695938E88827B"
		$"736D67615E5B5A5B5E61666C72787E82878B8E9091919191908E8C8B8A8A8B8D"
		$"909395979795918C857E77706A64605C5B5A5B5F63696F757A7F84888C8E9091"
		$"9191908E8D8B8A8A8A8C8F9194979796938E88827B736C66615D5B5A5B5D6166"
		$"6B71777C82868A8D8F919191908F8D8C8B8A8A8B8E919496979796918D857F77"
		$"706A645F5C5A5A5B5F63696E74797F84888B8E90909190908E8D8B8A8A8B8D90"
		$"9395979796948F89827B736C66615D5A5A5A5D60666A70767C8185898C8E9090"
		$"90908F8D8C8B8A8B8C8E919497989896928D867F77706A635F5B5A5A5B5E6368"
		$"6D73797E83878A8D8F9090908F8E8D8B8A8A8B8D90939698999794908A827B73"
		$"6C66615D5A5A5A5D60656A70767B8084888B8D8F9090908E8D8C8B8A8B8D8F93"
		$"9598999996938D867F777069635F5B5A5A5B5F63686D73787E82878A8D8E9090"
		$"908F8E8D8B8B8B8C8E91949799999895908A837B736C66615D5A595A5D60656A"
		$"70757A7F84878A8D8E90908F8E8D8C8B8B8C8D909396999A9A97938E877F7870"
		$"69635F5B5A5A5B5E63686D72787D8185898B8E8F908F8F8E8D8B8B8B8D8F9296"
		$"989A9A9996918B837C746D66615D5A5A5A5D60656A6F75797F83878A8C8E8E8F"
		$"8E8E8D8C8B8B8C8E919497999A9A98948E878078706A645F5B5A5A5B5E63676C"
		$"72777C8185888B8D8E8F8F8E8E8D8B8B8C8D909396999B9B9A97918B847C746D"
		$"66615D5A5A5B5D6064696F74797E8285888B8D8E8E8E8D8D8B8B8B8D8E919498"
		$"9A9B9B99948F888078706A645F5B5A5A5B5E63676C71767B7F84878A8B8D8E8E"
		$"8D8D8C8B8B8C8E9094979A9C9C9A97928B847C756D67615D5B5A5B5D6064696E"
		$"73787D8185888A8C8D8D8D8D8C8B8B8B8D8E9295989B9C9C99958F888178716A"
		$"64605C5A5A5C5F63676C70757A7F8386888B8C8D8D8D8C8B8B8B8C8D9194979A"
		$"9C9D9B98938D857D756D67625E5B5A5B5D6165696E73787C818487898B8C8D8D"
		$"8C8B8B8B8B8D8F9396999C9D9C9A9690898179726A64605D5B5B5C5F63676C70"
		$"75797E8285888A8B8C8C8C8B8B8A8B8C8E9194989B9D9D9C98938D857E766F68"
		$"635E5C5B5C5E6165696E73777C808386888A8B8C8C8B8B8A8A8B8D8F93969A9C"
		$"9D9D9A96918A8279726B65605D5B5B5D6063676C7075797D818487898B8B8B8B"
		$"8B8A8A8B8C8E9195999B9D9E9C99948D867E766F68635F5D5B5D5E61656A6E72"
		$"777B7F8285888A8B8B8B8B8A8A8A8B8D9093979A9D9E9D9B97918A827A726C65"
		$"615D5C5C5D6063676C7074787C808385888A8A8A8A8A8A8A8A8B8E9195999C9E"
		$"9E9D99948E877F766F6963605D5C5D5E6165696D72767B7E828487888A8A8A8A"
		$"8A8A8A8B8D9093979B9D9F9E9C97918A837B736C66615E5D5D5E6063676B6F73"
		$"787C7F828587888989898989898A8B8E9196999C9F9F9D9A958E877F776F6964"
		$"605E5D5D5F6266696D7176797D818385878889898988888A8A8D9094989C9E9F"
		$"9F9D98938B847B746D67635F5D5D5E6064676B6F73777B7E8184858788888888"
		$"88888A8B8E91969A9D9F9F9E9A968F888078706A64615E5E5E606266696D7175"
		$"797C7F828486878888888888888A8D9094989C9FA0A09D99938C847C756D6763"
		$"605E5E5F6164676C6F73777B7E81838586878787878788898B8E91969A9DA0A0"
		$"9F9C9690888178716B6662605E5F6063666A6D7175787C7F8184858687878787"
		$"87888A8C9094989C9FA0A09E9A948D857D756F696461605F606264686C6F7376"
		$"7A7D7F828485868686868787888A8E91969A9DA0A09F9C9791898179726C6763"
		$"605F60616366696D7074787B7E80828485858686868687898B8F93989C9FA0A0"
		$"9F9A948E867E76706A66626160616365686C6F7276797C7E8183848585858585"
		$"86878A8D91969A9DA0A1A09D98918A827A736D6764616060616366696D707377"
		$"797C7F818284848485858586888B8E93979C9FA1A19F9B968E877F77706A6663"
		$"6161616365686B6F7275787B7E80828384848484858587898D91959A9EA0A1A0"
		$"9D99938B837B746E6965636161626466696C6F7376797B7E8082828384848484"
		$"85878A8E93979C9FA2A29F9C9690888078726C67646361626365686B6E717577"
		$"7A7C7F818282838383848586888C90959A9EA1A2A19F9A948C847C756F6A6664"
		$"6362636467696C6F7275787B7D7F8181828282828485878A8E92979CA0A2A2A0"
		$"9C9790898179736D69666463636466686B6E717476797C7E7F81818282828384"
		$"85888B90959A9FA1A2A29F9A948D867E77716C686564636465676A6C6F727578"
		$"7A7C7E7F8081818182828485898D92979CA0A2A2A09D98918A827B746F6A6765"
		$"64646567696B6E707376787B7C7E7F80818181818284878B8F949A9EA1A3A2A0"
		$"9B968E877F78726D696765646566686A6D6F727577797B7D7E7F7F8080818182"
		$"85888C91969CA0A2A2A19D99928B837C75706C686665656667696B6E70737578"
		$"7A7C7D7E7F7F7F7F808183868A8E93999DA1A2A2A09C968F888079736E6A6866"
		$"666667696A6D6F727476797B7C7E7E7E7F7F7F808184878B91969B9FA2A2A19E"
		$"99938B847D76716D696766666768696C6E70737577797B7C7D7E7E7E7E7F8082"
		$"85888D93989DA0A2A2A09C969088817B746F6C6967676768696B6D6F72747678"
		$"7A7B7C7D7E7E7E7E7F8183878B90969A9FA2A3A29F9A938C857E78726E6A6967"
		$"6767696A6C6E70727577797A7B7C7C7D7D7D7E7F8184888D93989DA0A3A3A09C"
		$"979089827B76706D6A696868696A6C6D7072747678797B7C7C7C7C7D7D7E7F82"
		$"858A90959A9FA2A3A29F9A948D867F79736F6C6A6968696A6B6D6F7072757678"
		$"797B7B7C7C7C7C7C7E7F83878C92979CA0A3A3A09D98918A837C77726F6C6A69"
		$"69696A6C6E7072737578797A7B7B7B7C7C7C7D7E8185898E949A9FA2A3A29F9A"
		$"958E87817A75716D6C6A696A6A6C6D6F7073757678797A7A7B7B7B7B7B7C7E82"
		$"868B91979CA0A2A3A19D98918B847E7873706D6C6A6A6A6C6D6F707273757778"
		$"79797A7A7A7A7B7B7D7F83888E94999EA1A3A29F9B958E88827B76726F6D6C6B"
		$"6B6C6D6E6F7073747677787979797979797A7B7D81858A90969B9FA2A2A09D98"
		$"928B857F7975716E6C6C6B6C6C6D6F707273757678787979797979797A7B7E82"
		$"878D93989DA1A2A29F9B969089827C7874706E6D6C6C6D6D6F70727375767778"
		$"787979787878797A7C7F84898F959A9FA2A2A09D99938C86807A7672706E6D6C"
		$"6D6D6E6F71727475767778787878787878797B7D81868B91979DA0A2A29F9B96"
		$"908A847E797572706E6D6D6D6E6F7172737576777878787878787878797B7E82"
		$"888E949A9FA1A2A09D99938D87817B7774716F6E6D6D6E6F7072737475767778"
		$"78787877777778797C7F858B91979CA0A2A29F9B96908A847F7A7673716F6F6F"
		$"6F70707273747576777878787877767677787A7D81878D94999EA0A1A09D9993"
		$"8E88827C79757270706F6F6F70717273757676777777777676767676787B7E84"
		$"8A90969B9FA1A19F9B96918B857F7B78747271706F7070717273747576777777"
		$"77767676757677797C81868C93989DA0A1A09D99948E88827E7A767372717070"
		$"707172737475767676767676767575757576797D82888E959A9EA0A09F9B9691"
		$"8B85817C797673727170707172737475757676767676767575747576777B7F85"
		$"8B91979C9FA0A09D99948E89847F7B7875737271717272737375757676767676"
		$"76757473737475787C81878D94999DA0A09E9B96918C87827E7A777573727272"
		$"727374757576767676767675747373737476797E848A90969B9FA09F9D99948F"
		$"8A85807C797675747373737374757576767676767675747373727374767B7F86"
		$"8C93989C9E9F9D9A96918B87827F7B7876757473737474757576767676767675"
		$"75737372727375787C81888E94999D9F9E9C99948F8A85817E7B787675757575"
		$"757576767677777776767574737271727375797E848A91969B9D9E9C9995918C"
		$"8783807C797876757575757576767677777776767574737271707273767B8087"
		$"8D93979B9D9C9A97938E8A85827F7C7978777676767676777778787877767675"
		$"73727170707274787D82899094999C9C9B9995918D8884807E7B797877767676"
		$"767777787878777676757373727070717275797F858C9196999B9B9997938E8A"
		$"86827F7C7B7978777777777878787878787876767473727070707173767C8188"
		$"8E93989A9B9A9895908C8885817E7C7B79787878787878787878787877767573"
		$"72706F6F707175787E848A9094999A9A9996928E8A8783807E7C7B7979787878"
		$"787979797978787776757371706F6F7072767B80878D9296999A9A9794918D89"
		$"85827F7D7C7B79797979797979797979787776757372706F6F6F7073787C8288"
		$"8E939799999795928E8B8784817F7D7C7B7A797979797A7A7979797977767472"
		$"71706F6F707275797F858B91959799989694908D898783817F7D7C7B7B7A7A7A"
		$"7A7A7A7A79797876757372706E6D6E6F72767B81878D919697979794918E8B88"
		$"8582807E7D7C7C7B7B7B7B7B7B7B7A7A7978767472706F6E6E6F7074797E848A"
		$"8F939797979693908D8A878481807F7E7D7C7C7C7C7C7C7B7B7A7A7876757371"
		$"6F6E6D6E6F71757A80868C919497979694918E8B888583817F7E7E7D7C7C7C7C"
		$"7C7C7C7B7B7978767472706F6D6D6E7073777C82888E9294969694938F8D8A87"
		$"858281807F7E7E7D7D7D7D7D7C7C7B7B79777573716F6D6C6D6E7074797F858A"
		$"8F9395959493918E8B88858482807F7F7E7E7E7E7E7D7D7C7C7B797876747270"
		$"6E6D6C6D6F72767C81878D9093959493918E8D898784838281807F7F7F7F7E7E"
		$"7E7E7D7C7B79787573716F6D6C6D6D7073787E84898D9194949492908D8B8886"
		$"84828180807F7F7F7F7F7E7E7E7C7C7A78767572706D6D6C6D6E71767B80868B"
		$"8F92939392908E8B8A87858482818080808080807F7F7F7E7D7C79787573716F"
		$"6D6C6C6D7073787D83888D90929393918F8D8B88868484828181808080808080"
		$"7F7F7E7C7B78767572706D6C6C6C6E71757A80858A8E90929291908D8B8A8785"
		$"8483828181818181818080807F7E7C7A787673716F6D6C6C6D7073787C82878C"
		$"8F919291908E8D8A8887858483828281818181818180807E7C7B797674726F6D"
		$"6C6B6C6D70757A7F84888D8F9191908E8D8B8987858584828282828282828181"
		$"807F7E7C7A787673716F6C6C6C6D6F72777C82878B8E909191908E8B8A888785"
		$"8484828282828282828281807F7D7C797774726F6D6C6B6C6D7074797E84888B"
		$"8E9090908E8C8A888786858483838282828282828281807F7D7B787673706E6C"
		$"6B6B6C6F72767B81858A8D8E90908E8D8B8A8887858584848383838383838282"
		$"81807E7C7A7875726F6D6C6B6B6D7073787D82878A8D8E8E8E8D8B8A88878685"
		$"8584848484848484838282817F7E7B797674716E6C6B6B6C6E72767B8084888B"
		$"8D8E8E8D8C8B89888786858484848484848484848382817F7D7B7875726F6D6C"
		$"6A6B6D6F73787D82868A8B8D8E8D8C8B8A888787858584848484848484848483"
		$"82807E7C797774716F6C6C6B6C6F71767A7F84878A8C8D8D8D8B8A8888878685"
		$"858485858585858584848382807E7B787573706D6C6B6C6D7073787C8185888A"
		$"8D8D8D8B8A8988878685858585858585858585848482817F7C797774716F6C6C"
		$"6B6C6E71767A7E82868A8B8B8B8B8B8A88888785858585858585858585858584"
		$"82807E7C797673706D6C6B6C6D7073787C8184888A8B8B8B8B8A888887868585"
		$"8585858585858585858483817F7D7A7774716E6C6B6A6C6E7075797E8285888A"
		$"8B8B8B8A898887878585858585858686868685858482817E7C797673706E6C6C"
		$"6C6D7073777C808487898A8B8B8A898887878685858585868687878786858584"
		$"82807E7B7774726F6C6C6B6C6E7075797D818587888A8A8A8988878786858585"
		$"8686878787878787868583817F7C797673706D6C6C6C6D6F72777B7F83868889"
		$"8A8A8988888787868686868787878787878787858482807E7B7875726F6D6C6C"
		$"6C6E7174797C8184878889898888888787868585868686878787878787868584"
		$"82807C797673706E6D6C6C6D7073767B7E828587888888888887878686868686"
		$"8787878788888887868583817E7B7875726F6D6C6C6C6E7074787C8083858788"
		$"88888787868686858686868787878888878787858482807D7A7774716F6D6C6C"
		$"6D7073767A7E8184858787878787868585858585868787878888888887878584"
		$"817F7C797673706E6D6C6D6F7275787B7F828485878787868585858585858586"
		$"8787888888888887868482807E7B7875726F6E6D6D6F707376797D8082848586"
		$"868685858585858585868687878888888888878584827F7C797673716F6E6D6E"
		$"707275787B7E8183848585858585848484848585858787888888888887868583"
		$"817E7B787573706F6D6F6F717477797C80828484858585858484848484858586"
		$"8787878888888787868482807D7A777572706F6F6F717376797C7F8284858686"
		$"868585858584848585858686878787878786858482807E7B78767371706F6F70"
		$"7275777A7D808284858585858584848484848485858587878787878786848382"
		$"7F7D7A78757371707070727376797C7F81828484848484848484848484848585"
		$"86878787878786858482817E7B79767472717070717375787A7D808283848484"
		$"848484838383848485858586878787878786858382807E7B7876747271717173"
		$"75777A7C7F818384858585858484848484848485858586868686868585838280"
		$"7E7C79777573727171727476787B7E8082838485858484848484848484848585"
		$"858686868686858483817F7D7B787675737272727375787A7C7F818284848484"
		$"8484848383848484848585868686868685858482817E7C7A7876747372727375"
		$"76797B7D80818283848484848383838383848485858686868686868585838280"
		$"7E7B797775747373737576787B7D7F8182848484848484848483838484848484"
		$"858585858584848282807E7C7A787675737373737577797B7E80828284848484"
		$"84848383838484848484858585858585848482817F7E7B797776757373747576"
		$"787B7D7F8182838383838383838282838383848484858585858585848382817F"
		$"7C7B797775757374757678797B7E808182838383838382828282828384848485"
		$"858585858584848281807E7C7A7876757574757577797B7D7F81828384848483"
		$"83838382828383838484848485858484838281807E7C7B797876757575757678"
		$"7A7C7E8081828384848383838282828383838484848485858584848382817F7E"
		$"7C7A7877767675767678797B7D7F818182838383838282828282828383848484"
		$"8484848484838281807F7D7B7978777676767678797B7C7E8081828282838282"
		$"82828282828283848484848484848484838281807E7C7B79787776767677787A"
		$"7C7E7F8182838383838383838282828283838383848484848484828281807F7D"
		$"7C7A79787776767778797B7D7E80818282838382828282828282828282838484"
		$"84848484838282817F7E7C7B7A787877777778797A7C7E7F8181828282828282"
		$"828282828282828383848484848484828281807F7E7C7B797878777878797A7B"
		$"7D7E808181828282828282828282828282828383848484848483828281807E7D"
		$"7C7A797878787878797B7C7E7F81818282828282818181818181818282828282"
		$"82"
	}
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// This snd is used for holding the sample sound used as the harmony voice.

resource 'snd ' (rSampleMelody, "sample melody", purgeable) {
	FormatOne {
		{sampledSynth, 0}
	},
	{hasData,soundCmd {20}		// buffer data offset is 20 bytes
	},
	{13976,						// number of samples
	 0x2B7745D1,				// sample rate
	 13974,						// loop starting point
	 13975,						// loop ending point
	 0,							// standard sound header
	 60,						// base frequency
		$"80807F7E7B7B8185878788878989837F828078777A78777A7A787C8186898987"
		$"888C8C8782807E78777574777876787D81858A8C8A8B8F8C8786827B77747071"
		$"767672767E83878B8C8B90928D8989857C76706A6E737171767D848A8C8C8F94"
		$"95928E89837E756A676A6C6D6F757C848C8E8F92969897938D857E7267656464"
		$"676A727C868D8F92969B9E9B95908A7D6F635E605F60646F7E898E9194999FA4"
		$"A198928F816D5F5957595A5F6B7C8C9292939AA5ABA79E9893826C5B52515052"
		$"59687C8E9493939BA9B1ADA49C94866D574C49474B55657C919893919CADB8B3"
		$"A9A1998A6F55464140454F5F7C939995929AADBCBBAFA49C8E74574138393F4A"
		$"5C78929C989299ACBFC2B5A8A091795A3F3131384559758F9B999598AABDC4BC"
		$"AEA296815F3E2D2B314156718C9A9A9498A9BCC6C1B3A69B8764412B242C3F52"
		$"6B899A9A9698A6B8C6C6B8A99D8B6B452A2128394F678497999898A2B6C5C8BD"
		$"ADA18F714B2B202736496281969997989FB2C4CAC0B1A393764F2E202432465C"
		$"7C949998989DAFC2CBC2B2A6967D563221232F405779949A97989EADBFCAC4B6"
		$"A799825B3521222D3D5374939C9A999DAABDC8C5B6A89C876239222029384E71"
		$"919C9C9A9CA7B9C6C6B9AA9F8D693D231E25354B6D8D9C9E9B9CA5B6C4C7BBAB"
		$"A1926E41241D243348688C9C9E9D9DA3B2C2C6BCADA4977446251B203046658A"
		$"9C9F9E9DA1AEC0C7BDB0A79A794A26191F2E4362879BA09F9C9EACBDC7BEB1A9"
		$"9E7F4F27171C2B405F859CA09F9C9CA9BBC6C0B2AAA184542A1619293F5C819B"
		$"A2A09B9BA6B8C5C1B5ACA48A5A2D1616263C597E9AA19F9B9AA4B6C5C4B8AEA6"
		$"8E60311514243A547A98A1A09D9AA1B4C4C5BBAFA791663617132135507899A2"
		$"A09B989FB2C3C6BDB2A8946B3A19121E324D7598A2A09B989EAFC1C7BFB4AA97"
		$"6F3F1C101A2F4C7395A1A09C989DADBEC7C1B6AC9A75421C0F182E4A7092A0A0"
		$"9C989CABBEC7C1B6AD9E7A481F0F162C486C8F9F9F9B979AA9BBC6C4B9AFA07D"
		$"4E240F142945698E9F9E999699A6B9C6C5BAB0A281522811132643678B9F9E99"
		$"9596A3B6C4C5BDB3A485582B1213254063899F9D989495A0B3C2C5BEB5A88A5D"
		$"2F1412243E5F869D9D9894949EAFBEC4BFB7AB9163321412233B5C839D9E9794"
		$"939AABBEC5BFB7AD95693616111F3759819D9E9792929AA9BCC5C0B8AE976C3B"
		$"18101D36587E9B9E97939399A7B8C4C2B8B09A713F1B101A33557B999E979493"
		$"97A4B5C2C3BBB29E75431E1119325276969E98939296A1B3C1C3BDB6A3794720"
		$"11192F4C72959F98939195A2B4C1C4C0B7A37D4C2311172B496E93A19A939194"
		$"A0B2C1C5C0B8A7825025111528466A8FA19B9491939EB0C1C6C1B7A786572912"
		$"152640648DA19C9592929CAEBFC5C1B9AB8C5B2D1414233E6089A09D9592929A"
		$"ACBEC6C2B9AE906031161421395A839FA098929199AABCC6C2BAB19869361712"
		$"1E3557809FA199928F97A8BBC6C3BBB29B6D3B190F1A32527D9FA39A949095A5"
		$"B9C5C3BCB49F73401B0F172E507A9DA49B938F92A2B6C3C4BDB6A379451D1016"
		$"2B4B769CA59C938E909EB3C1C3BFB9A7804B1F0E142948739BA79C918C8E9DB1"
		$"C1C5BFB9AE884F1F0D122747719AA69C938D8D9BB0C1C5C0BCAF8B52220D1127"
		$"456B97A79E938E8D98AEC1C4BFBDB38F58240C0F24426895A8A0948D8C96ABBF"
		$"C4C0BDB4935D280C0D21406793A7A0958E8B96AABDC6C2BCB396632D0D0B1D3A"
		$"6694A79F958E8C95AABDC5C2BEB5996632100919396390A8A1948C8D96A9BAC5"
		$"C3BFB99F6A32100818365E8EA9A3958D8B94A7BAC4C4C1BCA36D340F0717355B"
		$"8AA8A4968D8B93A6BAC4C2C0BCA675370D0616335988A6A4968E8B91A4B9C4C3"
		$"C0BDAB79390C0315335786A5A4978F8D92A2B7C5C4C1BDAA7D3D0F04122F5383"
		$"A7A597918E929FB4C3C5C1C0AE7E401305112C4F7FA5A699918D90A0B5C2C4C2"
		$"C0AF824214050F284B78A3AA9D928C8FA0B5C3C4BFBEB2894916040C254675A3"
		$"AB9D928C8E9CB3C5C5BFBFB48E4E19030A234573A0A99D928C8E9CB1C4C7BFBE"
		$"B792531B04091F41709EA99D918D8F9CB1C2C5C1C0B894561D05091E3F6C9BAA"
		$"9F938D8F9CB0C0C4C2C1BA995B2106071C3C6998A9A0948E8F9AAEC0C5C1C2BD"
		$"9A5D230706193A6797AA9F938D8E9BAEBFC5C2C2BC9B6027070519366395A9A0"
		$"948E8E99ADBEC5C3C4BE9F6227070517366091A8A1958E8E99ACBEC5C3C3BFA1"
		$"662907041536608FA6A297908F97AABDC6C6C4BFA26A2C080313325C8EA7A297"
		$"918F97A8BAC6C8C6C2A26A310B03112F5689A8A6988F8D98ABBBC3C6C6C1A66D"
		$"300C030F2A5488A7A7998E8D98ACBDC5C5C5BFA673350D030D285185A8A7998E"
		$"8D99ADBDC5C4C2BFA7753C12020B254E81A5A89B908E99A9B9C2C2C2C1AE7C40"
		$"140309224A7EA5A89A908E97A8B9C1C3C3C1B0814316030921467AA5A99C918F"
		$"95A4B6C1C1C2C3B4874A1803061E4578A3A89B8F8E96A4B3BFC1C2C5B78B4C19"
		$"04051C4476A0A89C929096A3B3BBBEC3C5B990501B04061C3F719FA89B929094"
		$"9FB0BBBEC2C6BC9558200505193D6D9CAA9E938E939EACB9BEC1C7C399592105"
		$"0417396B99A89E958E919EAEBBBFC2C7BF9A5E24050315396697A9A09690919B"
		$"ACBABFC2C6C19C6028080313336396AAA2958F919AA9B8C1C6C8C09E60270804"
		$"142F5C95ACA4968E8F9AABB9C1C3C7C1A0652A0903112D5A93ADA6978D8D98A9"
		$"BBC3C3C7C4A4682A0702102E5890ABA5968D8D99AABCC5C4C5C4A76C2D08020D"
		$"2B578FABA6998F8D96A7B8C2C3C5C5AB713109020B28558AAAA89A8E8B93A4B6"
		$"C1C3C6C9AE74350C0209255287A7A79B8F8993A5B7C0C1C5C6AF7C3D0E020823"
		$"4D80A6AA9C8F8A90A1B4BFC2C5C8B38144120208224B7CA3A89D8F8991A2B4C1"
		$"C3C4C6B485471402071D447BA4A99D918B90A0B2BEC3C5C3B4894B1602071D41"
		$"77A3AA9D908A909FB0BDC3C5C3B58B4F1D05061B40709DABA1928B8E9AABBBC1"
		$"C1C4BB94581F0405183D6C99AAA2948B8C96AABCC2C3C7BD965D240504173A67"
		$"97AAA1948D8C96A8BAC2C2C3BD9B6329070415346396ABA3958B8A95A7B8C0C0"
		$"C2BFA1692D080313325F94ABA4958C8A94A6B7BFC0C1C0A46D2F0902102F5D93"
		$"ABA3958C8A94A4B6C1C2C1BDA571330B030E2C5C90AAA3958D8B92A2B3BFC1C2"
		$"BFA874370D020D2A588CA7A3968E8C91A0B3BFC0C2C3AC763A0F020B295687A6"
		$"A4988E8A91A0B2BFC2C3C2AC793E12020B265284A5A4988F8B909FB2BEC2C3C2"
		$"AE7E4415020B264D7DA3A5988F8C909EB1BEC1C2C2B2834618040A244879A1A5"
		$"998F8A8E9DB1BDC1C0C1B2884E1C060C224375A1A89A8D888C9AB0BEC0BFC0B6"
		$"8F531E070B2143729DA79B8E888A98ACBDC2BEC0B8945822080A1F406E9BA79C"
		$"90898A96AABCC0BDBFBA99612809071A3C6C9BA89D908A8A95A9BABFBDBFBC9D"
		$"652D090518396799AB9E908A8B96A6B8BEBEBFBEA1692F0A0416366396AAA193"
		$"898995A8B8BDBCC0C1A56C2E0A0414346293A8A1948B8A94A6B7BFBEC0C0A770"
		$"330C0313315D8FA9A3958C8A92A3B5BFC1C1C0AA74370F03112E5A8CA8A5968B"
		$"8991A2B4C1C2C1BEAB7C3E1204112D5586A8A7988D898E9DB3C0C1C0C2B08145"
		$"15030F2C5383A5A6998D888D9BB0C1C4C0BFB1874A17040F2A507FA3A79A9089"
		$"8C9BAFBFC2C0BCB08D511D060E274B7FA7A7988F8B8D99ADBDC1BEBEB38D5622"
		$"090D25497BA5A8998E8C909AAAB9BFBDBEB99458240B0C22497AA3A89B918B8D"
		$"99A9B8BDBDC0BB995F27090C234676A1A99B908C8D96A7B6BBBCC1BF9E632807"
		$"082045749DA89C908B8D96A5B5BCBEC2C1A3652A0A081D41729EA89B8F8B8D96"
		$"A5B3BBBFC3C1A3682C0B091E3F6D99A79D918D8E95A5B5BBBEC3C2A46A310D08"
		$"1D3F6993A6A0948D8C94A4B4BEBFC1C1A66F33100A1B3B658FA29E968E8B92A3"
		$"B4BEBFC0BEA7763C12091A385E89A1A096908C8F9EB3C0C2C0BDA97A42160816"
		$"355A85A1A2978F8C8F9DB1BFC3C0BCAB80461A0A15325580A1A399908C8F9BAE"
		$"BDC1C0C0B0864C1B0B152C4F7CA1A4988F8C8F9BAEBEC0BEBEB28A51200C1329"
		$"4C779DA69A908B8D9AACBBC0BDBDB69257240D122748739AA49B928B8B96AABC"
		$"C0BDBEB9955D280C0D24467097A49D938C8A94A8BBC1C0C0BA9A622E0E0D2243"
		$"6D93A49E948C8A93A6B8C1C1C1BC9D6A330F0B1E3E6994A59F948D8A90A3B6C0"
		$"C2C1BDA26E36100B1D3C6592A69F938B898FA0B4C0C1C0BDA77439130C1D3860"
		$"90A79F928A888FA0B3BEBFC0BFAC7A3C130B1B375D8DA7A0928B898E9DB1BDBF"
		$"C1C3B17D3F140B1B375D88A4A0938A888E9CAFBDBFBFC3B48444160A17335886"
		$"A3A2958D888B9AAEBCBEC0C5B78848170915315580A3A5978D878A98AEBEBFBF"
		$"C5B98D4E1A07122D527EA3A79A8E868897ADBEC1BFC2B993541B07122C4E7BA1"
		$"A69B90878694ABBDC1BFC3BC985A1F070F284D779EA79B8F878792AABEC3C0C2"
		$"BC9C5F22080D2647749FA89C8F888893A7BBC2C3C2BD9B60270A0D2446729DA8"
		$"9B8E878894A9BBC3C2C2BE9D63290B0B2043709CA99B8D888A96A9BBC2C2C3BE"
		$"A064290B0A1E416E99A89D8E878997ACBCC1C2C4BFA1682B0A091D3F6995A99F"
		$"8F888A95AABEC4C3C3BFA4692C0A071B3D6594A9A092888896A9BCC6C6C4BFA5"
		$"6D2F0B071B395F91A9A0928A8A95A8BBC6C6C4C0A873350C0516366090A8A092"
		$"898994A8BBC4C8C6C0AA78370D0414345D8CA59F908A8C96A7B8C4C5C3C3AD7B"
		$"3B100512315A89A49F918B8C95A5B7C2C6C5C3B27E3F1305102E5784A29F938C"
		$"8D95A5B5C1C5C4C3B3824215060F295283A29E938F8F94A3B5C1C2C4C7B48344"
		$"17050E294D80A2A1948E8F97A3B2BEC4C5C6B8864516070D234A7CA0A1968F8E"
		$"94A3B3BDC0C4C8BC8B4616060C2247769EA49A918E92A0B2BEC0C2C7C0914B18"
		$"03092144749FA59B928D8F9EB3C0C1C1C6C1934F1A03051C4175A1A69A918D91"
		$"9DB0BEC2C4C7BF9654190205194175A2A6988F8E929EB1BEC1C2C8C094511B05"
		$"03173F72A0A79A908D93A1B0BBBEC0C6C198541A0404163C70A2AA9C918D919F"
		$"B1BBBCBEC6C49C55190304163870A3AA9C928F929EAFBABDBEC7C59B561A0304"
		$"133971A3AA9C918D929FAFB9BBBDC6C69E56180304113875A4A699908F95A1B0"
		$"B8B7BCC8C79C54180304113A72A1A89B908E96A3AFB4B6BDC9C99E5418030312"
		$"3A71A0A69B918D93A2B1B6B6BDCBC99B541A0304133A6F9DA69B918E94A3B0B7"
		$"B8BCC9C69B571A0303123A6D9CA59A908E95A4B1B7B9BFC9C89C551A04030F38"
		$"6D9AA3988F8D95A7B3B5B8C1CCC69A571C040311366797A49A8F8E97A6B3B7B8"
		$"BDCAC89D571D050310326294A49C918E96A8B5B7B7BDC9C89D5920050311305C"
		$"91A59D949095A5B5BAB6BAC8C8A05C230604102D5A8FA39E959094A5B4B9B6BA"
		$"C6C5A3632507030E2A5A90A49E969095A5B3B6B5B9C4C5A6662908030E2B578F"
		$"A69F969196A3AFB4B4B8C2C6AA6A2C09030C28578EA6A0969196A3ADB2B3B5C1"
		$"C8AD6B2C0A030A2A5789A4A0979396A2ADB2B2B4C0C8AF6F2F09020B28548BA4"
		$"A19995969FABB1B2B6C1C7AD70310B030B285286A5A49A95969FAAAFB1B5C1C6"
		$"AD73340D030B264E86A8A69B95959EA9AFB0B3BEC7AE74370F020B275083A5A6"
		$"9C95959EA7ACAFB6BFC4AF773910020B264D7FA4A79D96959EA8AEB1B5BDC1AD"
		$"7A3D11030B234A7FA5A89F97959CA6ABB0B4BDC3AE7C3F13040D23487CA4A89E"
		$"96969CA5ACB1B3BAC1B37F3F15060B20477DA3A79E97949BA7ACAFB2B9C2B482"
		$"4418070B204678A0A69F98959BA6ACACB0BBC4B686481A06092144739EA8A097"
		$"969CA4AAAEB1BAC3B88A4E1E06081F42719CA9A2989399A4AAAEB3BCC2B89054"
		$"2007091E3F6D9BA9A2999497A1AAAFB2BAC2BA93592509081D3C6998A9A39995"
		$"97A0A7AEB3B9C0BC975C290C091B396597A9A49A95979FA7AFB3B9C1BD9A5F2A"
		$"0E0A19376393AAA69C95959CA7B0B4B9BFBD9D622D100A19366191A8A69D9694"
		$"9BA6AFB4B8BEBBA06933110A17345E8DA6A7A0989398A2ADB4B8BDBDA36F3914"
		$"0915325A88A5A9A19893969FA9B3B9BFBFA7743D1809132F5683A4AAA2979195"
		$"9FA9B3BABFBDA87A441B0C142B5080A4ABA39892929DAAB3B8BDBCAA814C1E0A"
		$"122A4B7CA4ABA198919199A8B4B9BBBCAE854F230E11274978A0ABA398909199"
		$"A6B3B9BBBAB08C55260F11244574A0AAA399918F97A6B3B9BCBBB18F5C2A1111"
		$"24446F9BABA499929095A2B2B9B9BAB497622D120F21416C99ACA79B908D93A1"
		$"B1BABBBBB79D6931110D1F3F6996AAA59B928C919FB1BABAB9B8A26F37130D1C"
		$"3C6792A8A79D938D8F9DB0BBBABAB9A4753D170C1938628EA8A79D938D8E9BAE"
		$"BABBBAB8A67A421A0D18355C8CA9A99D928C8D99ACB9BCBAB8A87E471D0D1834"
		$"5A88A8A99C928B8B97AAB9BCB9B6AA874F200E17315785A6A79B928B8A95A8B8"
		$"BCB8B6AF8D52210F17305581A3A69C938D8A91A6BABCB6B6B190572510152C52"
		$"7EA2A79C928D8A91A5B8BCB6B5B2935C2B13152A4D7AA1A89D938D8A90A3B6BA"
		$"B6B7B397642F1413294C769CA79E948D8A8FA1B3B9B6B6B59E69331413284972"
		$"9BA89F958D888E9FB2BAB7B7B4A17137161225466E98A89F948C878C9DB2BAB7"
		$"B6B6A4763E171022446B94A89F938D8A8C99AFBCB9B5B5A87C421B11203F6692"
		$"A8A0938D888C99ADB9B9B6B8AC81461B101D3C628FA9A1928C898B97ACB9B9B7"
		$"B9AE874A1D101D3A5E8AA7A1938D8A8C97AAB7B7B5BAB38A4D1F1019355C86A5"
		$"A4968C888B97A9B6B6B4BAB79053220F17325685A6A3938C898A96A9B6B7B6BC"
		$"B79457240F152F527FA5A6958A888B95A8B8B7B7BEBB975A250E132B4F7CA3A7"
		$"978B868895AAB9BCBABDBC9B5B230D13294B7AA3A6978C868796ABBABBBABEBD"
		$"9D5E240B11274A77A1A8988B868895A9BABEBDBFBE9E60270C0E244775A0A799"
		$"8C878996A9B9BFBFC2BF9F61270B0C2243719EA89A8D878897AABBC1C1C4C0A1"
		$"62270B0B20416C99A89B8D878996ABBCC2C0C1C2A466290C0A1C3C6796A89F8F"
		$"878896ABBDC2C0C3C3A4692C0A061A3B6496AA9E8F898A96AABDC3C0C1C1A76B"
		$"2D0B0616366496A99E8F898A97ABBDC3C1C2C3A76B2E0B0514326297A99D8E88"
		$"8B99AEBEC1BFC2C4A86B2D0A0612326196A79B8D898D9BAFBDBFBEC3C7AA682B"
		$"0A0310306095A79C8F898B9BB1BFBFBEC7C9A7652908020F2F6196A59A8E8B8D"
		$"9CB3C0C0BEC8CCA5632708020C306093A59C908B8D9EB5BFBEC0CBCCA7622507"
		$"030C2C5E94A59C908A8C9EB5C0BEC0CECFA7622306030A2A5F94A39C90898D9F"
		$"B6C1C0C4D0CFA86122050309295C91A29A908A8EA1B7C0C0C3CFD1AB601F0404"
		$"08265D93A09990898DA3BAC2BFC2CFD1AB601F040409245993A2998D8990A4B9"
		$"C2BCBECED3AE601F050407235E97A2978D8990A4BAC2BABCCFD5AC5E1D030405"
		$"2465979E958E8C93A7BCBFB8BECFD1A75A1C030406256498A0948C8C96ACBEBD"
		$"B7BBCED6A754170404062668989D948F8D96ADBFBDB4BBD0D8A6501303040729"
		$"69989D95908E97ACC0C0B6BBD5DB9F4810030407286B999D948F8F99AFC4C3B6"
		$"BDD8DB9C410D0404072A6B999E9591909AB2C6C0B3BFDBDA9B400A0304082A6D"
		$"9C9C94929099B4C9C2B4C0DBD8953D090404082C709A9C9592909BB6C8C1B6C2"
		$"D9D7933A080404092F7299999392929FB9CABFB4C0D8D99438070404092E739A"
		$"989292949FBACBBEB2C2DBD693360704030B36769594919295A4BCCABFB4C0D8"
		$"D5913606040408367B98948E9197A5BEC9BEB4C1DCD3873307040409387B9895"
		$"8E9097A9C2CBBBB0C1DFD4842D0505030B367A99968E9198A8C1CDBAAEC1E2D8"
		$"82280505030A3D7E9494929397A8C4CDBAACC2E3D582280305030D4082979391"
		$"9498A9C5CEB9AEC5E1CF7E260206030D4383989491969AABC5CCB9ADC6E3CD76"
		$"22030503104384999593979DB0C8C9B6AEC5E5CE721D0206031047879795969A"
		$"9DAECBCDB3ADC9E7C96B19020603134A839697969AA0B2CCCFB6ACC6E3C86A18"
		$"020603154E869495999DA0B2CBCCB7B1CBE4BE64160206031352899292989DA4"
		$"B7CECBB5B4CDE1BA5C1502060316518693959A9FA4B7CFC9B3B2CFE6B9571002"
		$"060418538690959D9EA1B8D2CAB0B2D1E1B357110205041B548793949AA0A7BB"
		$"CFC8B1B1CDE3B4510F0206031C5B8B90929BA1A7BDD2C6AFB3D0DEAF53100205"
		$"041C5A8C93949AA2ABBFD2C8AEAECFE2AB4D0F030503235E8790969DA2ACC4D4"
		$"C7AEAFCDDCA54C11030505225F8993979FA4ACC3D5C4AFB2CFD79D4911030505"
		$"20608B9399A2A7ADC4D7C7B0B2D2D897420D03050524628B949CA2A5AFC8D6C4"
		$"AEB4D3D496400B03050526658B929AA5A7AFCAD8C3AEB8D6D1903F0B03050729"
		$"648A959DA4A6B1CBD6C2AEB9D4D08D3C0B03040829658B969EA6A8B2CCD6C1AF"
		$"BAD5CF8C390904040729688E969FA7A7B2CED5C0AFBAD5CD8836090404082E6C"
		$"8D97A1A6A6B4CDD5C2B0B9D4CC843307040409326D8B969FA7A9B6CED6BFAEB9"
		$"D4C984340804030B326A8B98A1A7A9B8CDD3BDAEBCD2C483350904030C33698F"
		$"9AA1A8ABB7CCCFBBADB8D2C784370A04040B36729198A0A8A9B7CCCDB9ABB9D2"
		$"C381390A03040E377090989FA6ABB9CBCCB8A9B7D2C882350904030F396E8F99"
		$"A2A9AAB7CBCAB6AABAD3C180380B04041237698E9DA5AAABB8C9C7B5ABBAD1C0"
		$"82390B04041137698D9FA9ADABB5C6C6B6ACB9CDBF813C0E03040F33678EA2AC"
		$"AEA9B4C6C4B7AFB9CBBD823E0F03040D316692A6AEADAAB3C2C1B5ADBACBBB82"
		$"3F1004040E336692A8B0AFACB2BEBDB3AFBBCBBD85401003040F2E6293A9B0B1"
		$"ADB1BCBDB3ADB7C8C28B431003040E2B6093ACB3B1ACB0B9BAB2ACB3C6C39049"
		$"1403040C2C6194AAB0B1ACAEB6B6B1AAB2C4C295501904030A2D6294AAB1B0AB"
		$"AEB4B2AEABB1C2C198571E0503092A6196ACB0AEABADB1B0ABA6B1C4C59C5A21"
		$"06030B2E5E92ACB1AEA9AAB0AEA8A7B0C1C5A05E2206030D2C5D91ABB2B0A9A6"
		$"A9AAAAA9B0C0C3A3652807030E2C588FAFB4AEA6A4A7A6A8AAB0BFC4A7692D0A"
		$"020B2A578EAFB4ADA6A3A5A5A8AAB0BEC3AA6F320D030C2A5389AFB5ACA4A1A3"
		$"A3A7A8AEBDC4AD763A11030D284F85AEB6ADA49E9DA1A7ABAEB8C1B17E411304"
		$"0F284980AEB8AFA49B999CA5ACB0B7BFB4864918050C23467FADB8B0A3999597"
		$"A4ACAFB7BEB58D4F1D060B22437AADBBB0A2979394A0ABB0B2BABB96561E080C"
		$"1E417AAEBAAFA3978F919DA9ADB3BDBB9B5F250A0B1F4074ACBDB0A1968E8D98"
		$"A7ACAFBCC1A2652C0D0A1B4074A7BBB1A2948A8C97A5ABAEBAC0A76F320F091B"
		$"3E6FA5BBB2A1948A8993A2A9ADB8BFAC783A130A1B3D6BA0BCB39F9187858FA0"
		$"AAADB7BFAF8042150B1C3E6799BBB4A09186848D9EACAFB4BCB38B49180E1C39"
		$"6498BAB4A19084828B9CACAFB3BBB693501A0D1C375F95BAB49F8E8582899BAC"
		$"B0B2B8B69858210F18335D8FB6B5A08E85828797ABB0B0B8B99D5F2610172F58"
		$"8AB6B6A08E84818696ABB2B0B8B9A0662B12152C5686B0B6A18E84808796AAB3"
		$"B1B6B9A46C3013152A5080ACB5A28F84818695AAB4B4B7BAA873341312284E7B"
		$"A5B2A38F84808594A9B7B6B7BAAA77371413264873A1B4A49084808494AAB9B9"
		$"B8BAAC7B3C161225456D9AB1A692847E8493A9B9BAB9BAAD82441912223F6794"
		$"AEA592867F8190A8B9BDB9B7B08B4B1C101D3A638EAAA493877F808FA7BABEB9"
		$"BAB48F5020101A375E88A5A4948881818DA4BBBFB9BAB79557221019355880A1"
		$"A4958A83818DA4BABFBBBAB6995E28101630527B9EA3968C86818AA1B9BFBCBB"
		$"BA9E652B1113284D779BA4978C8481899EB7C0BFBEBCA26930131125497094A3"
		$"9A8E8580899DB6C1BFBFBDA67235120D23426992A49B908681879CB5C3C1BFBC"
		$"A8793D130B1E3D628EA69E8F8680859AB4C3C4BEBCAD7D3E150C1B375E8CA59D"
		$"9087808699B3C3C3BEBCB18644160A16335A88A49F9187818598B2C3C3BEBEB5"
		$"8847180913305784A29F9189838596B1C4C3BFC1B88C4A1908102B5082A4A093"
		$"89838496B1C4C5C1C3BB8F4D1A070F274C7EA5A2938A848596B1C3C5C2C4BE93"
		$"4F1B060B24497BA3A5978B848595AFC3C5C2C5C19A531A0409214479A6A7968C"
		$"858492AEC4C6C2C7C49B561A03061E4179A8A9978C858594AEC4C5C1CAC69C56"
		$"1A0404194078A8AA978A858698B2C4C5C3CBC79C55190304153C78A7A9998D85"
		$"869AB5C4C2C3CEC99B55190303153E75A5AC9B8D84879DB6C4C6C3CDCA9C5318"
		$"0303113B73A5AD9F8D83879EB7C6C6C5CFCD9D53160304123971A4AFA18D8387"
		$"9CB8C8C8C7D1D0A0521403040F3671A4ADA08E83869DBBC9C9C8D1CFA1561403"
		$"040C3474A5AB9D8F878A9CB9C9C8CAD4CFA1531503030B3574A5A89A8E888DA1"
		$"BAC9C8CAD7D09A4E150303093274A4A79A8F8990A8C0C7C3C9D9D29A4A130303"
		$"0A3275A6A7998E888FA9C4C9C0C8DDD695461103030C3675A1A69A8E8791ACC5"
		$"C9C3CCDCD495440F0304093277A4A69A8F8891AFC8C9C4CEDFD191420E030409"
		$"3076A3A89B8F8895B2C9CAC3CEE0D48F3C0C0404082E76A3A79E908796B6CCCA"
		$"C1CDDFD58F3B090304073179A0A39B918995B7CFCCC1CBE0CD883B0A03040933"
		$"78A0A199908D9AB8CECAC2CBDED086370B040408337BA0A0988F8A9CBDCFC7C1"
		$"CDE0CF8131080404093379A0A1978F8C9EBED0C7BECAE1CF813108040409327A"
		$"A0A2988F8B9EC0D1C6BFCDE2CD7E2E0605030A347AA0A199908DA3C3D1C5BECC"
		$"E2CC7C2C05050405337D9EA199928FA3C5D4C7BCCBE3CB7A2804050408357A9C"
		$"9F9A9591A3C5D3C8BFC9E1CA782704050407387F9D9C999791A3C6D4C6BECFE2"
		$"C16F25050504073B829C9D999494A8C7D1C2BECFE3C26A21050504083B839F9E"
		$"979395ABCBD1BCBAD0E8C2691E0206030841879F9F999393ACCDD2BBB8D2E7BD"
		$"631A0206030B46889E9D989498B3D0D1BBB7CFE4B95D170306020B4A899D9F9A"
		$"959AB6D6D1B7B6D0E4B458140206030C49899D9F9D9799B5D5D2B9B6CEE3B255"
		$"120306030E4F8B9A9D9D989BB6D4D2BDB9D2E2AA4D0F03060210558B979A9D9A"
		$"A0BBD6D2BAB9D6E0A54A0C030602155C8F99999B9BA1BBD7D0B7B8D4E0A4450B"
		$"0306021A619096989C9CA1BBD4C9B4B9D6DFA3460A0306011C679396959A9CA2"
		$"BCD2C7B0B6DBE3A0420A0306021F6B929393989AA6BFD0C1ACB7DCE6A03F0903"
		$"0602216A939593979BA8C3D3BFA8B4DDE7A03D060406012471939494989BA8C3"
		$"D3BFA6B2DBE39C3D070305032A73959594989CABC4D0BDA7B3DBDF943C0A0305"
		$"042C73939595989CABC6D0BBA6B7DBDC953A0A03050528709397989A9CAAC6D1"
		$"BAA5B5E0DE933A080405062A6F9498999B9AA8C6D0B9A7B5DBDD973B08040505"
		$"297196999B9E9AA7C3CEB8A6B7DEDC983E080305062B72959B9B9B9BA9C1CBB7"
		$"A7B6D9DA9C440B0305062A6D989E9C9D9EA9BEC3B1A5B6DAE0A1450C03050629"
		$"6F989E9C9B9CA9BEC1AFA2B3DDE4A2490D030505256B99A19D9C9CA9BDC1B1A1"
		$"AFD8E3AA5010030505246798A2A29F9CA7B9BCAFA2B0D8E5B257100205052363"
		$"97A6A5A19EA5B2B8AFA3B1D4E2B55E15020505226299A9A8A3A0A3ADB1AAA3B0"
		$"D2E2B8641B02050421669BA8A6A29EA2ACABA5A3B0D0DEB96B1E020504215F9A"
		$"ACA8A29EA3ABA6A29DADCFDFBF7122020406215D99ADABA19DA1A7A4A09EACCC"
		$"E0C27123030407225C98AFABA09C9EA3A3A1A1ACC8DABF7628050407245C96AD"
		$"ABA39D9D9E9E9FA0ADC9D8BF782C060307215595B0ABA09C9D9F9C9EA3AFC8DA"
		$"C2792C070308215190ADAEA49C9B9C9B9FA3AEC8D8C37E3007030822508AAAAF"
		$"A89F999798A0A8B1C6D6C58434070307214C84AAB1A9A0999696A0A9B1C6D6C8"
		$"89390803051E4881A9B0A89F9894939FAAB5C5D4C78D3F0B02051A447DA9B0A8"
		$"9E9794939DABB5C5D4C691470F0205183E79A8B1A89E958F8F9DADB7C5D1C897"
		$"4B120205163A74A7B3AA9E948D909DABB7C2CECA9C5015030412366EA6B5AB9E"
		$"928A8D9CAEB7C0CCCBA35918020410326AA3B5AB9D918A8B9AADB7BECBCBA863"
		$"1F03040D2D64A1B6AD9E91888897ACB7BDC9CDAF6A2404040C2A609EB6AD9E91"
		$"878796ACB8BDC7CEB3702A0603092A5C98B5AD9E90858794AAB7BCC5CBB67B33"
		$"080209275793B4B09F91858390A8B6BCC6CCBA833B0A020924538DB2B0A09083"
		$"7F8DA6B7BCC4C9BB8D46100208214E88B0B1A291827D89A0B4BDC4C9C0955016"
		$"02071F4880ADB3A391827C859BB3BFC3C7C19D5A1D03061C4278ACB6A491847C"
		$"8197B2BFC1C4C2A5642305051A3E73A9B6A591837C8195B0BDC0C2C3AB6E2A06"
		$"05173A6CA4B6A692847C8094AEBCBFC0C1B27A320A041337699CB3A895867D7F"
		$"8FAABDBFBDC0B6863D0C0311336396B3A997887E7C89A7BDC0BEC0B78E491202"
		$"0E305C90B0AC99887E7C85A2BDC3BFBFBB955117030D2B588CAFAC98887F7C84"
		$"9EB9C3BFBFBB9D5C1F060E295483ABAE9A8A807C8198B6C1BEBFBFA56728080C"
		$"27507CA6B09D8A807B7F94B2C0BCBCC0AE75310C0D264B75A1B09E8C81797D91"
		$"AFBFBBBAC1B5813C100C2245709BAFA08D81797B8EADBFB9B7BFB98C48180D1F"
		$"426A94AEA38E81797B8BA7BCBAB4BDBE96521F0F1D3D648CAAA69483787989A5"
		$"BBB9B3BCBFA060250F1C395E85A8AB9885787684A1BABAB1B8C0A96C2E111936"
		$"597FA5AC9A877873829EB7BAB2B5BEB07A37131831557BA1AB9C8A79727E97B3"
		$"BCB4B3BDB6854016152E537598ABA08B7A737B92B0BFB5B0BBBA8F4C1A142A4D"
		$"7294A99F8D7D73788DABBEB8B1B8BB9857211427466A8EA7A18F7F757889A6BB"
		$"BAB3BABC9E61281523426586A5A3917F757789A4B9B9B5BABEA4692F17223D5D"
		$"7F9DA49683757587A1B7BCB7BABEAA7134182039587798A498877774839FB7BE"
		$"B7B9BEAD7C3F1B1C33527093A59B897873809AB6C1BDB9BDB385441C1C2E496A"
		$"8FA39B8C7B737F98B3C0C0BCBCB58C491E1B2D436289A19C8E7D757C93B1C3C0"
		$"B9BFBC934F201928405D7F9D9E917F757A8EACC3C4BDBFC19C572317253C5579"
		$"999E928376778AABC2C5BFC0C3A56026141F345175959C9284797786A6C3C8C1"
		$"C2C5AA692C13192F4B6E929B92867B7884A2C0C9C1C2C8B2733215172944698E"
		$"9B948A7D77819DBCC9C4C4CAB879371414254164889A948A7E777F99BBCAC5C4"
		$"CCBC833F18101F3C5C819A988C7E757C96B8CAC7C5CCC18E47180D1B34537D9C"
		$"9A8D80757992B5CACBC7CCC397501B0D172D4D76999C918377788EB1C9CCC7CA"
		$"C9A0581E0C142744729B9F918578768AACC8CCC7CACCA95F220A0F23416D999F"
		$"9285787688A9C6CCC7CCD0AF67260B0D1C396897A194877B7785A7C6CCC9CBD0"
		$"B56F29070717356597A396897C7784A4C4CCC9CCD1B8742E090514325E92A599"
		$"897D77829FC1CECBCED4BB79300A030E2A5A94A799887C7884A0C1CDCBD0D5BF"
		$"7D3108030D275592AB9B877C78829FBFCCCACFD6C4843508020A244F8EADA089"
		$"7A75819CBECDCBCFD7CA8A34060307214E8EACA08C7D768099BCCECDD0D8CB8C"
		$"380703071F4B87AAA08C80787F97B9CECED1DBCE903D0903051C4784A9A38F80"
		$"798097B8CCCDD3DAD1983F090305194182AAA48E817A7F96B6CBCCD0DDD49B46"
		$"0A0304163C7EABA48E817B8095B5CCCED1DED79D480B0303103A7DA9A48F837C"
		$"8197B5C9CDD3DFD99F480B03040E3879A7A591857E8297B4C7CAD1E0DDA1490C"
		$"03040C3275A6A693857D8298B4C6CBD3E0DCA64D0D0304092E76A7A797887E80"
		$"95B2C5CBD2E2DDA64F0E03040A2C73A8A997877E8194AFC3CCD4E3E1AB520F02"
		$"04082971A7A897887D8497ADBFC6D2E6E7B0530C0204092B6EA7AA9586808596"
		$"AEBFC5D1E5E9B4560D0204092C6BA4A99686808697ADBFC3CEE4EBB75B110205"
		$"072766A1AC9B8A808395ACBEC3CCE3EDBD60140205041F66A5AE9B8A808093AE"
		$"BFC3CCE4EDBE63160205041F61A1AD9D8B808294ADBFC3CCE2EFC56618020603"
		$"195D9FADA090828093ACBEC2CCE3F0C86B1901060318599CADA292848291AABD"
		$"C2CAE3F0CA731C01060414519DB0A49285828EA5BAC2C9E1F1CD782202050314"
		$"5099B0A592838290A5B7BDC7DEF3D27D260106030E4F98B1A491848291A6B6BC"
		$"C6DDF1D581280106020E4F99B1A592858291A4B4BAC2DCEED2822E0405031048"
		$"92B1AA95858490A1B5B9BFDAEED6872E0405030B4493B1AC9988828FA2B1BAC1"
		$"D6ECD68D320405030A4795B1A9998A838E9EADB6BFD6EDD78D370604030A4494"
		$"B4AA9688858F9FACB3BDD5EBDA933B0704030B3F8EB4AC968888909DA9AEB8D2"
		$"EDDD963F0904030B3F8CB5AD978A87909AA5ACB6D0EBDF9D450A0304073A90B6"
		$"AC978B88919AA3AAB4CEEBE19F460A0304053891BAAF968988929CA3A6AFC9EC"
		$"E5A44A0A030505358EBCB2988685929CA2A6AEC6EBE8A74B0B030505358DBAB2"
		$"998A868F9BA2A5ACC5EAE7AA510E0205053187BAB49A89868E98A0A5ABC5EBE9"
		$"AE550E0205063082B5B29E8C858D98A0A5AEC6E7EBB1560F0205052C7EB4B4A0"
		$"8E868D969FA8B1C8E5E9B157100205052A77B0B7A48E878B939EA8B1C8E5E8B4"
		$"5D130205052872AEB6A3908888909CA9B4C9E5E7B76317010505256BACB6A490"
		$"888A8E98A7B4C8E1E5BB6C1D0105032068ACB6A49087898F96A3B2C8DFE7C071"
		$"210205031E61A8BAA790878A9095A1ADC0DCE7C579260205031C5EA7BAA99387"
		$"878E94A0ABBAD4E2C88232060403195CA2BAAC9587868B919CA6B8D3E2CB8839"
		$"0703041A55A1BCAC9386878B8F9DA8B7CFDFCE8E3D0A030316529AB9AE978685"
		$"8B8F9CAAB6CEE0D4953D080403154C94B8B09A8A84878E9EABB7CDE0D2994309"
		$"0304134590BAB09B8E87858999ABBAD0E0D49E4B0C0204104089B8B49D8C8687"
		$"8995A9BCCEDFD7A7540D02040E3B84B7B59F8E84828594A9BACBDFDCAF5B1202"
		$"040D3979B2B8A28E84818493ADBDC7DADDB4641802040B3373AEB9A691837E81"
		$"92ADBCC5D6D8BC701D0204092F6FABB8A894847C7E91ADBEC6D3D8BD77260204"
		$"072866A6BBA993847D7E8EADC0C7D0D7C17E2D0504052360A1B9AC97857B7D8D"
		$"AAC0CAD0D6C786320504041E5A9BB7AD99897D7B8AA9C1CBD1D5C58C3A080304"
		$"1D5391B5B19C8A7E7986A6C1CAD1D5C993440A03041A4C8AB3B19E8D7F7883A1"
		$"BECCD1D6CC9B4B0C0204194783AFB19F8E7F77829FBECCD0D6CD9F5613020417"
		$"457BA9B1A19080777F9CBECED1D3CEA65C180203134078A8AFA19081787F9ABB"
		$"CFD2D1CBAA661E0204123A6FA3B2A39283767B96B9CDD0D1CDAE6F2805030F34"
		$"67A0B4A79382777A91B7CDCECECBB57A3208020D316298B3AA9683777A8DAFC7"
		$"CBCACBC0893B09020B2A5B96B3A8948578778BAEC6CAC7CCC492440C01092856"
		$"91B3A994867A7989ABC5C7C5CAC5994D11020723538BAFAA97877B7887A8C2C6"
		$"C3C8C89F55170206214E82AAAB9A887A7785A3C0C7C3C8CAA75F1D03061D477C"
		$"A9AC9B8A7D7580A0BEC7C4C8CCAD68230304194477A4AD9C8B7D747C9ABBC8C5"
		$"C7CCB573280503164071A0AD9D8D7E747995BACAC6C6CDBC7B2E0602143B6B99"
		$"ADA08F80737792B5CAC7C5CABF8839080213376493ACA1918375748CB3CAC8C4"
		$"C9C18F4610020F325D8DA9A29284767488ADC8CAC4C7C4964E17030E2B5687A7"
		$"A39285787486A9C5CAC4C8C6A0571C040C285182A5A293877A7483A5C4CBC7C9"
		$"C8A65F1F050B264C799EA396897D7480A1C2CBC5C8CAAD6926060922477098A4"
		$"9B8F7F737C9BBDCBC7C8CDB3732E08081E3F6896A79D907F747B98B9CAC9C8CC"
		$"B77D370C071B396290A99F8F81767A94B6C9C8C7CCBE8640110616355B89A6A1"
		$"9181747991B4CACBC7CBC291471307142F5584A5A2938476778DB1CBCAC6CBC4"
		$"974E1806112C517DA1A4958477778AADCBCEC6C9C89E551C070F264A77A0A598"
		$"87777588ABC9CDC8CBCAA25D22080D2243709EA89A897A7385A8C7CDC7CCCCA9"
		$"6426080A1F3D6B9FAA9A8A7B7383A5C5CDCACBCEAF6A2909091B3B6797A99E8D"
		$"7C7384A4C3CECBCAD0B8722C090617356092A89F917F7481A2C3CFCACCD1B978"
		$"330A041430588EA9A2918177809CC1D2CDCCD2BD7E380C03112D558CAAA29181"
		$"767F9CBFD1CECDD2BE853E0E020E2A5086A9A29284777D9ABDD1D0CED3C48B40"
		$"0F020C264B81A7A29385787D97BAD0D1CFD3C7914410030C21467FA7A3958779"
		$"7B94B8D0D1CFD6CA91481303091E447AA3A496877A7D94B6D0D2D1D7CC974B15"
		$"03061A3F75A3A6998A7C7D93B5CFD2D2D8CE9A50170305173A6EA2AA9C8B7D7C"
		$"91B2CED4D4DAD1A054180304153568A0AE9F8C7D7A90B2CED5D5DBD3A35A1A03"
		$"04113166A1AFA08D7D7A8EB0CDD3D4DBD5A85E1D03040E2B62A0B1A18C7D7C91"
		$"B1CBD2D3DCD6AD631F04030A275EA0B2A38E7E7D93B2CAD2D2D8D8B166200304"
		$"08235DA0B2A48F807D92B2CBD3D2D8D8B46A210304051F5CA0B3A591817E92B2"
		$"CCD2D3DADBB368210305041B589EB5A891818093B0CBD2D4DCDEB7671F030503"
		$"1A569CB3A894838092B0C9D0D4DFE1B8692002050318569CB3A794848091B0C8"
		$"D0D5DEE2B96B2303050317539CB4A893858192B0C7CCD2DEE4BD6C2202050219"
		$"569AB1A996858193AFC4CBD3E0E5BD6E2403050314559CB2A895868393AFC2CA"
		$"D1DFE7BD6C2303050318579CB0A694878596AFC0C7D0DEE5BD6C250405041253"
		$"9CB4A995878596AFBFC4CBE0ECC06D2102050313549CB3A996868597B2C1C3CA"
		$"DEEBC06B2102050314529AB3A997898699B2C1C1C9DFEDBE6A2504050310519A"
		$"B1AA9989869BB5C1BFC8DFEABE6B220305030E4F9AB3AA9A8B869AB5C0BFC7DF"
		$"EBC36D200206020A519EB4AA98878499B5C0BFC7E1EEC36C1E0206020D549DAF"
		$"A79786869BB5BFBFC6DDEEC86E1F02060210539BAFA59589899CB3BCBBC4DEEE"
		$"C972220206020F509BB1A695898A9BB0BBBBC2DFF0CA76220206020C509DB2A5"
		$"94898B9EB2BAB8C0DEF3CC72210206020C529DB2A493888AA0B4B7B4C0DBECCC"
		$"77240206020B4F9FB5A7968B8B9FB1B4AFB9D9EFCD7A250206020C54A0B2A697"
		$"8C8D9DADAFAFBAD9EDCA7C2A0305030A4FA2B7A6948A8C9DABACACB8D8EFCF7E"
		$"2C0405020A56A4B3A4948A8C9FAEAAA7B5D7EBD07F2B0405020C53A0B4A6948B"
		$"90A2ADA7A3B1D5EDD0822D04050210569FB3A5938A91A4ACA7A4B0D3EFD0802C"
		$"0405010F569DB1A4938B92A4ACA8A3B0D2EBCE7E300505020C559EAFA1928E92"
		$"A1ACA6A3B2D7EAC982320705030B4E9AB2A4928C94A2A7A2A2B4DBEECC813006"
		$"05020D4D97B1A6938A92A2A8A3A1B4D9EED3822C0505020D4E97AFA49490929F"
		$"A9A4A1B2D4ECD5872E040503104C94AFA6968F929EA4A3A3AFD4F0D485330704"
		$"030B4995AEA59790929EA5A4A5AFD3ECD289340704030B4491B1AA998F909CA4"
		$"A5A4B2D1ECD489360804030B448CAFAC9B919097A1A5A5B2D1EAD88C38090404"
		$"0C3F89AFAB9D928F97A2A6A6B3D1EBDA8C390A04030B3E87AEAA9E939098A0A6"
		$"A7B4D4E9D68F3E0B03030B3F8CB5AD978A87909AA5ACB6D0EBDF9D450A030407"
		$"3A90B6AC978B88919AA3AAB4CEEBE19F460A0304053891BAAF968988929CA3A6"
		$"AFC9ECE5A44A0A030505358EBCB2988685929CA2A6AEC6EBE8A74B0B03050535"
		$"8DBAB2998A868F9BA2A5ACC5EAE7AA510E0205053187BAB49A89868E98A0A5AB"
		$"C5EBE9AE550E0205063082B5B29E8C858D98A0A5AEC6E7EBB1560F0205052C7E"
		$"B4B4A08E868D969FA8B1C8E5E9B157100205052A77B0B7A48E878B939EA8B1C8"
		$"E5E8B45D130205052872AEB6A3908888909CA9B4C9E5E7B76317010505256BAC"
		$"B6A490888A8E98A7B4C8E1E5BB6C1D0105032068ACB6A49087898F96A3B2C8DF"
		$"E7C071210205031E61A8BAA790878A9095A1ADC0DCE7C579260205031C5EA7BA"
		$"A99387878E94A0ABBAD4E2C88232060403195CA2BAAC9587868B919CA6B8D3E2"
		$"CB88390703041A55A1BCAC9386878B8F9DA8B7CFDFCE8E3D0A030316529AB9AE"
		$"9786858B8F9CAAB6CEE0D4953D080403154C94B8B09A8A84878E9EABB7CDE0D2"
		$"9943090304134590BAB09B8E87858999ABBAD0E0D49E4B0C0204104089B8B49D"
		$"8C86878995A9BCCEDFD7A7540D02040E3B84B7B59F8E84828594A9BACBDFDCAF"
		$"5B1202040D3979B2B8A28E84818493ADBDC7DADDB4641802040B3373AEB9A691"
		$"837E8192ADBCC5D6D8BC701D0204092F6FABB8A894847C7E91ADBEC6D3D8BD77"
		$"260204072866A6BBA993847D7E8EADC0C7D0D7C17E2D0504052360A1B9AC9785"
		$"7B7D8DAAC0CAD0D6C786320504041E5A9BB7AD99897D7B8AA9C1CBD1D5C58C3A"
		$"0803041D5391B5B19C8A7E7986A6C1CAD1D5C993440A03041A4C8AB3B19E8D7F"
		$"7883A1BECCD1D6CC9B4B0C0204194783AFB19F8E7F77829FBECCD0D6CD9F5613"
		$"020417457BA9B1A19080777F9CBECED1D3CEA65C180203134078A8AFA1908178"
		$"7F9ABBCFD2D1CBAA661E0204123A6FA3B2A39283767B96B9CDD0D1CDAE6F2805"
		$"030F3467A0B4A79382777A91B7CDCECECBB57A3208020D316298B3AA9683777A"
		$"8DAFC7CBCACBC0893B09020B2A5B96B3A8948578778BAEC6CAC7CCC492440C01"
		$"09285691B3A994867A7989ABC5C7C5CAC5994D11020723538BAFAA97877B7887"
		$"A8C2C4C1C6C69E55180307224E81A9AA99877A7784A2BFC5C0C5C7A560200609"
		$"20487CA7AA9A897D75809FBCC4C1C5C9AA692708091D4677A2AA9A8A7D747C98"
		$"B8C4C1C3C8B2732D0C0A1C44719EAA9B8C7E747993B6C5C1C1C8B87B330D091C"
		$"406C97A99D8D80747790B0C4C1BFC4BA873E110B1B3C6691A79D8F8276758AAE"
		$"C3C1BDC2BA8D4B1A0D1939608BA59E9083777587A8C0C1BCBFBC935322101A34"
		$"5A86A29F9084797585A3BCC0BBBFBD9C5C28131A335681A09D90867A75829FBA"
		$"C0BDBEBDA0632C161B325379999E92877D75809BB7BFBABCBEA56C34191B304F"
		$"72949E968C7F757C96B3BEBABBBFAA753C1C1C2E4A6C92A0988D7F767B93AFBD"
		$"BCBBBDAC7D44211D2D46678DA1998C80777B90ABBBBAB9BDB2844D271E2B4462"
		$"879E998C80767A8CA8B9BAB6BAB38C532B232D425F829B998E82787889A3B7B6"
		$"B3B7B2905A31242D405D7D96998E82797986A0B5B7B1B4B39461382A2F405979"
		$"959990847977859CB1B3B0B2B296683F2E313F5674939990857B778299AEB1AE"
		$"B1B1996D453233415472939A90857C778196AAAFADAEB09C724A35354255708D"
		$"9891877D788295A7ADACAAAEA0774E3A3842536D899791897F788092A5ACA9A9"
		$"AC9F7B543E3A44546A87969288807A808FA3ACA9A7AB9F7E59423E4554698695"
		$"9188807A7F8EA0A9A7A6A99F815E464047556882939088817B7E8C9DA6A6A5A7"
		$"A08561494448556780918F88817C7E8A9AA4A5A3A5A087634D494C56667F908F"
		$"88827C7D8898A2A2A1A49F8667514A4C55667C8E8E88827C7E8795A0A1A0A39F"
		$"896A544D4E57657B8D8E89837E7E87949D9E9EA19D896D574F515765788C8F8A"
		$"837E7E85919B9E9EA09C8A705B53545965778A908A847E7D8590999C9C9E9B8B"
		$"735D55565A65778A8F8A837E7D828D9697989A988A75615B5B5E6776878C8781"
		$"7E7E828B92949495938977676161636977858885807E7E81"
	}
};


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// a simple C major scale covering two octaves

resource 'snd ' (rScaleSnd, "C Major scale", purgeable) {
	FormatOne {
		{	/* array Synthesizers: 0 elements */
		}
	},
	{
/* 1 */  noData, freqDurationCmd {500, (kOctave5 + Ckey)},
/* 2 */  noData, freqDurationCmd {500, (kOctave5 + Dkey)},
/* 3 */  noData, freqDurationCmd {500, (kOctave5 + Ekey)},
/* 4 */  noData, freqDurationCmd {500, (kOctave5 + Fkey)},
/* 5 */  noData, freqDurationCmd {500, (kOctave5 + Gkey)},
/* 6 */  noData, freqDurationCmd {500, (kOctave6 + Akey)},
/* 7 */  noData, freqDurationCmd {500, (kOctave6 + Bkey)},
/* 8 */  noData, freqDurationCmd {500, (kOctave6 + Ckey)},
/* 9 */  noData, freqDurationCmd {500, (kOctave6 + Dkey)},
/* 10 */  noData, freqDurationCmd {500, (kOctave6 + Ekey)},
/* 11 */  noData, freqDurationCmd {500, (kOctave6 + Fkey)},
/* 12 */  noData, freqDurationCmd {500, (kOctave6 + Gkey)},
/* 13 */  noData, freqDurationCmd {500, (kOctave7 + Akey)},
/* 14 */  noData, freqDurationCmd {500, (kOctave7 + Bkey)},
/* 15 */  noData, freqDurationCmd {500, (kOctave7 + Ckey)}
	},
	{
		// array DataTables: 0 elements
	}

};


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The opening to Mussorgskys Promenade in four voices.  This is the main melody.

resource 'snd ' (rMelodyPart1, "Promenade melody", purgeable) {
	FormatOne {
		{
			// array Synthesizers: 0 elements
		}
	},
	{
// ~~~~~~~~~~~~~~~~~~~~~~ measure 1 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Fkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Ebkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Abkey)},
/* 4 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Bbkey)},
/* 5 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Ebkey)},
/* 6 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ckey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 2 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Bbkey)},
/* 2 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Ebkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ckey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Abkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Bbkey)},
/* 6 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Fkey)},
/* 7 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Ebkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 3 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Fkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Ebkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Abkey)},
/* 4 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Bbkey)},
/* 5 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Ebkey)},
/* 6 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ckey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 4 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Bbkey)},
/* 2 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Ebkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ckey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Abkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Bbkey)},
/* 6 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Fkey)},
/* 7 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Ebkey)},
/* 8 */  noData, quietCmd {},
/* 9 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 5 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Ebkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Fkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Ckey)},
/* 4 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Ebkey)},
/* 5 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Fkey)},
/* 6 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 6 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Fkey)},
/* 2 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Gkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Ebkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ebkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ckey)},
/* 6 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Bbkey)},
/* 7 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Abkey)},
/* 8 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Ebkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 7 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Ebkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Fkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Ckey)},
/* 4 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Ebkey)},
/* 5 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Fkey)},
/* 6 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 8 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Fkey)},
/* 2 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Gkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Ebkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ebkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ckey)},
/* 6 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Bbkey)},
/* 7 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Abkey)},
/* 8 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave5 + Ebkey)},
/* 9 */  noData, quietCmd {},
/* 10 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 9 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kMelodyBeat * 3), (kOctave5 + Ebkey)},
/* 2 */  noData, freqDurationCmd {(kMelodyBeat * 2), (kOctave4 + Abkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 10 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kMelodyBeat * 3), (kOctave4 + Ebkey)},
/* 2 */  noData, freqDurationCmd {(kMelodyBeat * 2), (kOctave3 + Abkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ END ~~~~~~~~~~~~~~~~~~~~~~
	},
	{
		// array DataTables: 0 elements
	}

};


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The opening to Mussorgskys Promenade in four voices.  This is the first part
// of the supporting harmony.

resource 'snd ' (rMelodyPart2, "Promenade voice 1", purgeable) {
	FormatOne {
		{
			// array Synthesizers: 0 elements
		}
	},
	{
// ~~~~~~~~~~~~~~~~~~~~~~ measure 1 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 5)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 2 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 6)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 3 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Dkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Bbkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ebkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Gkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ckey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 4 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Bbkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Gkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Abkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Dbkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Fkey)},
/* 6 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave7 + Gkey)},
/* 7 */  noData, quietCmd {},
/* 8 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 5 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 5)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 6 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 6)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 7 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Bbkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Abkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ckey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ebkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Fkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 8 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Bbkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Bbkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ebkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ckey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Dbkey)},
/* 6 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave7 + Bbkey)},
/* 7 */  noData, quietCmd {},
/* 8 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 9 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Abkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Gkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ckey)},
/* 4 */  noData, freqDurationCmd {(kMelodyBeat + (kMelodyBeat / 2)), (kOctave6 + Ebkey)},
/* 5 */  noData, quietCmd {},
/* 6 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 10 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Dbkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ebkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ckey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Dbkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Bbkey)},
/* 6 */  noData, freqDurationCmd {kMelodyBeat, (kOctave5 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ end ~~~~~~~~~~~~~~~~~~~~~~
	},
	{
		// array DataTables: 0 elements
	}

};


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The opening to Mussorgskys Promenade in four voices.  This is the second part
// of the supporting harmony.

resource 'snd ' (rMelodyPart3, "Promenade voice 2", purgeable) {
	FormatOne {
		{
			// array Synthesizers: 0 elements
		}
	},
	{
// ~~~~~~~~~~~~~~~~~~~~~~ measure 1 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 5)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 2 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 6)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 3 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Fkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Gkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Abkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Bbkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ebkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 4 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ebkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ckey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ckey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Fkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Abkey)},
/* 6 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave8 + Bbkey)},
/* 7 */  noData, quietCmd {},
/* 8 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 5 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 5)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 6 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 6)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 7 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ebkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Dbkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ebkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Gkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 8 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Dkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Ebkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Gkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ebkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Fkey)},
/* 6 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave7 + Ebkey)},
/* 7 */  noData, quietCmd {},
/* 8 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 9 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Dbkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Bbkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ebkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Gkey)},
/* 5 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave6 + Abkey)},
/* 6 */  noData, quietCmd {},
/* 7 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 10 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Fkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Gkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ebkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Fkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Dbkey)},
/* 6 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ END ~~~~~~~~~~~~~~~~~~~~~~
	},
	{
		// array DataTables: 0 elements
	}

};


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The opening to Mussorgskys Promenade in four voices.  This is the third part
// of the supporting harmony.

resource 'snd ' (rMelodyPart4, "Promenade voice 3", purgeable) {
	FormatOne {
		{
			// array Synthesizers: 0 elements
		}
	},
	{
// ~~~~~~~~~~~~~~~~~~~~~~ measure 1 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 5)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 2 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 6)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 3 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Abkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Bbkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ckey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ebkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Abkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 4 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Gkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ebkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Fkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Bbkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Ckey)},
/* 6 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave8 + Ebkey)},
/* 7 */  noData, quietCmd {},
/* 8 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 5 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 5)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 6 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, restCmd {(kMelodyBeat * 6)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 7 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Gkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Fkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Abkey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Bbkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Dbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 8 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Fkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Gkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Ckey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Abkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave8 + Bbkey)},
/* 6 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave7 + Gkey)},
/* 7 */  noData, quietCmd {},
/* 8 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 9 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Fkey)},
/* 2 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ebkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Abkey)},
/* 4 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave7 + Bbkey)},
/* 5 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave7 + Ebkey)},
/* 6 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave7 + Ckey)},
/* 7 */  noData, quietCmd {},
/* 8 */  noData, restCmd {(kMelodyBeat / 2)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 10 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave7 + Bbkey)},
/* 2 */  noData, freqDurationCmd {(kMelodyBeat / 2), (kOctave7 + Ebkey)},
/* 3 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Ckey)},
/* 4 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Abkey)},
/* 5 */  noData, freqDurationCmd {kMelodyBeat, (kOctave7 + Bbkey)},
/* 6 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Fkey)},
/* 7 */  noData, freqDurationCmd {kMelodyBeat, (kOctave6 + Ebkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ END ~~~~~~~~~~~~~~~~~~~~~~
	},
	{
		// array DataTables: 0 elements
	}
};


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The song Schaut, ihr Snder arranged by JS Bach.  This is the soprano part.

resource 'snd ' (rCounterPt1, "Bach soprano", purgeable) {
	FormatOne {
		{
			// array Synthesizers: 0 elements
		}
	},
	{
// ~~~~~~~~~~~~~~~~~~~~~~ measure 1 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Akey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 2 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Dkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Ckey)},
/* 5 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 3 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Akey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave7 + Akey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 4 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Ckey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Dkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 5 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Dkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Ebkey)},
/* 5 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Dkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 6 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Ckey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Ckey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave7 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 7 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Dkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Ckey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 8 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Ckey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Ckey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 9 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Akey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave6 + Fkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 10 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Ckey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Dkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 11 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Dkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Ckey)},
/* 5 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Gkey)},
/* 6 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave7 + Akey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 11 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Akey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave6 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ END ~~~~~~~~~~~~~~~~~~~~~~
	},
	{
		// array DataTables: 0 elements
	}
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The song Schaut, ihr Snder arranged by JS Bach.  This is the alto part.

resource 'snd ' (rCounterPt2, "Bach alto", purgeable) {
	FormatOne {
		{
			// array Synthesizers: 0 elements
		}
	},
	{
// ~~~~~~~~~~~~~~~~~~~~~~ measure 1 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Dkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Dkey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Dkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Dkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 2 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Gkey)},
/* 5 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Gbkey)},
/* 6 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 3 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Gkey)},
/* 2 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Gbkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Gkey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Gkey)},
/* 5 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave6 + Dkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 4 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 5 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave7 + Bbkey)},
/* 5 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave7 + Akey)},
/* 6 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 6 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Bbkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Akey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave6 + Fkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 7 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kBachBeat + (kBachBeat / 2)), (kOctave6 + Fkey)},
/* 2 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Gbkey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Dkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 8 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {(kBachBeat + (kBachBeat / 2)), (kOctave6 + Gkey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 4), (kOctave6 + Fkey)},
/* 5 */  noData, freqDurationCmd {(kBachBeat / 4), (kOctave6 + Ekey)},
/* 6 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 9 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Ekey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave6 + Ckey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 10 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave7 + Akey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave7 + Bbkey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Abkey)},
/* 5 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 11 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
/* 5 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 12 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gbkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave6 + Dkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ END ~~~~~~~~~~~~~~~~~~~~~~
	},
	{
		// array DataTables: 0 elements
	}
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The song Schaut, ihr Snder arranged by JS Bach.  This is the tenor part.

resource 'snd ' (rCounterPt3, "Bach tenor", purgeable) {
	FormatOne {
		{
			// array Synthesizers: 0 elements
		}
	},
	{
// ~~~~~~~~~~~~~~~~~~~~~~ measure 1 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Bbkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Akey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Gkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 2 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Bbkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Ckey)},
/* 5 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Dkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 3 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kBachBeat + (kBachBeat / 2)), (kOctave6 + Dkey)},
/* 2 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Ckey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave5 + Gbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 4 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Dkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Ckey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Bbkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Dkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 5 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Ebkey)},
/* 5 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 6 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Gkey)},
/* 2 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Fkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Ebkey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave6 + Dkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 7 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Bbkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Ckey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Dkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 8 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Ckey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Bbkey)},
/* 5 */  noData, freqDurationCmd {(kBachBeat + (kBachBeat / 2)), (kOctave6 + Akey)},
/* 6 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 9 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Akey)},
/* 2 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Bbkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Ckey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Bbkey)},
/* 5 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave6 + Akey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 10 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Dkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Ebkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 11 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Fkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Ekey)},
/* 5 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Bbkey)},
/* 6 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Ckey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 11 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kBachBeat + (kBachBeat / 2)), (kOctave6 + Dkey)},
/* 2 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Ckey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave6 + Bkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ END ~~~~~~~~~~~~~~~~~~~~~~
	},
	{
		// array DataTables: 0 elements
	}
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The song Schaut, ihr Snder arranged by JS Bach.  This is the bass part.

resource 'snd ' (rCounterPt4, "Bach bass", purgeable) {
	FormatOne {
		{
			// array Synthesizers: 0 elements
		}
	},
	{
// ~~~~~~~~~~~~~~~~~~~~~~ measure 1 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kBachBeat + (kBachBeat / 2)), (kOctave5 + Gkey)},
/* 2 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Gbkey)},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Gkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave4 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 2 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave4 + Gkey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Akey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Bbkey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Ckey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 3 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Dkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Ebkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave5 + Dkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 4 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Dkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Bbkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Akey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Bbkey)},
/* 5 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 5 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Bbkey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Abkey)},
/* 5 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Gkey)},
/* 6 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Fkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 6 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Ekey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Fkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave5 + Bbkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 7 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Bbkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave6 + Bbkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave6 + Akey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Gkey)},
/* 5 */  noData, freqDurationCmd {kBachBeat, (kOctave4 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 8 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Ekey)},
/* 4 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Fkey)},
/* 5 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Dkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 9 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Ckey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Ckey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave4 + Fkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 10 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Gkey)},
/* 2 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Fkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Ebkey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Dkey)},
/* 5 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Bbkey)},
/* 6 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Ebkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 11 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, quietCmd {},
/* 2 */  noData, restCmd {kBachBeat},
/* 3 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Bkey)},
/* 4 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Ckey)},
/* 5 */  noData, freqDurationCmd {(kBachBeat / 2), (kOctave5 + Dkey)},
/* 6 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Ekey)},
// ~~~~~~~~~~~~~~~~~~~~~~ measure 11 ~~~~~~~~~~~~~~~~~~~~~~
/* 1 */  noData, freqDurationCmd {kBachBeat, (kOctave5 + Dkey)},
/* 2 */  noData, freqDurationCmd {kBachBeat, (kOctave4 + Dkey)},
/* 3 */  noData, freqDurationCmd {(kBachBeat * 2), (kOctave4 + Gkey)},
// ~~~~~~~~~~~~~~~~~~~~~~ END ~~~~~~~~~~~~~~~~~~~~~~
	},
	{
		// array DataTables: 0 elements
	}
};


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// This is the sampled sound used in the about window.

resource 'snd ' (rMoofSound, "Moof", purgeable) {
	FormatOne {
		{sampledSynth, 0}
	},
	{hasData,bufferCmd {20}		// buffer data offset is 20 bytes
	},
	{27616,						// number of samples
	 Rate22K,					// sample rate
	 0,							// loop starting point
	 27615,						// loop ending point
	 0,							// standard sound header
	 60,						// base frequency
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80807F7F"
		$"7F7F7F7F 7F7F8080 80808080 80808080 80808080 7F7F7F7F 7F7F7F7F 7F7F7F7F"
		$"7F7F7F7F 7F7F7F7F 7F7F8080 80808080 80808080 80808080 7F7E7D7D 7C7B7A7A"
		$"79787674 72727272 74767878 7877787A 7C7D7D7D 7D7F7F7F 7F7F7F7F 80828383"
		$"83838282 82838587 89898A8A 89878686 87898989 88868484 84848484 83838282"
		$"817F7E7D 7D7E7F7F 7F7F7E7C 7B797A7C 7D7D7E7E 7D7D7D7D 7D7F8080 80808080"
		$"80808080 80808080 7F7E7E7E 7F7F7F7F 7F7F7F7F 7F7F7F7F 7F808080 7F7F7F7F"
		$"80808080 80807F7F 7F7F8080 80808080 80808080 80808181 81818181 80808080"
		$"81818181 81818181 81818181 81818181 82828282 83838383 82828282 83838382"
		$"82828282 82828282 82828181 807F7F7F 7F7F8080 7F7F7F7F 7F7F8080 81818181"
		$"80808081 82828181 81818180 7F7F7F7F 80807F7F 7E7E7E7E 7F7F7F7F 7F7F7F7F"
		$"7F7F7F7F 7F7F7F7F 7F7F8080 80807F7F 7E7E7D7C 7B7A7978 77757372 71717274"
		$"75777776 7677797C 7D7B7C7E 7F7F7F7F 7F808181 82848484 83828282 83858688"
		$"88888786 85848587 88888886 84848382 82828383 82818080 7F7D7D7D 7D7D7D7E"
		$"7E7D7B79 787A7B7C 7D7D7C7B 7B7B7B7D 7E7F8080 807F7F7F 7F7F7F7F 8080807F"
		$"7F7F7F7F 80808080 80807F7F 7F7F7F80 81818080 80808080 80808181 807F7F7F"
		$"80808080 81818080 80808080 81818181 81818181 80808081 82828181 80808080"
		$"81818282 81818181 81818181 81818282 82828282 83838484 83838281 81818181"
		$"81807F7F 7F7F7F7F 7E7E7E7E 7F7F7F7F 7F808181 82828181 81818282 81818181"
		$"81807F7F 7F7F7F7F 7F7F7F7E 7E7E7F7F 7F7E7E7E 7F7F7F7F 7F7F7F7F 7F7F8080"
		$"80808080 7F7F7E7C 7A787878 77757371 706F7072 75777776 77797B7C 7C7B7C7E"
		$"80808181 81818080 81848686 85848382 82838587 88898988 87858486 87888887"
		$"86868583 82828282 82828282 807E7D7D 7D7D7D7E 7E7D7B79 7878797B 7C7D7E7E"
		$"7D7C7C7D 7E7F8080 81818080 7F7F7F7F 80828281 80807F7F 7F7F8080 80807F7E"
		$"7E7E7E7F 80818180 7F7F7F7F 7F808181 80808080 7F7F7F7F 80818181 80808081"
		$"82828282 82828181 81818182 83838383 82808081 82828282 82828282 82828383"
		$"83838383 82828282 82828282 83838383 82828282 83838281 80807F7F 7F7F7F80"
		$"80808080 7F7E7E7F 80808080 80818181 80808081 82828181 807F7F7F 7F7F8080"
		$"7F7F7F7F 7F7F7F7F 7F7F8080 7F7F7F7F 7F7F8080 80808080 7F7F7E7E 7D7C7B7A"
		$"79777573 71707070 71737575 76767678 7A7C7C7B 7B7D7E7F 80808181 81818385"
		$"85848484 84848485 86888989 88878685 85878888 88868585 84838383 82828181"
		$"8181807E 7D7D7D7D 7D7D7D7C 7B797878 797B7C7C 7C7C7C7C 7C7E7F7F 7F808181"
		$"80808080 81818182 82828180 80808080 80808080 7F7F7E7E 7E7E7F80 807F7F7F"
		$"7F7F7F7F 8080807F 7F7F8080 7F7F7F80 80808080 80808080 81818282 82828181"
		$"81828383 82828181 81818181 81818181 81818181 81828383 84848383 82828282"
		$"82828282 82828282 82828282 81808080 7F7F7F80 80807F7F 7E7E7E7E 7F7F8080"
		$"7F7F7F7F 80808080 80807F7F 7F7F7F7F 7F7F7F7F 7E7E7E7E 7E7E7F7F 7E7E7E7E"
		$"7E7E7E7E 7E7F8080 7F7F7F7F 7F808181 807F7E7C 7B797878 77777573 716F6F6F"
		$"71747676 76767779 7B7B7B7B 7D7F8080 7F7F7F7F 7F818385 85858482 82828486"
		$"888A8A8A 89878585 86888989 88868583 82828282 83838383 82807F7E 7E7E7F7F"
		$"8080807E 7C7A797B 7C7E7F7F 7E7C7C7C 7D7F8081 82828181 80808080 80828282"
		$"81808080 7F7E7E7F 8080807F 7E7E7E7E 7E808181 807F7F7F 7F7F8080 80808080"
		$"7F7F7F7F 7F7F8080 80807F7F 7F7F8080 81818181 80808080 80818181 80808080"
		$"80808181 81818181 81818181 82828282 83838282 82828282 82828282 81818181"
		$"81828282 81808080 80808080 7F7F7F7F 7F7F7E7E 7E7E7E7E 7F7F8081 81808080"
		$"80808080 7F7F7E7E 7E7E7F7F 80808080 81818181 81818181 82828282 81818080"
		$"80808181 807F7E7E 7D7D7C7A 79797878 77757371 70707173 75777777 7777797B"
		$"7C7C7D7F 80807F7F 7F808182 83858585 84848383 83858789 8B8B8A88 87858586"
		$"888A8A88 86848281 81818282 82818080 7F7D7D7D 7D7D7E7F 7F7D7C7A 79797A7C"
		$"7D7D7D7C 7B7B7B7C 7D7F8080 80807F7E 7E7F8080 80818180 7F7F7F7F 7F7F7F7F"
		$"7F7F7E7E 7E7E7E7E 7F80807F 7F7F7F7F 7F7F8080 807F7F7F 7F7F7F7F 7F808080"
		$"80808080 80808181 80808080 80808081 82828282 81818181 81818282 82828282"
		$"82828283 84848585 84848484 84848484 84848383 82828282 82828180 7F7F7F7F"
		$"7F808080 80807F7F 7F7F7F7F 7F7F8080 80808181 81818181 81818080 80808080"
		$"80808080 80808081 81808080 81818181 80808080 81818181 807F7E7E 7D7D7C7A"
		$"79787777 76757371 6F6F6F70 72757675 7576787A 7B7A7A7C 7D7F7F7E 7E7F7F7F"
		$"80828484 84848383 83848688 898A8B8B 89878686 87888989 88868482 82828282"
		$"82828282 817F7E7D 7D7D7E7E 7F7F7E7C 7A79797B 7C7D7E7E 7D7C7C7C 7C7E7F7F"
		$"80808080 7F7F7F7F 7F808181 807F7F7F 7E7E7E7E 7F7F7F7F 7E7D7D7D 7E808181"
		$"807F7F7F 80808080 7F7F7F7F 80807F7F 7F7F8080 80808080 80808080 81818180"
		$"80808080 81818282 81818181 81818282 82828282 82828283 84848484 84848383"
		$"83838383 83838484 83828282 82828282 81808080 80808080 80808080 7F7F7E7E"
		$"7E7F8080 7F7F7F81 81808080 80808080 80808181 81818181 81818181 81818080"
		$"80808181 80807F7F 7F808080 80807F7F 7E7E7D7B 79787777 76767471 6F6F6F6F"
		$"70747575 74747578 7A7A7B7B 7C7E7E7E 7E7E7F7F 7F808284 84848382 82828486"
		$"888A8B8B 89878686 86888989 89878584 83838383 83838383 82807F7D 7D7D7E7E"
		$"7E807F7D 7B797878 797B7C7C 7C7B7A7A 7A7C7D7E 7F7F7F7F 7F7F7F7F 7F7F8081"
		$"81818080 7F7E7E7E 7F7F7F7F 7E7E7E7E 7E7F8080 807F7F7F 7F7F7F80 80807F7F"
		$"7F7F7F7F 7F7F7F7F 80808080 80808080 81818282 81818181 81818282 82818181"
		$"81818282 83838282 82828282 83838383 84848383 83838484 83838383 83828282"
		$"82828181 80807F7F 7E7E7E7E 7E7E7E7E 7F7F7E7D 7D7E7F7F 7E7E7E7F 80808080"
		$"81818181 81818181 80808080 80808080 80808181 807F7F7F 7F7F8080 81818080"
		$"80807F7F 7F7F7E7E 7D7B7977 76767574 73716F6E 6E707274 75757677 797B7C7A"
		$"797B7C7E 7F7F7F7F 7F7E7F81 83848585 84848484 8587888A 8A8A8987 86868789"
		$"89888786 85838281 81818181 82828280 7E7D7D7D 7D7F8080 7E7C7A78 78797A7C"
		$"7D7D7D7B 7B7B7C7E 7F7F7F80 80808080 7F7F7F80 81818180 7F7F7E7E 7E7E7F7F"
		$"7F7F7E7E 7E7E7E80 81818080 80808080 80808181 81818181 80808080 81818181"
		$"81818181 81818282 81818181 81818181 82828282 82828282 82828282 82828282"
		$"82828383 83838484 84848484 84848484 83838282 82828282 81807F7F 7F7F7E7E"
		$"7E7E7E7E 7E7E7E7E 7E7E7E7E 7E7E7E7E 7F7F7F7F 7F7F7F80 81818181 81818181"
		$"82828281 81818282 82828180 80808181 81818080 80808080 81818282 82828181"
		$"80807F7E 7D7B7A79 77757371 6F6F7072 74767674 7476787A 7A7A7B7D 7E7E7E7E"
		$"7E7F8080 81838484 83838282 83858789 8A8A8987 86858586 87888886 85848382"
		$"82828383 83838282 817F7F7F 7F7F8080 80807E7C 7A7A7A7B 7C7D7D7C 7B7A7A7B"
		$"7C7E7F7F 7F7F7E7E 7E7E7F7F 7F808080 7F7F7E7E 7E7E7E7E 7F7F7E7E 7E7E7E7E"
		$"7E808080 7F7F7F7F 7F7F8080 807F7F7F 80807F7F 7F7F7F7F 7F7F7F7F 80808080"
		$"81818181 81818181 82828282 81818080 80818181 81818181 81818181 82828383"
		$"84848383 83838484 84848383 82828282 81818080 80807F7F 7F7F7F7F 7F7F7F7F"
		$"7E7E7E7E 7E7F8080 80808080 7F7D7D7E 7F7F8080 80808080 81818282 82828282"
		$"82828282 81818181 82828181 80808080 81818181 81808080 80807F7F 7E7C7B79"
		$"78777572 706E6D6D 6F717374 74737476 78797A7A 7B7D7D7C 7C7C7D7D 7E7F8082"
		$"83838282 82828385 87898B8B 8A888786 8688898A 8A888684 82828282 82828282"
		$"8181807F 7F7F8080 81828280 7E7C7A7A 7A7C7D7D 7D7C7B7A 7A7C7D7E 7F7F7F7F"
		$"7F7F7F7F 7F7F8081 81818080 7F7F7E7E 7E7E7F7F 7F7E7D7D 7D7D7E7F 80807F7F"
		$"7F7F8080 81818080 80808080 7F7F7F7F 7F7F7F7F 7F7F7F7F 7F7F8080 80808080"
		$"80808181 82828180 80808181 82828282 82828282 83838484 84848585 84848484"
		$"84848484 85858482 81818181 81818080 7F7F7F7F 7F7F7F7F 7F7F7F7F 7E7E7E7E"
		$"7F7F8082 83838281 81828383 84848484 83828282 82828383 83838281 80808080"
		$"80808181 81818181 81818182 83838280 7F7D7B79 79797878 76726F6D 6C6C6D71"
		$"73737373 74767779 7A7A7B7D 7F7F7E7E 7D7D7D7F 81848584 83818080 81838689"
		$"8B8B8A88 86858587 88898987 86848281 81818282 82828281 807E7D7D 7E7F8080"
		$"807F7D7A 7979797B 7C7D7D7C 7B79797B 7C7D7E7E 7F7F7F7F 7F7F7F7F 7F808181"
		$"81807F7F 7F7F7F7F 80807F7F 7E7D7D7D 7E808080 7F7F7F7F 80808080 80808080"
		$"80808080 80808080 7F7F7F7F 7F7F7F80 81818181 80808080 81818282 82828181"
		$"81818282 81818181 82828282 83838484 85858484 84848484 85858484 83828181"
		$"81818181 80808080 7F7F7F7F 80808080 807F7F7F 7F7F8080 81828282 81818181"
		$"81818282 82828181 81818080 7F7F7F7F 80807F7F 7F7F8081 82828282 82828282"
		$"81807F7D 7B797776 75747371 6F6D6B6B 6C6E7173 74747476 787A7B7A 7B7D7E7F"
		$"7F7F7F7F 7F7F8082 84848483 82818183 85888A8B 8B898886 85868789 89878686"
		$"85838282 82828383 83848482 807F7F7F 7F818181 7F7C7A78 77797A7C 7D7D7C7A"
		$"79797B7D 7E7E7E7F 7F7F7F7F 7F7F7F80 81828281 80807F7D 7D7D7E7F 80807F7E"
		$"7D7D7D7F 80808080 80808080 81818080 80808080 80808080 80808080 80808181"
		$"81818181 82828282 81818182 82828282 81818181 82828282 82828282 82828282"
		$"83838383 82828282 82828282 83838383 82828282 82828282 81808080 80807F7F"
		$"7F7F7F7F 80807F7F 7F7F8081 82828281 81818282 81808080 80808181 80807F7F"
		$"7F7F8080 81818181 81818181 82828383 8282817F 7E7C7A78 76767575 73706D6B"
		$"6A6A6D71 73737373 74767778 797B7D7F 80808080 80808081 83858686 85848382"
		$"8385878A 8C8C8B89 87858587 898B8B89 88868483 83838484 84848484 83817F7E"
		$"7E7E7F80 80807E7B 79787879 7B7D7D7B 7A78787A 7B7B7C7C 7D7D7E7E 7E7E7F7F"
		$"7F7F8080 80807F7F 7E7D7D7D 7D7D7E7E 7E7D7D7D 7E7F8080 80808080 80808181"
		$"81808080 80808080 80807F7F 7F7F8080 80808080 80808181 80808080 80818282"
		$"81818080 80808181 82828282 82828282 83838383 84848484 84848383 83838484"
		$"84838282 82828282 81818080 80808080 7F7D7D7D 7D7E7F7F 80807F7F 7F818283"
		$"84848484 83838383 82828282 82828281 807F7F80 81818282 81808080 80818283"
		$"84848381 807E7C7B 7A787775 7472716F 6D6B6A6A 6B6F7172 72727375 77797A7A"
		$"7B7D7E7E 7F7F7F7F 80808183 84848383 83838385 87898A8B 8B8A8988 8888898B"
		$"8B898886 85848484 84848484 84848482 81818080 80818181 807E7C7A 79797A7B"
		$"7C7C7B79 78797A7C 7C7C7C7C 7D7D7E7F 7F7F7F7F 7F7F8080 7F7F7E7E 7E7E7E7E"
		$"7F7F7E7E 7D7D7D7E 7F808080 80808080 80818180 80808080 80807F7F 7F7F7F7F"
		$"80808080 81818181 81818181 80808080 81818282 81818080 80808181 82828282"
		$"82828282 83838383 83838383 83838281 81828383 82828282 82828282 82828181"
		$"80807F7F 7F7F8080 80808181 82828282 83848484 84848383 82828282 82828181"
		$"81818080 80808080 81818181 80808080 81818181 807F7E7C 7A787676 7573706C"
		$"69686869 6C707170 70707377 7878797B 7D7F7F7F 7F80807E 7F828486 86848382"
		$"82828486 888A8B8B 8A888787 888A8B8B 8A888785 84848585 84848485 85838280"
		$"80808182 83848381 7E7B7A7A 7A7B7C7D 7D7B7978 797B7C7C 7C7C7D7D 7D7D7D7E"
		$"7E7E7E80 81818180 7F7D7C7C 7C7D7E7E 7E7E7D7B 7B7C7D7E 7F7F7F7F 7F7F8080"
		$"80808080 80808181 807F7F7F 7F7F8080 80808080 80808181 82828180 80808081"
		$"81818181 81818181 81818181 81818181 81818282 82828282 82828383 83838383"
		$"84848484 83838383 84848484 83828181 80808080 80808181 81818182 83838485"
		$"86868585 84848484 84848484 83818080 80818181 81818080 80808181 82828282"
		$"817F7E7C 7A787776 74726E6A 67666667 6A6E6F6F 6F6F7074 7676777B 7D7D7E7E"
		$"7F7F7F7F 81858787 86858482 82838588 8A8B8B8A 89878686 87898B8B 8B898784"
		$"83838485 86868686 86848281 81818181 82838381 7E7A7979 79797A7B 7B797876"
		$"77797A7B 7C7C7C7B 7B7B7C7C 7C7C7C7E 7F7F7F7F 7E7D7C7C 7C7D7E7E 7E7E7D7C"
		$"7C7C7D7E 7F7F8080 7F7F7F80 81818181 81818282 81818080 80808181 81818181"
		$"81818181 82828281 81818282 83838282 81808080 81818181 80807F7F 7F7F7F7F"
		$"80808181 81818282 82828383 84848383 82828282 83838381 81818080 80808080"
		$"80818282 82828282 83838383 84848484 84848383 83838383 83818181 82828181"
		$"81818282 82838382 81818080 7F7E7D7B 79777573 706C6967 6666676A 6C6C6D6D"
		$"6E707273 7476787A 7C7C7D7F 80808284 85878786 85848383 83858789 8A8A8987"
		$"86868688 898B8B8B 8A888685 85868787 87888887 86848383 83838384 8484827E"
		$"7C7A7A7A 7B7C7C7C 7B79797A 7B7B7C7C 7C7C7C7C 7C7C7C7C 7C7D7E7F 7F7F7F7F"
		$"7E7C7C7C 7C7D7E7E 7D7D7C7C 7C7D7E7F 80808080 80808181 81818181 81818282"
		$"81818181 81818282 83838383 83838383 83838282 82828383 83838282 81818181"
		$"81818181 807F7F7F 7F7F7F7F 80808181 80808080 81818181 82828281 81818181"
		$"81818080 80807F7F 7F7F7F7F 80818181 81818181 82828384 85858585 85858484"
		$"84848484 84848484 83828282 82828383 83838282 8282817F 7E7C7B79 7775726F"
		$"6C686665 6668696B 6B6B6B6C 6E707273 7577797A 7B7D7E7F 80808285 86868585"
		$"84828283 85888A8A 8A8A8987 8787898B 8D8D8C8A 89878787 87878787 88888785"
		$"83838282 82838484 83817E7C 7A7A7A7B 7C7C7C7A 7979797A 7B7B7B7B 7B7B7B7B"
		$"7B7B7B7B 7B7C7D7D 7E7E7D7C 7B7B7B7B 7C7C7C7C 7C7C7C7C 7D7D7E7E 7F7F7F7F"
		$"80808181 81818181 82828282 82828282 82828383 84848484 84848383 82828282"
		$"83838484 83828282 82828181 8181807F 7E7E7E7E 7E7E7F7F 80807F7F 7F7F7F7F"
		$"80818282 81808080 80808081 81808080 7F7F7F7F 80808181 82828282 83838383"
		$"84868787 86868686 86868686 85858585 85858585 85858585 85858686 86858483"
		$"82807F7D 7B797774 706C6866 64646568 6A6A6969 696B6C6E 70727577 787A7B7B"
		$"7B7B7D81 83858585 84828181 8285888A 8B8B8B89 8787888A 8C8E8E8D 8B898787"
		$"87878888 88898988 87858585 85858686 86858381 7E7C7B7B 7C7D7D7D 7C7A7A7A"
		$"7A7B7C7C 7C7C7C7C 7B7B7B7B 7B7B7C7D 7E7E7D7D 7C7C7B7A 7A7B7C7C 7C7C7B7B"
		$"7B7B7C7D 7E7E7E7E 7F7F7F7F 80808181 81818282 82828282 82828383 84848484"
		$"84848484 85858484 84848484 84848382 81818181 81818181 80807F7F 7F7F7F80"
		$"81818181 81818080 80818282 82828281 81818181 81818080 80808080 7F7F7F7F"
		$"7F808181 82828282 82828384 85858585 85858585 85858686 86868585 84848484"
		$"84848586 87878685 84838280 7E7C7A78 75716C68 65636466 68686867 67686A6C"
		$"6E6F7072 74777979 7A7A7B7D 7F818282 8281807E 7F818385 86888887 86848587"
		$"898B8C8E 8E8C8A88 88898989 898A8B8B 8A888888 87878789 89898886 83807E7D"
		$"7D7E7F7F 7F7D7C7B 7B7C7D7D 7D7C7C7C 7C7C7B7B 7B7B7B7B 7C7C7C7B 7A7A7978"
		$"78797A7A 7A7A7979 79797A7A 7B7B7C7C 7C7C7C7D 7E7E7F7F 7F808181 81818181"
		$"81818182 83838383 84848484 85858483 83838484 85858584 83838282 81818181"
		$"80807F7F 7F7F7F7F 80818181 81818181 80808080 81818282 81807F7F 7F7F8080"
		$"80807F7E 7E7E7E7E 7F7F8080 80808080 81818183 84858686 85858585 85858686"
		$"86868685 85858585 86868788 89898888 87858482 7F7D7A76 726D6967 6666686A"
		$"6A686868 696B6D6F 70727476 7778797A 7B7C7E80 82828180 7F7E7E80 82848688"
		$"88878685 8586888A 8B8D8C8A 88878787 88888989 89898887 87878888 898A8B8B"
		$"8A888683 817F7F7F 80807F7E 7D7B7A7B 7C7D7D7D 7C7C7C7C 7C7C7B7B 7B7C7D7D"
		$"7C7C7B7B 7A797979 79797A7A 79787878 7878797A 7B7B7A7A 7A7B7C7D 7E7E7F7F"
		$"80808080 80808080 81818283 84848484 84858686 86868686 86868787 87878686"
		$"85858484 83838282 81818181 81818282 82828181 81818282 82828282 83838282"
		$"81808080 80807F7F 7E7E7E7E 7E7E7E7E 7E7E7D7D 7D7D7E7F 80818282 83838484"
		$"83838383 83848585 85848484 85858687 888A8A8A 89898886 8583817F 7B76716C"
		$"69676769 6A6B6B69 696A6B6D 6E707274 76787979 797A7C7E 80828484 82807E7D"
		$"7E808385 87888886 84848486 888A8C8C 8A888686 86868787 88888887 86868686"
		$"8688898B 8B8A8987 84828080 80807F7F 7E7D7C7B 7B7D7E7E 7E7E7D7D 7D7D7C7C"
		$"7C7C7C7D 7E7E7E7D 7C7B7A7A 7A7A7A7A 7A7A7978 77777778 79797A7A 7A7A7A7A"
		$"7B7C7D7D 7E7E7F7F 7E7E7E7E 7E7E7F7F 80818282 82828283 84848484 84848485"
		$"86868686 85848484 84848383 82828181 81818282 83838383 83838383 83838383"
		$"83838484 83818182 82828180 7F7F7F7F 7E7D7D7D 7D7D7D7D 7C7C7C7C 7D7D7E7F"
		$"80818181 81828383 83838383 83838484 84848585 85858687 88888989 89878684"
		$"817F7B76 716C6967 67696A6A 69676668 6A6D6F71 72737577 78797A7C 7D7F8183"
		$"83838280 80808183 85878888 87858484 8587898B 8C8C8A86 85858586 87888888"
		$"87858484 84858688 89898989 88868381 80808080 80807F7D 7B7B7B7D 7E7F7F7E"
		$"7D7C7C7C 7D7D7D7D 7E7E7E7E 7E7D7C7C 7B7A7A7A 7B7B7B7A 79787878 7878797A"
		$"7A7A7A7A 7A7B7C7C 7D7E7F7F 7F7F7E7E 7E7E7E7E 7E7F8081 82828282 83838484"
		$"85858585 85868787 86868686 86868585 85858483 82828283 84848585 85848383"
		$"83838384 84848383 82828282 82828282 81807F7F 7E7E7E7F 80807F7E 7D7D7D7D"
		$"7D7D7E7E 7F7F7F7F 80808080 81818282 83838383 83838383 84848585 86878889"
		$"8A8A8988 8785827F 7B75706B 6868686A 6B6B6A68 68696B6D 6E707173 75777878"
		$"7A7C7F81 82848482 817F7F7F 81838587 88888785 83838486 898B8C8A 88868585"
		$"85878888 88878684 84848486 888A8B8B 8A8A8886 84828080 80807F7F 7E7C7C7C"
		$"7C7E7F7F 7F7E7D7C 7C7C7C7C 7D7D7E7E 7E7E7D7D 7C7B7A7A 7A7A7A7A 7A7A7978"
		$"78787878 79797A7A 7A7A7A7B 7C7D7E7E 7F7F7F7F 7E7E7E7E 7E7E7F7F 80818282"
		$"83838484 85858585 85858585 86868787 86868686 85858484 83838282 82828383"
		$"83838383 82828282 82828383 84848382 81818181 82828281 807F7F7F 7F7F7E7E"
		$"7D7D7D7D 7C7C7C7C 7C7D7E7F 80808080 80818282 82828282 83838383 83838484"
		$"85868787 88898989 88878684 817E7A75 706A6765 66686969 69686868 696B6D6F"
		$"71737577 7979797B 7D7F8284 84848381 80808183 85878989 88868483 8486888B"
		$"8C8C8A87 85848486 87888989 88868585 8586888A 8B8B8B8B 89878583 81818181"
		$"81807F7E 7D7D7D7E 7F7F7F7E 7D7C7C7C 7C7C7C7C 7D7D7E7E 7D7D7C7B 7A7A7A7A"
		$"7A7A7A7A 79787777 77787979 7A7A7A7A 7A7B7C7C 7D7E7F7F 7F7F7E7E 7E7E7E7E"
		$"7F7F8081 82828383 84848585 85858686 86868787 87878686 85858585 84848383"
		$"82828282 83838484 84828181 80808080 80808080 7F7F7F7F 80808181 80807F7F"
		$"7F7F7F7F 7E7E7D7D 7C7C7C7C 7D7E7F80 81838485 86868787 87878787 86868686"
		$"85858484 84848587 898B8D8E 8E8E8C8A 8784807C 766E6458 504C4C50 545A5D5D"
		$"5E606367 6C727678 7A7A7A78 77757576 797D8082 84848585 878C9299 9EA09E99"
		$"948E8B8D 90959898 958D8681 8082868C 90929291 908E8C8C 8C8E8F91 908E8983"
		$"7C76716F 6E707172 72706F6F 72767A80 83848381 7E7A7876 75747474 75757575"
		$"75757779 7C808385 85848381 81828486 8583807C 79777779 7B7F8181 81807F7E"
		$"7E808285 87898987 84828080 80828280 7E7C7977 76787A7E 81838383 83838385"
		$"87898988 8683807D 7C7C7C7E 7F7F7F7E 7D7D7E80 82848687 87878685 84838282"
		$"82828180 7E7B7979 7A7C7F81 83838280 80808284 87898A8B 8B8A8989 88878787"
		$"88888888 87858484 8587898B 8D8E8E8E 8D8D8D8D 8C8A8681 7A726759 4D413A39"
		$"3D464F57 5B5B5C5F 656D747A 7D7D7B78 7573716F 71757A80 85898C8D 8E8E9197"
		$"9DA4A8A8 A49C948D 8A8C8F95 9898948C 85818083 878D9092 9291908E 8D8D8D8F"
		$"9193928E 887E756D 6867686B 6D6D6C6A 6A6E737B 82898C8B 88827C76 716D6C6C"
		$"6C6D6E6E 6F6F7174 787E8389 8C8C8B88 86848384 8585837F 7B777575 787C7F83"
		$"84848383 83838486 888A8B8B 89878583 83838484 82807D79 76767779 7B7E8082"
		$"83838485 87898B8B 8A86827C 78747273 75787A7C 7D7E7F80 82868A8E 9091908C"
		$"88827F7D 7C7E7F7F 7D7B7978 797B7F83 86898A89 8783807E 7E818488 8A8B8A87"
		$"85838486 898D8E8C 89858280 8185888C 8F919393 92919090 8F8D8880 74635240"
		$"342E313D 49565E60 605F6268 707A8082 7F79736D 6967686D 747C8389 8C8D8E8F"
		$"9399A0A9 AEAEA99F 958B8687 8C959B9D 9990877F 7C7F858D 9294938F 8C888788"
		$"8B909497 97948D83 796F6A69 6A6C6D6C 69656365 6B76818C 92949087 7E746E6A"
		$"69696A6A 6A696868 6A6E747C 838A8E8E 8D898582 82848688 86817B75 7171747A"
		$"7F858888 86838181 8284878A 8C8C8B89 86848383 84858583 7F7A7672 71727579"
		$"7C7F8182 8384868A 8C8D8C89 847E7872 6F6F7074 777B7D7E 7F808387 8C929595"
		$"928C8783 80808080 7F7D7C7A 79797A7E 8286898C 8D8B8781 7C7A7A7D 81858888"
		$"87858484 868A8D8F 8E8C8883 80808184 888C8E8F 90909191 9191908E 8A837765"
		$"523E312B 2E3C4A58 5F60605F 626A727C 81817E78 726C6764 6569707A 828A8E8E"
		$"8F8F939B A4AEB4B4 ADA19488 83878E99 A0A29E94 8A817E82 878F9393 928E8A87"
		$"86878A90 95999995 8E82786E 6868696C 6D6B6763 61636A76 818C9294 90887F75"
		$"6D696767 67686969 69696B6F 757D848B 8F8F8D88 84807F81 8284827E 79737171"
		$"757B8084 87878785 83818182 84878A8C 8C8A8785 84858687 8684807A 76727173"
		$"767A7D7F 80808284 878B8D8D 8B868078 736F6E70 7275787A 7C7E8287 8C929698"
		$"97938F8A 86827F7E 7E7E7E7E 7D7C7C7D 7F83878C 8F8F8C88 837E7B7B 7C7E8082"
		$"83838484 8587888A 8B8B8B89 86838180 8184888C 91959898 9795918C 857B6E5C"
		$"4B3B3231 36424C54 5A5E6268 6E767B7D 7D7A7773 6E6A6662 63676E78 80878C8E"
		$"909298A1 A9B0B2AF A79B928C 8B90969C 9D9A948C 8785878C 90939492 8F8B8886"
		$"87898D91 9395928C 83797069 66686A6C 6B686664 666C7581 8C949693 8D837971"
		$"6B686767 68686868 686A6D73 7A82898E 908F8C87 84828182 8282807C 78747374"
		$"777D8286 89898886 84828283 85898B8B 8A878584 84848585 84807C78 75737375"
		$"777A7D80 83858789 8A8A8A88 857F7A74 706C6B6C 6F73787C 8084888C 90949799"
		$"97938D87 82807F80 8181807E 7C7C7D7F 83888C8E 8E8C8884 807D7B7B 7C7E8082"
		$"83838281 8284878A 8C8C8A86 83818185 898E9296 98989897 9490897F 72625141"
		$"3632343C 4650575C 61656B73 797E807E 7A76716B 66626163 69727B83 888B8E91"
		$"979FA8B0 B3B2ADA3 9B938F8F 92979998 958F8A86 85878A8E 91929290 8D8B8A8A"
		$"8C8F9294 938F877C 726A6563 64666868 67656569 707B8690 95979389 7F746C67"
		$"6566686A 6A6A6A6A 6D71777F 878E9293 918B8783 81818182 817E7A75 7373757A"
		$"7F858889 89878584 84848688 898A8A88 87858585 84848280 7D797775 7576787A"
		$"7D808385 888A8A8A 8987837E 79736F6D 6C6C6E70 7274787D 83898F93 95969694"
		$"918D8985 83818181 807E7D7D 7E808387 8B8E8F8E 8C888581 7F7E7E7E 7E7E7F7F"
		$"80808183 84868787 87858381 8183868B 9095999D 9E9C9892 897E705E 4C3B312D"
		$"2F39424B 52565C62 6A747B81 827F7B75 706A6561 60636971 7981868A 8E9299A3"
		$"ACB5B9B7 B0A49A91 8E90959B 9D9B968E 87838387 8B909393 928F8C8A 898A8C90"
		$"93969591 887C7066 605E5F63 65656462 62666E7A 87939A9C 988E8478 6E676464"
		$"64666767 68686A6E 747D868F 95979690 8B858282 8283827F 7B75716F 71757B81"
		$"868A8B8A 88858484 86888B8D 8D8B8987 85858484 83817E7A 77757575 76787B7F"
		$"8286898B 8D8D8C8A 86827D77 736F6D6B 6B6B6D70 74787E84 898F9396 97969490"
		$"8C888482 807F7E7D 7C7C7C7D 8084888D 91939290 8C888583 81818182 83838383"
		$"83838486 888A8B89 8783807F 81868C92 989C9E9E 9D99938B 7D6B5741 3026242A"
		$"333F484E 5459616C 76808585 837E7870 69615C5A 5E66707A 82888C8E 939CA7B4"
		$"BCC0BCB1 A5978F8E 91989D9F 9C948D85 8282858B 8F929393 918D8B89 8A8C9094"
		$"9798948A 7E70655D 5A5D6064 64626060 656F7B89 959D9F99 8F817468 615D5D60"
		$"63666868 696A6E76 808C969D 9F9B948A 84808082 8383807A 75717072 777F858B"
		$"8D8D8B87 85838486 888B8D8D 8C8A8886 8482817F 7D7A7775 73727274 767A7E82"
		$"878B8D8E 8E8C8985 7F78726D 6A68696B 6D6F7377 7D838A90 94969795 928E8A87"
		$"84828181 81818080 80808185 898D9092 918F8B87 837F7C7C 7D7F8284 85858482"
		$"82848688 8A8A8885 83838589 8F969CA0 A09E988E 806C5741 31292932 3B434848"
		$"4A4E5766 75838A8A 867F7870 69635F5F 62687078 7F838688 8E98A4B2 BCC1BFB5"
		$"A99B9492 959DA2A4 A1978D84 80808389 8F939696 94908D8B 8A8A8D91 9597958D"
		$"8272655B 575A5F65 67676563 656B7583 8F999C99 9185796D 645E5C5E 6165696B"
		$"6D6D6E72 79838D97 9B9B978F 89838081 8283827E 7A757272 74787E84 888A8B89"
		$"88868585 87898C8E 8E8D8B89 86827F7B 79777575 75757575 76787B7F 83888C8E"
		$"8F8D8984 7E76706B 69696A6D 70737679 7D82888E 94999B9A 97928D87 84828284"
		$"85868582 7F7D7D7F 82888C90 908E8A84 807C7B7B 7D7F8284 84848381 81818488"
		$"8B8D8C8A 8785868A 91999FA3 A3A19B91 826F5A44 342A2932 3C464C4E 4F515865"
		$"73818A8C 89817971 69635F5F 62686F77 7D818486 8A929DAB B4BAB9B2 A89B9390"
		$"929AA0A5 A49C9389 82808186 8B8F9294 9492908E 8C8B8C8F 92969692 88796B5D"
		$"5554575F 64686866 676B737F 8A969C9C 968B7F71 665E5B5C 60666C70 72727273"
		$"78808A94 9B9D9991 89817E7E 80838380 7D797674 75797D81 85898A8A 89878686"
		$"87898B8D 8F8F8F8E 8B87827C 78757474 74757676 7777797D 81878B8E 8F8D8983"
		$"7D756E68 6565666A 6D707377 7C82888F 95999B9B 9995908B 87858486 888A8A88"
		$"85818080 83888C8E 8F8D8A86 817D7A7A 7B7E8185 86858380 7E7E8085 8A8E8F8D"
		$"8B89898D 929A9FA1 9F978973 5B413028 2B394652 5754504C 505C6C7E 898E8D87"
		$"7E736A62 5E606670 78808381 807E818B 98AAB6BD BBB1A496 8E8D929C A5ABAAA2"
		$"978A817D 7E848B93 97999793 8F8B8989 8B8F9294 928C8172 64585355 5B63686A"
		$"69656468 707E8A95 9A989084 786C645F 5D5F6268 6C707272 72747880 88909596"
		$"938D8781 7F808284 84817E7A 77767779 7D818486 87878685 85858789 8B8D8F8F"
		$"8E8C8A86 827D7975 73737476 77787979 7B7E8286 8A8C8C8A 86807A73 6D686666"
		$"676B6F73 777A7D81 868C9195 97979693 8F8A8785 85878A8D 8E8D8A85 82818387"
		$"8A8D8E8C 8985817F 7D7D7D7F 81848687 8683817F 8185898F 9292908E 8E909296"
		$"97958E80 6C543F2E 29303C4D 585E5E58 56585F6D 78818584 807A726A 64606268"
		$"717C8488 88838181 8896A4B2 B7B5AC9E 938D8D95 9DA6AAA8 A2988E86 8282858B"
		$"91979A9A 97928E8A 898B8E92 9290887A 6C5D5553 5660676D 6E6C6A69 6D77828E"
		$"94969288 7C6E635D 5B5D6268 6E727575 76767A80 878F9395 928C8680 7D7E8185"
		$"8787847F 7B797879 7C808386 87878684 83838486 898D8F90 908E8B87 827D7975"
		$"73737476 77797A7A 7C7E8286 8A8D8D8A 857D7670 6D6B6A6A 6A6C6E72 767C8187"
		$"8B8D8F91 92949492 908C8886 8587898D 8F8F8D89 86838283 85878888 87858482"
		$"817F7F7F 8185898C 8C8A8682 81858A92 96989793 908E8D8D 887E6E58 45352F33"
		$"3D4B545A 5C5A5B5F 656F767B 7D7B7876 726E6A66 666A717B 82888987 85848993"
		$"9EAAB0B0 AB9F968E 8D9299A1 A4A4A09A 938D8886 86888D93 989C9C9A 958F8A87"
		$"888C8F91 8B7E6F5D 514C4F59 636D7171 6F6B6D73 7B868E93 928B8173 665B5656"
		$"5A646D76 7B7B7976 777B838D 94989690 88807B7A 7C818588 8886837F 7C7A7A7C"
		$"7F838688 89878482 81818488 8D919292 8F8A847C 77737272 74777979 7979797B"
		$"7E83888C 8D8B867E 76706C6A 6A6C6D6F 7173767B 81878C90 92929292 9292908E"
		$"8B878686 878A8D8F 8F8D8B87 85848484 84858585 84848381 81818285 898D9090"
		$"8E8B8A8A 8C909496 96928D87 7E746757 493C383C 434F575C 5D5B5B5E 646E767C"
		$"7E7C7874 6F6A6767 6A70777F 84868785 84868B93 9CA5AAAB A8A09993 9092969C"
		$"A0A19E97 90898584 868A8F94 989A9B9B 9995918D 8C8C8D8D 887E715F 534B4B53"
		$"5D697073 74727377 7D858B8F 8E898074 685D5654 575F6973 7A7D7D7B 7A7B8088"
		$"8F959793 8D857F7B 7B7E8185 87878582 7F7B7878 7A7E8286 88898987 86848587"
		$"8A8E9092 908B847A 736D6B6E 73797C7E 7E7C7B7B 7E82878B 8B88837B 746D6967"
		$"686A6E72 767A7E83 888C9092 93939291 90908E8C 89868585 868A8D8F 8F8D8A87"
		$"85838282 81818182 83838383 8383858A 8F939595 93908F8F 90939494 90887D6F"
		$"5F4D4037 373F4854 5B5F5F5C 5C5F6670 777C7E7D 7A76716C 69676A70 79838A8C"
		$"8B878484 88929BA5 AAAAA69E 97919092 969DA1A3 A0999187 817E8086 8E969B9D"
		$"9D9A9793 908F8F91 918F887C 6E5D5149 49515A66 6E727473 75797E86 8B8D8C87"
		$"7F73695F 59565860 69737A7F 817F7E7E 80868C92 9493908A 85817F81 8386888A"
		$"8987837D 7977777B 7F85888A 8A898887 888A8C8F 9193918D 857B726A 686A6E74"
		$"797D7E7D 7D7D7F83 86888886 817B746E 68646365 696F767E 84888C8F 91929393"
		$"94949391 8D898581 7F808389 8D8F8F8D 8A868382 82848585 85848383 83838485"
		$"888C9095 98989693 908E8D8D 8B867B69 553F312B 303E4B59 6163625D 5C60666F"
		$"767A7B79 7773706C 6B6D727A 838D9293 918B8684 878F99A3 A9A9A59B 94909096"
		$"9CA4A7A7 A2988F85 7F7C7E84 8B959CA0 A09C9892 8E8E8F92 92908676 65534743"
		$"46525D67 6E707273 777D8389 8C8C8880 766A6159 55555A64 6D777E82 83818080"
		$"8286898D 8E8C8985 83818284 87898B8B 8A87837D 78757578 7D838789 89888888"
		$"898C8F92 93938F87 7E736B65 65696E74 797B7D7E 7F808284 86878682 7E78736D"
		$"6967676B 70777E85 8B909496 96969594 94949391 8E88837F 7E82878E 9395938D"
		$"87838182 83858585 84828284 86888A8D 9195989B 9C9A9690 8982796E 5E4A3828"
		$"252D3A4E 5C666965 63656A72 777B7A76 716D6A68 67696E76 7F8A9297 9896928E"
		$"8C8E9298 9DA1A09C 97918F8F 939CA4AA ABA79E90 857C7A7E 858F979D A1A19E9A"
		$"96929090 91938E82 715A483C 3B455364 6F737471 72767E88 8E908D85 7B6E6258"
		$"5250545E 69757F86 89878583 84888B8D 8E8C8985 817F7F81 848A8E92 92908A82"
		$"7B757375 7A818688 89878685 878B9096 999B9890 84766A61 5F636870 75787A7A"
		$"7B7D7F83 85868581 7D77726C 67646569 6F777F88 8F939799 9A999896 94949391"
		$"8E89847F 7E81878F 95989792 8D878383 83838382 82828385 888B8E90 94989C9E"
		$"9C968E82 74645240 3126252D 394A5862 66646468 6E777D80 7E766F69 65636468"
		$"6F77828E 979B9B97 95939498 9B9D9D9B 9793908F 9093979B A0A4A5A3 9D938A80"
		$"7C7E848E 969C9F9F 9E9C9997 9593918E 877B6C58 473A3840 4D5F6B73 76757679"
		$"7F878B8D 89817669 5E544E4C 515B6674 7F878B8B 8A888789 8A8C8C8A 88848280"
		$"80818488 8E949796 91877E75 72747981 878B8B89 8785878B 8F95999B 99918677"
		$"6A5F5B5D 636C7378 7B7B7B7D 7F818181 7F7B7773 6F6B6765 6569707A 858F979B"
		$"9D9D9C9A 97959392 91918F8C 89858587 8C949A9E 9D99938C 87848383 84848586"
		$"888B8E92 95979898 95908678 6347311F 191D293C 4B575E60 62656B75 7C828380"
		$"7B736C64 605E6169 7585929C A1A19F99 97979AA0 A4A6A49C 948B8787 8D979FA6"
		$"A8A49E94 8B837F7F 838B929A 9FA1A19E 9B979696 9799958C 7C644F3D 34353E4E"
		$"5C686F71 74777C83 888B8B87 7F74685A 5049484D 57667480 888C8D8B 8A8A8C8E"
		$"90908E8A 86817E7E 80848A90 95979690 887E7875 767B8187 8B8C8B89 88888B8F"
		$"94989996 8F837668 605C5D63 6A707578 7A7A7B7B 7C7C7B7B 7A787571 6D696768"
		$"6D757F89 92989C9E 9E9C9995 93939394 9494928E 8B8A8D93 999FA19F 9A928C86"
		$"84848485 8687898B 8E929494 918B8173 5F46301C 1418253A 4B575D5B 5C5F6877"
		$"828A8A84 7C746D67 6463666D 7887949E A4A4A19B 9A9C9FA5 A9ABA8A2 9A908A87"
		$"8A929BA6 ACADA698 8A7C7678 7F8D98A0 A4A3A09C 98959496 989A978F 80685039"
		$"2E2F394B 5A666C6C 6D707680 86898882 7A6F6459 5049484C 56657482 8A8E8E8C"
		$"8A8A8B8D 8F91918F 8D898581 81848A92 999D9B93 887C7472 757D848B 8F8F8D8B"
		$"8A8B8D91 95999997 8F837566 5D595A61 686E7272 72717171 73767A7E 8081807C"
		$"78747170 7379818A 92989B9B 9A969390 90929599 9A989590 8D8D8F95 999D9D99"
		$"948C8784 83838485 86888989 87817666 523A291D 1B243040 4B535654 565A626E"
		$"78808484 817C7772 6F6F7279 838F99A0 A3A2A09C 9A9CA0A6 AAADACA6 9F979290"
		$"9094989D A0A19E97 8E827A77 7B85909C A4A7A59F 99939090 90918E86 7A685747"
		$"3E3C424E 5963696D 70727477 79797875 70686057 504B4B51 5A687581 898D8D8B"
		$"89888889 8B8D8F90 908E8C8A 89898C91 95999892 897D7673 757D858D 9191908D"
		$"8B8B8C8F 92959592 8C82776B 625C5B5D 62686D71 7577797A 7B7D7F82 84858482"
		$"7E7A7674 74777D85 8D969DA1 A19F9B96 9494969A 9C9D9C98 94919194 979B9C9C"
		$"9995918D 88847F79 736B5F50 3F2B1C14 16233346 535B5D59 595D636C 72767674"
		$"72717274 767A8089 94A0A9AF B0ACA69F 9C9C9FA5 AAAEAEAC A79F9994 93979BA1"
		$"A29E9688 7D75747A 84909BA3 A8AAA9A5 9F97918B 8886827A 6E5C4C3E 3B414D5F"
		$"6C757979 797A7B7C 7B777169 625C5650 4C4A4C52 5C6A7783 8A8D8E8C 8B898787"
		$"888B8E92 9392908C 8B8D9096 98989287 7D736E70 76808991 96989896 9492908E"
		$"8D8C8985 7D71665A 5351545C 65707A82 898D8E8C 8A888684 8383817F 7C797777"
		$"7A808791 99A1A6A8 A6A29D97 93939498 9A9B9B99 97969799 9B9D9C9A 958F867B"
		$"6A523616 0501040D 203E525C 5F595656 5C667078 7A78736D 67636264 6B798AA0"
		$"B1BDC1BE B8AEA7A3 A3A9AFB5 B7B5B0A8 A0989597 9BA2A6A6 A1958777 6C676A75"
		$"8496A5AF B3B0AAA0 97908C8A 86807462 4F3D3332 39475666 727C848A 8D8D8A83"
		$"7A70665D 554D4844 4448505D 6A788289 8D8D8C8A 88868688 8B8F9295 96969595"
		$"96999B9D 9A92887A 716B6C72 7B879096 9A9B9B99 9795928E 8880766B 6157514F"
		$"50555D69 7684909A 9F9D9991 8B858282 82828180 7F7F8082 868C949C A4AAADAB"
		$"A59B938C 8A8D9299 9EA0A2A2 A19F9D9A 948A7A66 4A260F05 0000081A 304A595F"
		$"6161646C 72777773 6D676361 6161656B 798DA1B5 C3CAC9C1 B8AEA9A9 ABB1B5B8"
		$"B8B4AFA7 A29E9FA3 A8AEACA2 927C6D63 636D7A8A 97A3ABAF AFADA79D 9389827C"
		$"72655440 31282B39 4A5E6D78 81878D93 94928A7D 6F5F534B 45414144 4B556270"
		$"7C858A8C 8B898682 807E8085 8B919496 9695979B A1A7A6A0 9482746A 686E7884"
		$"8E949797 97959189 827C797A 7B7B766C 61565357 6171808E 989EA0A0 9D97918B"
		$"8785868A 8C8D8D8B 89898B91 979FA5A9 ABA9A49E 98928F91 949A9D9F 9C968975"
		$"59371C09 00000715 24343B39 38384155 697F8A8A 8375695F 5856575C 67798EA7"
		$"BAC6C9C3 BBB2B0B4 BCC6CDCF CBC1B5A8 9F9B9CA3 ABB3B5B2 A898897B 74747983"
		$"8E99A1A6 A6A29B93 8C878380 786C5B45 3529262E 39485662 6E7C8997 9E9F998C"
		$"7E6E635B 544F4A46 464B5462 6F7B8488 8A8A8A8A 89878686 86878889 8B8D9094"
		$"999E9F9C 95897F76 7375797F 8487898B 8D8F9192 92908D89 8785817C 766E6762"
		$"61656C78 84919A9F A09D9993 8E8B8A8A 8B8B8A8A 8988898B 8F959DA7 AFB5B7B5"
		$"B0A8A097 8E867967 5137200C 04060F20 2F3B4345 4646484C 5155595C 5F63666A"
		$"6C6C6D70 7884919F AAB0B1AE ABA7A8AC B2BAC2C8 CACAC6C0 B9B3AFAD ABA9A6A0"
		$"978C837B 7A7E8590 9BA5ABAF AEAAA296 8A7C7167 5D52473C 3531333B 45515C66"
		$"6F778088 8C8C8880 766B635D 5A5A5C60 656B7176 7A7C7D7D 7E7F8080 81838589"
		$"8D919597 97979798 9898938A 7E6E625C 5D677280 8A909598 9CA0A3A3 A19D9893"
		$"8F8D8983 7D777371 73798088 8E929494 928E8984 807C7B7D 80868B8F 92939495"
		$"989CA1A7 AAAAA59D 8F7C6344 280F0202 0E263D51 5A585248 4549505A 6165645E"
		$"59535050 55617083 95A5B0B4 B2ACA59F 9DA0A7B2 BDC7CBCA C5BBB3AC AAADB1B6"
		$"B6B1A89A 8F868385 8A939CA5 ACB1B3B1 AA9E9183 7870675E 513F3125 22293545"
		$"51596064 6A747D85 87847E74 6B645F5C 5957585C 65717D87 8D8D8C88 85838486"
		$"8A8E9296 98989591 8D898888 86847F77 716B696C 737D8894 A0AAB1B5 B5B1AAA0"
		$"99938F8E 8D8B8781 7A746F6D 6E737A84 8C939694 9088817C 7B7D8084 86878888"
		$"8A8E9296 9A9C9D9B 917F623C 1E0A0207 1A3B5464 69635F5B 5D666D73 74706B64"
		$"5E585452 555E6E84 9AB0BCBF BBAFA49A 9699A0AA B3BBBDBB B5ADA6A1 A1A5A8AC"
		$"AAA49B8F 86828389 919BA5AF B6BAB9B5 ACA09488 7C706151 402F2626 2D3B4854"
		$"5C61676D 757E8384 80776C60 554B4644 48525E6E 7B878E90 908E8C8A 8A8A8C8E"
		$"92969999 96908A84 84888D93 938E8578 6E686973 7F8F9CA7 AEB2B3B2 AFA9A39B"
		$"94908C88 847F7971 6C68696E 76818B94 999A9892 8D898684 84848688 8B8E9193"
		$"9290887B 674B2E11 0200091B 2F475559 5B5B606C 78868C8A 84786E64 5E5A585A"
		$"606C7B8E A0B0B8BA B7B1ABA6 A5A7AAB0 B3B4B1AB A49C999B 9EA3A5A3 9D938A83"
		$"8183878D 949CA5B0 B9BFC0BA B1A39587 7767543F 2E201D23 2D3C4850 585E6874"
		$"808E9595 8D7D6D5B 4C403A3A 404C5B6D 7C899092 908C8885 85878A8E 9091908D"
		$"8B8A8D93 9BA3A8A9 A59B9186 7F7B7B81 89939DA8 B1B8BBBB B6AEA499 9089837D"
		$"78726C66 63616369 727F8A94 9A9C9A96 918C8886 85868789 857D6745 270D0000"
		$"06132C51 666A665B 585C687B 8A949895 90888076 6E66656B 77899BAB B2B0A99F"
		$"99989EAA B8C8D2D6 D1C3B29F 94909197 9B9C9890 89838185 8A92999F A6B0B8BE"
		$"BEB8AD9F 9286796B 58422E1C 161B2636 424B5154 5B657180 8B918F86 79675646"
		$"3C383A43 50607181 8C949692 8B837E7D 818A949F A6A8A59E 9A989BA3 A9ADAAA0"
		$"94867D7A 7C838B94 9DA5ACB3 B7B7B3A9 9E928882 7F7F7D7A 746C6662 63697482"
		$"8F999FA0 9F9B9894 8F8B8377 654B2E0F 00000000 0D273E54 60636668 6E788189"
		$"8E8E8D89 86848280 81838892 9DA9B1B4 B1A79D93 8F9199A8 B8CAD4D6 D0C1B2A4"
		$"9B979595 91898078 7579808A 93999FA7 B1BDC5C9 C6BAA995 82705D4B 39271B17"
		$"1B273442 4C535A62 6C78838B 8E8A8174 66584C42 3E3E444E 59656F77 7E828484"
		$"8383868C 96A6B4C0 C4C1BBB1 AAA7A6A8 A59D9181 756D6C72 7A848F99 A4B0B9C0"
		$"C1BCB2A2 93847A75 7272706D 69646365 6C798899 A6AFB3B3 B0AAA298 8872532B"
		$"11050000 00010F2A 3C444B52 5E6F7F8D 95979693 92949492 8F8A8889 8F99A2AA"
		$"ABA59E94 8E8C8E95 A0AEBBC9 D1D3CFC5 BBAFA69E 948A8078 74747982 8B939BA1"
		$"AAB6C0C9 CCCAC1B3 A2907B63 4A301B0D 09111D2D 3A42494D 535C6773 7B7F7D76"
		$"6D61564C 4745464C 545E666C 71747678 7A7C8189 95A5B5C7 D3DBDDD9 D3CBC5BF"
		$"B7AFA599 8F878383 868A8F93 979DA4AC B1B5B3AB A0928579 716F6E70 70706F6E"
		$"7074797F 83858075 634B2F0F 00000000 050F1F35 43494D4F 545D6874 7D828482"
		$"80808287 8F9AA8BA CADAE3E5 DFD1C1AF A1999699 9FA9B0B4 B5B1ACA8 A6A8AAAC"
		$"A89F9488 807C7D84 8B939AA0 A8B1B8BD BBB1A492 81706050 3F2D2119 1A263446"
		$"52595C5A 5B5D626A 6D6B6458 4C403B3B 3F48535F 6A757E85 8A8C8C88 837F7F83"
		$"8C9BACBE CAD1D3D0 CCC8C4C0 B9AFA395 8C87878B 8F95999B 9EA2A7AD AFAEA79B"
		$"8E807771 7072767B 7F838688 898A8986 7C6B5437 1E0A0000 0000040E 1D31424E"
		$"59626A72 787C8082 82807E7C 7A7A7E86 93A3B3C2 CCD2D2CC C3B8AFA8 A5A5A7AD"
		$"B1B4B4B1 ADA8A5A4 A3A19C94 8C848181 848B939C A5AEB7C0 C6C8C4BA AB99846C"
		$"533B2717 11131E30 4355626A 70747779 78746E64 57493C31 2A26272D 37455567"
		$"76838C90 918F8D8D 9097A1AD B7C1C7C9 C7C3BEB9 B5B1AAA0 978D8785 868C929A"
		$"9FA3A6A8 AAACACAA A49A9084 7B736E6E 70747A81 89929797 8E7A5F3D 210B0000"
		$"00000510 1A242E38 45556474 808A8F90 8F8D8B89 89898D95 9FACB8C3 CACDCAC0"
		$"B6ACA5A1 A0A0A0A2 A3A4A5A7 A7A6A5A3 A19E9C9A 9896938F 8E8E9197 9EA6AEB4"
		$"B7B7B3AC A1927E66 4E362519 14161E2A 36424D57 606A727A 7D7D786F 655B5149"
		$"43404042 47515D6B 7A88929A 9D9D9D9D A0A6ACB4 BBC1C6CA CAC8C3BB B2A8A19B"
		$"97959595 9696989A 9C9EA0A2 A3A3A19D 98908981 7B767270 6C675E52 4333251B"
		$"130D0907 080E1725 313D454B 50545A60 676D7377 7A7D8288 8F98A1AA B2BAC0C6"
		$"CBCFD0D0 CDC9C3BC B5AEA9A5 A3A3A2A0 9D999490 8E8E8E8F 9193979B 9FA3A4A4"
		$"A29E9D9D 9E9E9B95 89796754 44362C24 201E212B 38495865 6E727577 797C7D7C"
		$"78716A63 5F5F6268 6E767D83 878B8D8F 8F8D8C8A 8A8C8E90 93989DA1 A4A4A19D"
		$"98938F8B 88868585 868A8E94 989C9EA0 A1A1A1A1 9F9B9793 908E8F91 93938B7B"
		$"674E3620 0F050000 040C1B32 44525B5F 656B7480 8A929696 938F8B89 898B8E94"
		$"9AA2AAB2 B9BDBDB9 B5B1AFAD AAA7A4A0 9C989591 8D898684 84858687 898D9298"
		$"9EA5A9A9 A8A5A4A4 A6A8A8A4 9988745E 48342315 0E0C121F 2F425260 6A72797F"
		$"84888988 847C746B 635D5A5B 5F666E78 81899197 9C9E9F9F 9FA1A3A5 A8ACAFB1"
		$"B2B0ACA7 A29E9A96 918D8B8B 8D9299A2 A8ABACAA A7A5A29E 9B999590 87796449"
		$"2D0F0000 00000000 0207162F 45576267 6B6D7177 7D838585 837F7F81 868F99A5"
		$"B0BAC4CE D5D9D9D7 D4D0CCC8 C6C5C4C2 BFBBB4AC A2968B83 7D7A7876 76777C85"
		$"8F9AA2A7 AAAAAAAB ABA9A295 826A523A 291E1715 1518212F 41576979 83878784"
		$"817E7A74 6D645C54 4F4D5056 5C646A6E 7379808A 92999EA0 A09FA0A3 A7ADB0B1"
		$"AFABA6A1 9D9B9896 94929293 98A0A7AE B2B4B4B3 B4B6B8BA B8B2A89A 866B4D2C"
		$"15070000 00000000 0A1E344C 5E696F71 7273767A 7C7D7C7A 797B7F85 8B92989C"
		$"A3ABB4BE C3C4C2BE BAB8B8BB BDBDBBB7 B1A9A096 8D857E78 74747579 7E858F9B"
		$"A6AFB5B8 BBBDBEC0 BEB8AC99 836A5646 3A322C28 272A3341 50606B71 7472706E"
		$"6A645C53 4B454243 464C525A 61697076 7C848A8E 93999FA5 A9ADB0B4 B7BABCBC"
		$"BAB7B3AF ACAAA9A9 A8A7A7A9 ACB1B5B9 BAB8B6B2 AEA9A197 8670573B 23110602"
		$"00000000 00000D29 41566367 6A6C6E70 72737372 7272767D 869099A1 AAB3BBC3"
		$"C9CED1D1 CFCBC7C3 BFBDBBBA B8B4AEA6 9D958F8B 89898886 85868A92 9AA2AAB0"
		$"B3B5B5B5 B1A99C8A 78685B52 4B463F35 2C242226 2E3A4754 5D636563 605C5958"
		$"57565450 4F4F5157 5D656C74 7C848D97 9FA5AAAC ADAEB1B5 B9BDBFBD BAB6B2B0"
		$"AFB1B2B2 B1AFAFAF AFAFADA9 A6A4A5A9 ACAEA99E 8B705539 220E0301 00000000"
		$"00000A1E 3244525A 6168717B 82888A88 85828284 8A929AA2 AAB3BDC7 CFD5D8D8"
		$"D5CFCAC5 C1BFBDBB B7B1AAA2 9B95908D 8A888684 868A9099 A1A8ACAE B0B4B6B7"
		$"B2A89985 725E4B3B 2D211812 0E0C101A 28384857 636B6F71 72726F69 625B5652"
		$"535A636D 767E8386 8B9199A4 ADB3B7B8 B9BABCBF C1C3C1BD B7B1ADAB ABABABAB"
		$"AAA9A9AB ADB1B3B3 B1ADA9A3 9E988E7F 6B52381C 0B030000 00000000 0103122D"
		$"43535F67 70787F87 8B8D8D8D 8D8D8E90 939599A1 AAB4BFCB D4DADEDF DFDDD9D5"
		$"D1CDC9C6 C1BBB2A7 9C928A84 7F7A7777 7A828B96 9FA5A7A7 A7A7A7A7 A39D9385"
		$"76665545 35251910 0D111A2A 3B4D5B66 6E72767A 7E81817F 7A726C66 62606062"
		$"64666B71 7880888E 9396999C A0A4A9AF B2B3B2AF ADACABAA A9A7A6A4 A5A7ABAF"
		$"B3B8BBBC BAB5AC9E 8B72573B 220C0100 00000000 02061327 39475561 6D7B8791"
		$"989B9D9D 9E9E9E9E 9D9C9D9F A3A8ACB0 B2B3B4B5 B6B8B9B9 B9B9BBBD BDBBB4AA"
		$"9F948B83 7F7D7D7D 7E828890 989FA4A8 ABADACA9 A3998B7A 6957473A 312B2828"
		$"2B323B47 525C646B 70737678 7978756F 69646160 61636668 696A6D72 787E8488"
		$"8B8E9296 9CA2A8AC B1B5B7B8 B8B6B3AF ABA8A7A7 AAAEB3BA C1C7CBCD CBC7BDAF"
		$"997C5D3C 210B0000 00000000 040D1D35 46525D67 717C8387 8783817F 7F818386"
		$"898B8F95 9EAAB4BC C0C0BFBE BFC1C4C6 C8CACBCB C9C5BFB5 ABA1978F 88827E7A"
		$"7A7D848E 98A2ACB4 BAC0C2C0 B9AD9C87 725D4A39 2A1E1510 11192433 414E5A64"
		$"6B717475 74706C68 65656668 69696968 686A6D71 7477797B 7D81868C 9298A0A8"
		$"AFB6BBBE BFBFBDB9 B4AEAAAA ADB3BAC0 C5C9CAC8 C2B7A997 8065492D 17070000"
		$"00000000 06121F2B 3844515D 6A778189 8E90918F 8C8A8A8C 919AA4AE B7BDC0C0"
		$"BEB9B6B4 B4B4B5B5 B6B7B8B8 B7B3AFAB A6A09B95 8F8A8785 868A8F97 9FA7AEB2"
		$"B3B1AAA0 94867769 5B4E4238 2F292627 2C343D47 50596065 68686969 69686766"
		$"676B6F73 74747372 72737577 78787879 7C818890 99A3ABB3 B9BDC0C0 BFBDBDBD"
		$"BFC1C4C6 C7C7C5C1 BBB3A694 816C5640 270D0000 00000208 162C3C46 4D515861"
		$"6B757B7F 807E7E80 81828282 83878E9A A6B2B9BD BDBBB9B7 B5B1ADA9 A8AAADB2"
		$"B6BABBB9 B7B4AFA9 A096908E 9199A0A7 ABACADAE AEAEAAA4 998A7B6B 5E53493F"
		$"33261B13 12182434 424E565C 60626363 6362615F 5E606368 6D727679 7C7E8185"
		$"888A8B8B 8D90959B A2AAB2B8 BDC1C6CA CDCFCECC CAC8C5C2 C1C1BFBC B5A99B89"
		$"77655038 210B0000 00000717 27373F41 43464C55 5D63686A 6C707478 7A7A7874"
		$"74798391 9EAAB2B8 BCBEBEBD BCBABABA BABABBBB BCBCBCBD BEBEBBB6 B1ABA8AA"
		$"ADB3B7B9 BABABABA B8B4ADA3 9687796B 5D4F3E2B 19080000 05111E2E 3A444B51"
		$"575D5F5F 5D595654 55585C62 676B7074 7A828A92 97999B9D 9FA1A3A6 A9ADB1B7"
		$"BDC3C7CA CDCFD2D4 D5D4D2CE CAC6C2BE B7AD9C84 694B2E12 03010000 0000091C"
		$"2B353B3F 42454B53 5C656C70 73747575 74716F6F 727A838F 99A2A9AD B2B8BDC2"
		$"C7CBCFD4 D9DEE1E1 DED9D2C9 C0B6ADA3 9B959395 999FA3A5 A6A7A7A5 A29C9488"
		$"79675440 2B170903 00000511 223A4F63 737F8688 89898785 817D7874 706D6B6A"
		$"6C707478 7C808488 8B8D9093 969A9EA2 A7AEB7C1 C9CFD1D1 CFCBC6C2 BDB8B2AA"
		$"A39F9B97 8F816C52 35170602 00000000 030A1117 1D232C38 46586878 85909AA4"
		$"AAACAAA2 9A928E8E 929AA2AA AEAEB1B5 BAC1C7CB CDCECFD0 D2D6D9DD DDDBD7D1"
		$"CAC4BCB4 ACA59D95 8D85807E 7C7A756F 675F574F 463A2B1B 0E040000 03091424"
		$"33414D58 61686F75 7C848A8F 91908F8D 8D8D8F91 9395989D A2A8AAAA A8A5A3A3"
		$"A5ABB0B5 BABEC3C7 CCD2D4D2 CFCBC7C4 C0BAB3AD A49A876C 491E0602 00000000"
		$"00000612 1F2B3742 4D58636E 777E848A 8F939491 8C86817D 7E848F9D ACBAC5CB"
		$"D2DAE0E4 E4E2DED9 D2C8C1BD BBBDBEBE BDB9B2A8 A0989597 9A9FA3A6 AAAEB2B5"
		$"B4AEA394 826D563C 230B0000 00000000 091D344E 6374818B 939A9D9D 9A948C83"
		$"7A726C68 6666676A 727E8C9B A5ABAEAE AEAFB2B6 BBBFC1BF BEBCBDBF C0C0BDB7"
		$"B1A9A4A0 9A94856D 4D250C04 00000000 07162636 41474A4A 4947484C 535D6975"
		$"8393A1AB AFADABA7 A6A8ABAF B2B4B3B0 AFAFB2B6 BABCBFC1 C5CBD0D6 D7D4CDC1"
		$"B5AAA29C 99979797 999B9EA2 A5A7A6A3 9C918372 604C361E 0D040000 00000716"
		$"283C4F5F 6E7C868D 90908E8B 89898A8C 8D8C8B89 898A8C8F 93979896 94908E8E"
		$"8F92979D A4AEB8C3 CCD2D3D0 CBC5BFB9 B4AFABA7 A29C9182 6C4E2F0F 00000000"
		$"0000040E 161D2327 2F394759 6A7A8894 A0ABB1B1 AEA69E96 918F939D A9B7C0C6"
		$"C8C7C5C1 BEBCBABA B8B5B3B1 B2B4B6B7 B6B2ADA6 A2A1A3A7 A9AAAAA8 A8AAACAE"
		$"ADA89E8E 7D6B5A49 37231206 00000000 0A1E3246 56636F7A 848D9396 97969490"
		$"8C86827F 7B777472 757B8189 8F939595 9494979D A3ABB3BC C4CACECF CECBC6BE"
		$"B7B2AEAC A2927958 36120000 00000000 0C243848 52565756 56585A5C 61697381"
		$"8D979EA2 A4A5A5A3 A4A8ADB4 B9BCBFC1 C0BDB7AE A9A7A6A6 A6A6A7AB AEB0B1B1"
		$"AEA8A19A 989CA2AC B3B8BBBD BDBDBAB4 AA9C8C79 634A3117 07020000 0000081A"
		$"304B6276 848E9498 99989795 9494918C 88848180 80808183 8587898B 8D8F8E8B"
		$"8A8B9099 A4B0BCC6 CCCECDC9 C4BCB6B2 ADA79D8F 76512F0F 00000000 02071732"
		$"454F575B 5D5F5E5A 59595E66 738595A2 A9A9A7A1 9FA0A3A8 ADB3B6B6 B7BABEC2"
		$"C3BFBAB2 AAA4A2A4 AAB2B8BC BBB6B0A8 A29E9DA0 A4A8ACB0 B6BCBFBF BAB0A18F"
		$"7B675139 220B0000 00000000 0B233A52 64727E88 91999EA1 A2A09E9D 9B989490"
		$"8B868381 82888D92 93918D87 8483878F 99A5B0BB C3C7CACC CBC8C3BC B5AEA69E"
		$"8C704C20 07020000 00000C24 394B575B 5E5F5F5D 5B585656 5A646F7D 89939A9C"
		$"A0A4A7AB B0B6BBBF C3C5C6C6 C5C1BBB4 AEAAA7A5 A8B0B7BF C0BBB5AD A6A09D9D"
		$"A1A7AEB4 BBC1C5C9 C8C2B9AB 99846D54 381A0802 00000000 030B1D3A 53677885"
		$"8F95989A 99969493 9393918D 8A878584 85898D93 97999996 938F8D8C 8F959EAA"
		$"B5C1C9CD CFCECBC5 BEB8B1AA 9E8C6D41 210B0000 00000613 2743555D 62656767"
		$"635B544D 4C505B6D 7D8B9396 999CA2AB B4BDC3C5 C6C4C4C5 C5C3BEB7 B0A9A4A0"
		$"A2A8AFB8 BCBCB7AF A69E9896 979BA0A6 AEB7C1CB D1D3CFC5 B5A18A70 54381F0A"
		$"00000000 00000D28 435D717D 868C9195 96949391 90908F8E 8C898682 8081868F"
		$"979EA19F 9B95918D 8D9198A3 AFBDC6CA CCCDCDCB C6BEB7AF A69A815D 38120000"
		$"00000206 1A3E5765 6C6E6E6C 6760584E 49494F5D 6B79838B 91969DA5 AFB9C5D1"
		$"D8D9D7D2 CCC4BAB0 A8A3A0A0 9F9EA1A7 ABABA8A0 9B99999D A2A8ADB3 B7BBC0C7"
		$"CDD3D3CD C1AF9B85 6B4E3010 00000000 0000091B 34546D7F 8B919597 9796938E"
		$"8B898683 807C7876 75777D88 939DA4A6 A6A29C95 908E929C A7B3BDC4 C9CCCBC5"
		$"BCB1AAA6 A0988260 3B130000 0000030B 1F405868 747C8080 7B6F6151 47444957"
		$"65758088 9098A2AE BAC8D3DB DCD6CEC6 BDB4ABA3 9D999490 8D8B8F9A A4ACAEAC"
		$"A9A5A2A0 A1A5ABB3 BCC4CCD4 DADEDCD6 CAB8A38B 72583B1B 08020000 00000511"
		$"284C6676 8084888E 9193918B 87858484 84848381 80808590 9CA8AFB1 AEA79F95"
		$"8F8C909B A7B5BEC2 C2BEB9B3 B1B3B1AB 936B4115 00000000 00011438 54697C8C"
		$"9AA4A59F 92807268 63636363 625E6169 778B9FB3 C5D5DEE0 DED8CFC3 B7A99E94"
		$"8B837F7D 818A939D A2A4A4A2 A2A2A4A8 ABADB1B7 C2D0DCE5 E7E3D9C9 B7A28C74"
		$"57341A08 00000000 00001236 56748997 A1A9AFB3 B1A79D91 88827E7D 7A76716C"
		$"6E788698 A7B2B6B2 ACA29B96 979EA5AC B0B2B1AF ACA8A7A7 A39C8763 3C140000"
		$"00000000 133B5B73 8797A4AF B2AC9F8D 7C6C6059 54514D49 4B536582 A0BFD8EC"
		$"F7FAF8F0 E4D6C5B2 A08F8072 6A666B77 85939DA1 A4A6A9AF B5BBBDBC BCBCC2CD"
		$"D7E1E3DE D3C2AE98 80684C2C 15070000 00000000 0F2D4D6F 899CABB5 BABCB7AB"
		$"9E90847B 75716E6A 68686F7D 8FA7B8C3 C7C3BCB2 ACA8A8AA ABABABAB A9A7A39E"
		$"9B9B978F 78523010 00000000 0001153B 596F8395 A4B0B3AD A1908175 6C66615B"
		$"5551535D 6E869EB8 CCDCE5E7 E5E0D8CD C0B1A292 8477706E 7179828A 9197A0AA"
		$"B5BFC6CC CFCFD0D2 D6DCE0E2 DFD9CDBC A78F765C 3F200C04 00000000 00000D28"
		$"45647C8C 9AA6AFB5 B3ABA093 89817C79 75716E6E 758394A8 B5BBBBB3 ACA5A3A5"
		$"ABB3B7B7 B6B2ADA6 A2A09F9F 988C734D 2C0E0000 00000105 1A40607A 8E9CA8B0"
		$"B1ACA292 83756C68 635D5753 545B697D 94AEC3D3 DDDFDFDD D7CEC1B0 9F8E7F71"
		$"6A696E7A 858F969C A2AAB1B9 C0C6CACC CDCFD3DA DFE2E0D9 CDBBA68E 755B4022"
		$"0E040000 00000000 0C264468 8396A6B2 BABEBCB3 A89C9085 7D797571 70707680"
		$"8E9FACB4 B5B1ACA6 A5A9B0B8 BDBEBCB6 AEA49D99 948E7E66 43160000 00000000"
		$"0A1E3755 6F8599AC B9C1C0B6 AA9A8C80 75695D4F 47434856 6A849CB4 C5D0D9DF"
		$"E2E3DDD1 C1AE9A85 73656064 6D7B868E 96A0A9B3 BDC7CFD5 D8D9DDE4 EBF1F2EE"
		$"E4D5C1A8 8D705233 1B090000 00000000 00011438 5873899B A9B4BABA B7AFA599"
		$"8E867E78 726D6D72 7D8F9FAD B6B8B6B0 ADADB1BB C3C9CBC7 C0B4A99F 97908474"
		$"572F1406 00000000 030A2044 627A8FA1 B0BDC3C1 BAADA195 8A827666 55433731"
		$"384A6280 9DB9CFDF E9EFEDE5 D7C4B19D 8C807978 7B81888E 9295989B 9FA5ACB4"
		$"BBC1C9D1 DAE4EBF0 EFE8DBC8 AF91714F 2F0F0000 00000000 00000A1E 3858738B"
		$"A0B4C3CD D0CBC2B4 A89C928A 827A726C 6C727F91 A1AFB6B7 B7B7BAC0 C8D1D5D3"
		$"CCC0B4A8 A09A9185 6B43240C 00000000 00000E2A 435A7189 A1B9C8D0 CEC4B6A6"
		$"988C7E6D 5A46372E 313F5677 96B2CADE EDF8FBF6 EAD6BEA2 89756B6C 727E868C"
		$"8F8F9195 9FADBBC9 D2D5D8DA E1EBF2F6 F5EDE0CD B79D8062 411E0903 00000000"
		$"0000040E 2446637B 90A4B4C0 C8CAC4B7 A99B8F85 7B6F665F 5F65748A A0B5C2C6"
		$"C6C3C5CD D5DFE2DF D6C6B8AA A19D978F 764C2A0E 00000000 00000C25 3D536C88"
		$"A2BBCACE CCC2B8AE A49B8E7C 6851423C 3E48576B 7F93A6B7 C8D9E4E9 E8E0D1BC"
		$"A48A7464 5B595A60 67717D8B 9CB1C5D7 E4EBF1F5 F9FDFFFF FEFDF2DC C3A68869"
		$"48250F05 00000000 00000208 19365067 7E95A7B3 BABAB7AF A59B9086 7B6F6660"
		$"636D7C8F 9FACB3B3 B3B5BCC9 D6E2E7E5 DCCBB9A6 988F857B 6440230B 00000000"
		$"00000E2C 49657E94 A9BBC5C8 C5BBB2A9 A29D958B 7C68584C 474B5464 75899BAD"
		$"BED0DCE4 E2D8C7B0 987E6A5B 54555A62 686C737D 8A9CAFC3 D2DEE6EB F2FAFFFF"
		$"FFFFF5E1 C9AD9174 55351B09 00000000 00000000 0F2F4C68 8096A8B7 C1C5C3BD"
		$"B4A89E94 8C847D75 71727882 8D99A0A4 A7A9B0BC C7D3DADC D8CCBFAF A2968879"
		$"613F220B 00000000 00000E2B 48647C90 A4B8C7D0 D1C9C0B7 B1AFA89C 8B735F4F"
		$"484A5464 748493A2 B2C4D0D8 D9D3C8B8 A5917E6D 625C5B5F 656F7984 909CAABA"
		$"C8D6E2EC F5FBFFFF FFFFF9ED D9BD9F7E 5C381C09 00000000 00000000 0D29445F"
		$"778B9BA9 B3BBBDBA B4AAA198 908A847F 7E828994 9EA8ADAE AFAFB4BE C9D5DBDD"
		$"D7C9BAAA 9C8F7A5D 3B130000 00000000 01031435 536F8CAA C3D7E1E1 DCD2C8C0"
		$"B7AD9F8B 7763564F 4D515967 758494A6 B6C6CFD1 CDC3B4A0 8B75655C 595C6064"
		$"6B737E8E 9FB2C4D4 E0E8F1FA FFFFFFFF FFFFF2D9 BC9C7A58 35110000 00000000"
		$"00000208 1A38536B 8299ABB9 C1C3C1BB B2A69B91 877F7B7C 828E99A5 ADB1B4B4"
		$"B8C0CBD9 E2E8E5DB CCB8A491 7A5F3D14 00000000 00000000 0D274361 7F9BB5CD"
		$"DDE7E9E4 DCD0C6BE B2A28D73 5F4F4644 4C5C6C7C 8A96A4B5 C1C7C8C2 B7A7947F"
		$"6F656165 696E7274 797F8B9C AFC5D7E5 EFF7FCFE FFFFFFFF F7E8D1B1 906E4C28"
		$"11050000 00000000 0000091C 3656738F A6B8C4C8 C8C4BCB1 A79E968E 86807F84"
		$"8C96A1AC B4B9BEC4 CBD3DADE DCD6CCBE AF9F8D7B 5D331607 00000000 00000612"
		$"FFFFFFFF FFFFF7E7 D4BEA790 75563511 00000000 00000000 00000819 2E465B6B"
		$"767C8187 909DAAB6 BEC2C8CE D6E0E7EC EDEBE7E3 E1E2E5E9 E9E3D2B5 86441A08"
		$"00000000 00000000 00000E2B 4C7191AB C3D8EAF8 FFFFFFFF FAF0E0CA BBB1ACAB"
		$"AAA8A6A4 A19E988E 857B6F61 5343352B 27292C30 353B434D 5B6B7B8D 9DADBFD5"
		$"E7F7FFFF FFFFFFFF F8EAD9C3 AC957B5F 3F1B0702 00000000 00000000 05112135"
		$"485A6A77 828C959E A9B5C0C9 D1D7DCE2 E6E9E9E7 E4E2E1E2 E5E9E9E5 D6BC8F4D"
		$"210B0000 00000000 00000000 0A1E3A60 82A2BED6 E9F7FFFF FFFFFAF0 E0CABCB4"
		$"B1B3B2B0 ADABA7A3 9B91877E 74685A48 382A2220 2229323D 47515E6E 8092A1AE"
		$"BED0E2F5 FFFFFFFF FFFFF8EB DAC6B19B 83674826 0F050000 00000000 00000208"
		$"152B4157 69778289 939FACB9 C2C6CBD0 D8E2E9EE EEEAE5DF DDDDE1E8 EAE6D6BA"
		$"863C1105 00000000 00000000 00000C25 446888A2 BBD3E7F7 FFFFFFFF F8EADAC6"
		$"BAB7B7BA BAB6B3AF ACA8A096 8A7D6F61 52423429 221E1D21 28334254 67798793"
		$"9EA8B8CD E1F5FFFF FFFFFFFF F7E7D6C2 AE9A8164 431D0802 00000000 00000000"
		$"040E1F37 4A58636B 747F8EA1 B1BDC5C9 CED4DCE6 ECEEEDE8 E4E2E4EA EDEFECE4"
		$"D1B48138 0F050000 00000000 00000000 0F2D4E72 8EA2B7CD E1F5FFFF FFFFF5E2"
		$"D0BFB7B7 B6B6B5B5 B5B6B4AE A79D9286 77655342 3428211E 1F232A33 4052657A"
		$"494B4F50 4C48443C 36322E2E 312E2A2A 2E363F46 536B87A3 B5BED8F5 FEFFFFFF"
		$"FFFFFFFF FFFCF9F4 F2F9FCFC FBF5EDE3 D2BBA79F 9D8D786C 6765625E 5C5B5B5D"
		$"5E5C5A54 4A3F3832 2A242222 262C3035 3C3F4147 50565F6A 6D6B6A6A 67666564"
		$"65635F5F 5F5B554F 4C4C4B47 4A4F5153 53565B63 6E727171 72737575 7676777E"
		$"868D9496 96969591 8E90908D 8A827C7B 7D848C91 94969AA0 A8AEB0B1 B2B1B1B5"
		$"BFCDD3D3 D2D0CCC9 C6BFB6B0 ACAAAAA9 A69D8E84 7E787672 68676C6E 6C696A6C"
		$"6E716E68 6B717678 77787C81 888A8986 7E7A7C83 837E7D7A 76797E80 86888488"
		$"8B8C9091 9499A1A8 AAAAACB7 CBDDEDFF FFFFFFFF FFFFFFFF FFFFFFF6 E8E0DCD9"
		$"D2C9C4BF B9B5B1AA A0979492 897C7167 666A7279 766C655E 5C5F554B 473C3634"
		$"342F231C 21282C36 35262021 1E1E211C 100A0E12 141A1D1C 18191C17 16181818"
		$"1511131A 1F242C33 36363439 3A353535 383E4245 48494B51 59606972 7C8CA1B4"
		$"BED7FFFF FFFFFFFF FFFFFFFF FFFFFFFF FFF9F4F2 EDE5D9CA BBABA19F 90746053"
		$"463E3B3A 3C3E3F42 4444474D 5154575D 62656D74 73798081 899299A0 A5A8AAB1"
		$"BDC1BAB3 ADA7A5A4 9E928170 696E747D 84848B8B 8C96948E 8A888D96 9FA3A4A4"
		$"A4A7B0B0 AAADACA7 A3A1A1A0 9C947F70 6C696663 65605A5F 68768793 9A9FA2A5"
		$"BAD9EDF8 FEFBF3ED E6DACBBD B3AA9F99 968B745C 4D403735 3A444E53 585F5E60"
		$"6C798797 9E9DA0AA AFA79FA6 ABA9ACAE A59A9899 99989582 6D686055 4B38200F"
		$"09040000 00000210 1611131C 1714140F 1212141C 212B435E 7992A5AD B3CADFE8"
		$"F0EEE4E1 DED9D4CD C7C0BBB6 A9A0A19F 9480664C 3C2E2728 2F32343A 3D454A41"
		$"34281F20 2A302C22 1B1F2F45 49423E3F 42372A2B 2D27231F 1C273D3F 38373430"
		$"3A44433F 3830281F 1D1D2026 2E38424E 554A3B31 2A272829 2624282D 323B4B52"
		$"54606B6C 73746960 5E5D5B60 6D747D8B 90929A9C 9796928F 949CA5B2 BFC6C8C9"
		$"D2E6FBFF FFFBF1E6 DBD3CBC6 C4C3C3C7 CDCFD3D5 D2C9BFBA BAB8AF9F 958C7F7D"
		$"807E7F81 7D7C8083 86837C76 6F71767A 82827B78 7264657C 8C959997 928E9296"
		$"8E847C70 71777870 68797F6A 656A6E67 69727579 7D787374 71707E81 7D7A6863"
		$"6D717885 8C939C9C 9B9E9B96 928C8071 747B7470 6E6C7583 96A7B4B9 B8B6B1A4"
		$"98999593 A0AEB6BA BABABABA B9B7B6B2 A696867E 79777F85 83828484 80848D8A"
		$"7E757272 7173797E 79706E6C 7076746E 665C5650 55605F5A 585E7280 88939293"
		$"9B9DA0A7 A6A29E99 9592969E A2A6B2B8 B8B8B8B8 B7B7B6B4 B2B0ABA3 9986787A"
		$"81888C8B 8B8D90A0 AA9A8D84 7C7A7A77 767E8178 747A746E 77888F80 6E6F736F"
		$"7B7F6E78 887C7883 82777B8B 887F796E 685F5462 746F737C 77727C83 827E7566"
		$"5754616E 6D727572 77828987 82756A6A 645D5B56 50525A64 70767470 6E686568"
		$"6C6F7582 8F9CB0BD BFCEE2E8 EBEFF0ED E6DCD4CC C9C9CDD5 D9D7D5D1 CCCBC7BE"
		$"B7B1ADAD AAA49C9A 9588817D 7A7D7F77 70737671 6866686A 6C737268 696E6D70"
		$"7985918D 878F948E 827E827E 7A7C837B 6E6E7179 827F7266 5F595354 50423F46"
		$"4B596462 5E676A60 6B6F6467 6263757A 7E868C94 9BA0A6A5 9F9C9AA0 9E999C9E"
		$"9F9C9EA0 989AA3A1 9F9E9797 98A0A8A3 A4ACADAD AEAEADB0 B6B6B1AD ACABABAA"
		$"A8A5A4A0 998F8B89 85868985 78717576 787E796C 696A6A6E 716F6965 666C6D65"
		$"5C565963 68696967 686D757A 77706D6F 76787270 75787474 75788289 8A8D8D8C"
		$"8D91928A 8A8C8A8E 93969BA3 A9ABACAC AEB5C4D4 D7D7D8D3 D2D9D4C9 C8C1B1A9"
		$"AAADABA7 A5A2A2A2 9E92807A 7A76736D 6968686E 6F6F6E70 70665F63 686B726E"
		$"5D525662 6A6D6C6A 6A6C7079 75686D6E 6866686A 6C6F7479 6A636D6D 6B6B6E6B"
		$"5C5A6059 575E5856 6470747A 7E786B66 645E6061 5A5B605F 61666975 7B777678"
		$"79726862 625D5B65 66626667 656C6F6C 70757274 7C7C7873 695D5860 6D7B91A7"
		$"AFAFB0B1 B2BCC5C3 B8AFADAD ADACACAC AAA8A49E 9A9D9B97 989BA0A1 9E9A9289"
		$"878C8985 867E6F69 6C71777F 807D7B77 7A7E7D7F 7D6C6566 6976807D 777E8077"
		$"7B7C7774 6E6B666D 72676162 61626769 6C6B646A 6A5F5E5E 66727173 80837771"
		$"6A5B5F76 87857E70 646E7878 70666664 69808774 6A6A636C 808D7D6F 7C7B7281"
		$"8A817F83 82848381 7D7B7C77 75859696 979A9491 91909699 958E8D99 9C969594"
		$"92959897 97948E89 8483888D 92939394 9291908D 8C8C8785 87848790 8B848381"
		$"81827F79 76797A79 838B8884 7D7D8584 868B837F 8384858A 8C89848C 949097A0"
		$"94898887 88878483 7A809289 8792867D 88877E77 70768184 8385877F 737C8E80"
		$"727F7764 79857C89 93928A82 837A6A64 6461606A 6F7A8881 706D787D 75747266"
		$"61656669 6E6C6D71 80897E71 6E6E7482 827A797A 7677827B 6E737C7E 7E808387"
		$"82746F7A 817D7C7C 7677807C 6E6E7571 75837E79 7A798086 81797572 75818784"
		$"858C8F8E 949C988E 88868079 7C84817E 8487898E 928F8E94 9CA3A8A8 A4A1A3A7"
		$"A59F9E98 91949899 9795979B 9D9FA4A6 9E9CA3A1 9B9A9692 908E8D8A 86847A72"
		$"76767477 7877777A 7C7E7B76 74707269 6679807C 776A6166 676D7D80 746A7180"
		$"7267868E 736A7273 706E7983 80797C96 90675E6A 6C7E8770 626E7C79 73787D81"
		$"7C6E6B74 78746E6F 746D7383 8282837C 767E8B81 7A7F7875 7B807F7A 7D7C7983"
		$"8F918887 87828493 9481848B 7C787E85 88807C7E 7A838F8E 88817980 8682848A"
		$"82727789 88858C89 83888A87 89847D78 777F898E 8E8C8F94 908A888A 8E8F8784"
		$"8B8D8889 8C898A8F 897F7A78 797E8787 8C8F8887 85858A85 7E82898E 8F87878C"
		$"8F8E8784 88888382 8D8A8282 828D9084 7F7E8588 7E7C7F7E 878E8174 75818E8D"
		$"80888E7C 787B746A 68696A78 88827973 636A7B79 7A7D7F7E 786D626B 7166636D"
		$"7F887D74 72707374 7A837477 9391847B 7E807275 86837674 858E817E 827F8087"
		$"83716E6F 67676D76 817E767C 85827E83 807A7A7D 888A8081 82757583 84878983"
		$"83858586 86868B8A 8E8E8489 88797882 80818C88 7A7B8686 85878485 80767983"
		$"817C7C81 88909A93 888E8E82 80867E7D 8483818A 8D899192 888A8584 958F848F"
		$"82768381 7B84837D 80848687 867E767F 82737177 78777E85 7D747D85 7D7E7E7C"
		$"81807D7E 7F7C7C7F 79858F7A 78807B7D 78808476 7D868085 877D7676 76767A84"
		$"837F8886 7C83847A 71727473 74767C7F 7975756E 7C948E87 84796E68 717E838A"
		$"837A8686 7A7D817D 848A898B 90968D7F 82828486 77748985 7A939E84 8591847F"
		$"857A767F 7D70707E 86868580 89877B89 8C767681 84817E82 807C8485 7C76747B"
		$"888A8891 8A78808D 8C837A78 797C7D7E 827E7C88 8F8A8986 83878985 807D7E80"
		$"7B737882 7F7C8188 88818080 7C848C7C 78877D76 807B7D88 7A707F82 7E878075"
		$"7A797982 878C807A 7E7C7B7B 81817B7B 838A8583 89827782 8A858282 7D6D7180"
		$"74748486 868D8A86 8B81747B 817C7A7D 7F7A787A 7C82868C 8D847F84 88847C80"
		$"7E767B82 8A8B8387 8F87878C 7C7C8A8E 867E807B 74767F81 7F81858C 96988672"
		$"7883827E 817D7882 8D95968C 84898984 82807B7B 7E7D7F85 84838C8A 868C837E"
		$"85897D72 81878080 807F7E7E 7C7B7B78 7F8B8986 867F7D80 7F7D7D84 84787D80"
		$"7A85857C 8A8D8380 7B777E88 827D8683 7C7E8081 84838080 82888479 7D7B737B"
		$"7E747275 7C827F7C 7A7E7E80 857E7576 7377888C 887F7D84 85898A88 877D7D83"
		$"81888A7C 76848E84 7E8A8273 7E81787C 8385817B 7A77757B 7A74767C 837C7A84"
		$"857B737B 89807C82 7A7B827E 87897F7D 80878C8E 8A84827F 80837F7F 8482787A"
		$"8A888084 8177777A 7D7D8689 7C7C7F7C 84807680 89818089 84787A80 8581808A"
		$"7C728180 8389878A 817A7E82 8884767A 827A7D86 7E7A848A 86888A7D 727F7D74"
		$"858A776F 7777777F 7A768184 7B7D7D74 7A857A73 7C7A747E 81787A84 7F7C8A8D"
		$"898B8A8A 827B7C7B 7B7D7B82 85808C8C 828C8D86 84817F7A 77808480 81818386"
		$"85817A7B 81838D8C 8387827C 84847C79 81848286 847E8284 888A8689 8C817C86"
		$"89848686 86868183 84828680 848E8280 857C8189 817B8182 81817B7F 827D7B80"
		$"85848282 85888584 837B797C 7B7E7B75 7C80868A 8285877A 7E8A8B84 7E7E7E84"
		$"887D727C 85767A8C 8476797A 787A7A80 7D737A83 847D7B83 8784827E 7D7E7E7F"
		$"84838080 7B828B84 87898584 80847E76 85897A7D 8A877E80 86837C82 8680878C"
		$"8582807C 77777B7A 77828881 7E83827F 807F7E84 847C7A7E 7E7D7D80 837F7E83"
		$"847F818C 8F817B84 7D778688 80847F7D 8A877B7E 837D818C 8880817E 7A7C8086"
		$"88827F87 88838685 78798176 79858180 82827E81 89868284 7E71747B 797B837C"
		$"76858D8B 90908378 7A81807D 807E7883 89848986 7C797B84 827E847E 7B86887E"
		$"7D817F79 77808280 8485847F 7D808384 7D7E8079 7A858484 86828283 8284837C"
		$"7E7E7C7D 78838A7A 75848987 87837C7E 85837A79 817B7780 847F7D7F 848A887F"
		$"7E7B767F 857D7D82 807B7C81 827E8084 82848889 89878482 7F7C7977 7C848581"
		$"868C8883 84878E89 80837A73 7C7F7E81 80828483 8286847C 7B83827B 8386767B"
		$"8A7F8088 7D7C8079 7C888682 86898281 87838186 7E7A8181 80858686 86858587"
		$"87838282 7F80817D 7F7E7D80 7E7E7E7E 81827E7D 80807F82 86827F81 82817F80"
		$"827D7F84 7F7F8281 84858687 827E817A 7B87867F 82858080 837F7F83 7F80837F"
		$"7F7E7C7B 7B808382 85868284 837E7D7C 7C7C7B80 847E8086 84878A83 7E7D7C82"
		$"837F7E7B 7C87837E 88897E7E 827F7E80 80828781 7C82817C 7F838082 827E7B84"
		$"89858383 7F7C7D82 827C7E87 827D7E7D 80828183 85817E81 7E7E8384 82838682"
		$"7C7F827F 807E7A7E 817D7B80 827F7D7E 84848384 7F81847F 7E7F7E7E 8488857E"
		$"8087837F 81808080 7E7E8384 807F7E7B 80858684 83858580 7C797A82 86828085"
		$"827D8388 8684817D 7C7F807E 80868680 81818285 8181807E 7F82817C 7C7E7F80"
		$"7E7E7A78 7E7E7E85 83808281 7E828380 7E7E8084 827F7F83 83808386 86848488"
		$"88807E7F 7D7F817E 7D7F8387 86858581 81808082 7F7D7C7C 7F808283 82818181"
		$"8182827E 82837D85 877F8082 86868484 827F8484 8084807F 84828384 7F7D7C81"
		$"87837D7E 7D7C7D7D 7D858482 89878280 817F797C 81827C7D 827E7E84 8582807E"
		$"7E807C7C 82828280 7F807E80 85828080 82828081 82817E7D 7D7D8081 868C827C"
		$"82817E80 817E7E80 82827E7D 7E7E8387 847D7E82 807D7E81 807E7F82 80808280"
		$"81838386 87828182 81808080 807D7B81 85817F80 82878781 7B7C8385 807D7C7C"
		$"7B7F8682 7D828481 8385817C 7E838684 7F7E7F7F 7C7E807C 7B808482 8186837D"
		$"7E7F7C7D 7F7C7C82 85817F82 7F7D8084 84807F80 7E7C8082 7E7F817E 7D7E7C7E"
		$"7F7E8485 84858382 827E7B7C 7D7C7E7E 8184837E 7D7E7C80 84808082 7E7E8181"
		$"80838481 81808080 80807E7E 817E7D7E 7E7F7E80 8482807E 7C7F8280 7E81857E"
		$"7E848280 83817F7E 7E83847E 7D807E80 83818182 83828184 83807E7E 80808184"
		$"817F8281 81818180 7F7F8281 7D7E7F7F 84878481 83858789 857E7E80 8286847F"
		$"7F7F7F82 86898784 827E8083 7F7E7F7E 81848483 7F7F8080 84848485 807F807E"
		$"7E808282 81808482 81818184 82848481 8084837F 84888584 84808081 827F7D7F"
		$"7D7D8486 84828180 81807E80 807C7C81 84828080 7F828282 82818080 7F828383"
		$"85878685 8481807F 7F7F7E80 82808487 827E8182 81818386 817D7F80 80828483"
		$"82828285 8482817D 7F828283 82808082 8382807F 7F808487 87817C7E 7F808381"
		$"7E7F8082 84837E7F 7F7D8084 84827F80 84868581 81838484 7E7C7D7E 84858182"
		$"84818281 7D7E7F7E 81817D7F 817E7D80 85868685 7F7C7E7D 81868582 81818081"
		$"84868883 82827D7F 85817C7E 82828484 81818180 807D7E83 82817E7B 7E82807F"
		$"81807A7C 81807E81 83807E80 83858582 7E7F8281 82828181 807F7D7C 7F828182"
		$"81818384 85817E80 82827F7A 7C828586 83808080 8180807F 7B7C7E80 84847D80"
		$"85858280 807E7D7E 7E808080 82838081 8382807F 81848584 7E7B7D80 82858786"
		$"817F7E7F 82828081 807E7D80 84828082 807F8282 81808083 84828281 7F7E7F82"
		$"827F7F7F 7C7E8180 81818486 807C7E7F 82858688 857F7D7B 7D82807F 83847F7F"
		$"817F7F81 807C7C7D 7F82807E 80808084 84828283 807D7E80 7F807F7E 7F818281"
		$"80838483 82828483 7F7E7F82 8281807F 80818080 81808181 807E7D7D 7E7F7E7B"
		$"7C7D7C7F 82838285 84808182 8384817E 7F807F81 83848586 87868684 807F8080"
		$"7F80807D 7E808386 86848382 7F7E7F7F 80827E7E 85858182 83808083 81808180"
		$"7E848582 82848381 80808080 7F7E8083 807F827F 80847F7C 7F82807F 807F7E7E"
		$"7F818281 82848081 82808180 7D7D7E7E 7D7E7C7E 82818486 837F8080 7E7E7F7E"
		$"7E82817F 82807D80 84838380 7D7C7B7E 83837F7F 807D7C7F 83828182 8385827F"
		$"80807D7C 7E7E8081 81858786 8683807F 7F7F7A79 7E828082 85848484 82818180"
		$"807E7C7E 8182807F 82807F82 83828182 827F7F81 80807F7F 7E7E8080 82858582"
		$"82828281 7F7F7E7E 7F7D7F7F 7C7D8081 81848683 8284817C 7E808181 8284827E"
		$"7E7F7F81 83828283 83858685 84827F7E 7E7F7E7D 7E808182 83868988 85828180"
		$"8081807E 7F828384 85868483 827E7F81 82848584 85848280 82817E7E 7F7E8186"
		$"85828384 84848484 7E7B7F81 82838182 84838383 81808181 81807F80 82818281"
		$"80818283 827F8182 81817F80 81808184 85848381 7F7F8182 83817F7F 7F808284"
		$"86848284 84818184 857F7F81 80828483 82818384 83828184 827F8181 80808182"
		$"82807F80 8182817E 80828283 82808182 83828180 7F808181 81818281 81828181"
		$"83827F80 7F808283 83858481 83838281 7E7E7D7D 7E7E7F7F 80807F80 82807D7C"
		$"7E7F7F80 8082827F 7F808080 80807E7C 7E818283 82828281 80808283 83807E7E"
		$"7F7E8084 8482827F 7D7C7B7C 7D7E7F7C 7C80807E 80828282 7F7F8180 7F7F8080"
		$"8082817F 81817F81 84858484 84848282 8384827F 7E818280 81838280 7F808486"
		$"84828080 7F7D7E7E 7C7C7F80 80818281 7F818283 827F7F7F 7C7D8080 7F7F8182"
		$"7F808284 84858584 817E7E7F 8082817F 7F808283 82828484 84828080 7D7B7D7E"
		$"80828383 82818281 82827F7E 7E7D7B7C 80828482 82848282 82807E7E 80808282"
		$"81818182 82838483 82838181 83828485 82838584 82807E7D 7F7E7C7A 7A7D7C7A"
		$"797A7A7B 7F7F7E7F 80808183 84838182 86878585 86868482 7E7E8387 85827E7A"
		$"787B8084 837E7B7A 7B7D8283 807F8082 84858481 7F7F7F81 84848081 82807F81"
		$"82828080 81808182 81808081 81808284 84838280 81818081 807F7F7E 7E818281"
		$"7F7F7F7F 807E7D7E 80818182 82807F80 83838382 82838588 87868887 84838381"
		$"7E7B7979 7B7C7A7A 7A797979 7A7D8081 81818281 83868685 84858585 85838282"
		$"82807F80 817E7B7E 80818585 8483817F 80818281 7F7D7D7C 7A7C7F7E 7E7E7E7E"
		$"7E7F8283 83827F7E 80858380 82848484 86878583 817F8181 7E7D7B7A 7C7F7F7E"
		$"80807E7E 82807C7E 807F7F80 80808082 83838280 7F808182 82818282 81828381"
		$"7F7E7F82 83828282 82848484 85838183 82828483 82828182 82828180 82838280"
		$"81828181 8282807F 7E7D7D7F 82828081 83848382 8281817F 80808080 7F7F8284"
		$"85858584 84838382 8183827F 7F808281 81807E80 82807F81 807F7F80 8080807E"
		$"7D7E8081 81828485 82818284 84838280 7F808284 85858481 80828484 8484817F"
		$"81828183 83807E7E 7F808181 7F808181 80808184 83808081 80818384 82828282"
		$"82838382 81808081 82808080 80818182 83838180 80828485 827F7F80 80818282"
		$"817E7E7F 7E7D7D7D 7E818282 83817F7E 7D7D7F7F 7F808081 80818384 82808081"
		$"82828281 807E7E7F 80807F7E 7E808182 8281807F 7F7E7E7E 7F7F8081 82828080"
		$"81808081 807F7F81 817F8081 807F8081 82807E7E 7F818384 84828181 8181807F"
		$"7E7F8082 84838180 80818182 8180807F 7F808283 83828281 8180807F 80828282"
		$"8281807F 80828181 807F8182 82828484 84848281 80807F7F 80828281 807F8083"
		$"84858584 82828281 82828080 81828383 84838180 82818080 7F7E7F81 80818280"
		$"7F7F7F81 86868482 81828283 84848381 7E7D8082 82828180 80818081 82818283"
		$"82807F7F 7F808282 82828182 82818181 81818181 807F7F7F 80818282 8383817E"
		$"7D7E8080 80807F7F 80818282 82828182 83828281 80818282 82828282 82807F7F"
		$"80807F7F 81838382 80808081 81818080 81817F80 83838284 84828181 8081817F"
		$"7F818384 84828181 82828180 81818283 83818080 82828180 80808080 80808080"
		$"82838382 81818181 81828282 80808282 81818181 81818181 81818181 82818284"
		$"83828181 80808180 7F80807F 7F808080 7F808181 807E7D7E 7F808283 827F8082"
		$"81808180 81828280 8181817E 7E808281 80808080 7F7F8080 80808182 81807F7F"
		$"80807F7F 80808182 81828181 80818281 80808181 80808080 80818282 80807F7F"
		$"81818181 81807E80 8181807E 7D7E7F80 7F7F8080 80808182 81808081 81817F7D"
		$"7D7F8080 81818080 81818180 8080807F 7E80807E 7E7F8081 82828281 81818181"
		$"8180807F 8182807E 7E7E7E7E 80818181 82828283 83838280 807F7E7E 7E7F8080"
		$"81818182 82818080 8080807F 807F7E7E 7E7F7F80 82828383 83828282 82828281"
		$"807F8080 81828180 81818283 8381807F 7F808181 82828181 82818182 81807F7F"
		$"7E7E7E7F 7F7F7F80 81828382 81818180 80818181 81818282 83838282 81808080"
		$"807E7D7E 7F808181 80808081 81828484 83828383 81818181 80807F7E 7E7F8181"
		$"82818282 8281807F 7F7F8081 81818282 83848383 83828181 80808080 81808080"
		$"80808182 82828282 82828281 80808080 81818181 807F8082 82818181 80808080"
		$"80818080 80807F80 80818181 81818080 81818080 80808080 81818282 81818080"
		$"80808080 80808080 80808080 7F808080 81818080 80808080 80808181 81818282"
		$"82828282 81808081 80807F80 80808181 81808080 8181807F 7E7E7F7F 7F808080"
		$"80818181 81818181 81808080 80808080 7F7F8080 80818181 80818181 81818181"
		$"81818180 80808180 80808080 8080807F 80808181 80808080 80808080 80808081"
		$"81828180 80808080 80807F7E 7F808080 80808080 80808080 80808080 80808080"
		$"81818181 81818080 807F7F7F 80808080 80808080 80808080 80808080 80808080"
		$"7F7E7E7F 80807F7E 7E7F8080 8180807F 7F808080 80808080 80818282 80808081"
		$"81808080 807F7F80 80818080 80808180 80808080 80808081 80808080 80808080"
		$"8080807F 7F808080 80808080 81818181 80818181 81808080 80808080 7F7F7F80"
		$"80808080 80808080 80818181 80808080 80808080 80808081 81818180 80808080"
		$"80808080 80818181 81818181 81818080 80808080 80808080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80808080"
		$"80808080 807F7F7F 7F7F7F7F 7F7F7F7F 80808080 80808080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808181 81818080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80808080"
		$"80808080 80808080 80808080 80808080 80808080 80808080 80808080 80808080"
	}
};

